/*
 * Aligner.hpp
 *
 */
#ifndef ALIGNER_HPP
#define ALIGNER_HPP

#include "Util.hpp"
#include "DPRegion.hpp"
#include "DPPath.hpp"
#include "DPDeck.hpp"
#include "AlignProbData.hpp"
#include "PairProbData.hpp"

class DPDeck;
class GuideTree;
class Weight;
class ScoreParam;
class DPPath;
class MSA;
class DPRegion;

class Aligner {
public:
  static Aligner& create(const string& type);
  virtual ~Aligner() {}
  int getType() const {return type;}
  virtual string getLogString() const {return "";}

  virtual int getLengthX() const {errorAbort(); return (-1);}
  virtual int getLengthY() const {errorAbort(); return (-1);}
  virtual int getStripWidth() const {errorAbort(); return (-1);}
  virtual int getChunkSize() const {errorAbort(); return (-1);}
  virtual float getScore() const {errorAbort(); return IMPOSSIBLE;}
  virtual void setScoreParam(ScoreParam& scoreParam) {errorAbort();}
  virtual void setScoreParamList(Vector<ScoreParam>& scoreParamList) {errorAbort();}
  virtual void setPurpose(const string& purpose0) {errorAbort();}
  virtual void setScoreMode(const string& modeName) {errorAbort();}
  virtual void setGuideTree(GuideTree& guideTree) {errorAbort();}
  virtual void setWeight(Weight& weights) {errorAbort();}
  virtual void setAlignProb(const NMatrix<float>& alignProb)  {errorAbort();}
  virtual void setPairProbData(const PairProbData& pairProbData){errorAbort();}
  virtual void setAlignProbData(const AlignProbData& AlignProbData){errorAbort();}
  virtual void setSequenceSimilarity(const float& z) {errorAbort();}
  virtual void setRefinementStrategy(const string& strategy) {errorAbort();}
  virtual void setCYKMode(const string& cyk_mode){errorAbort();}
  virtual void setMSAs(const MSA& msaX, const MSA& msaY) {errorAbort();}
  virtual void setMSAs(const MSA& msaX, const MSA& msaY,
		       const UpTriMtx<float>& pairProbX, 
		       const UpTriMtx<float>& pairProbY) {errorAbort();}
  virtual void run() {errorAbort();}
  virtual float computeWorkspaceSizeInMB() {errorAbort(); return IMPOSSIBLE;}
  virtual float computeDPIterationCountInUnitsOfMegaTimes(){errorAbort(); return IMPOSSIBLE;}
  //virtual void printWorkspaceSize(ostream& fo = cout) {errorAbort();}
  //virtual void printDPIterationCounts(ostream& fo =cout) {errorAbort();}
  virtual int getParamVectorSize() const {errorAbort(); return (-1);}
  virtual void setParamVector(const Vector<float>& pv) {errorAbort();}
  virtual const char* getParamNameFromOffset(int offset) const {errorAbort(); return "";}
  virtual Vector<float> computeFeatureVector() const {errorAbort(); return Vector<float>();}
  virtual Vector<float> getParamVector() const {errorAbort(); return Vector<float>();}
  virtual Vector<float> getParamVectorFromCountVector(const Vector<float>& countVector) const {
    errorAbort(); return Vector<float>();
  }
  virtual string featureVect2String(const Vector<float>& fv, const int& indentLevel) const {
    errorAbort(); return "";
  }
  virtual void getAlignProb(NMatrix<float>& alignProb) const {errorAbort();}
  virtual void getPairProb(UpTriMtx<float>& pairProb) const {errorAbort();}
  virtual float getDEDistance() const {errorAbort(); return 0.0;}
  virtual const DPPath& getDPPath() const;

  virtual void setDPPath(const DPPath& dpPath1) {};
  virtual void setDPAlongPath(bool dpAlongPath0) {}
  virtual void setDoRefine(bool doRefine0) {}
  virtual void setSkipSize(int skipSize0) {};
  virtual int getSkipSize() const {errorAbort(); return 0;}
  //virtual void setLockFlag(bool lockFlag0) {};
  virtual void setMaxStripWidth(int maxStripWidth0) {}
  virtual void setChunkSize(int chunkSize0) {}
  virtual void setMemRequested(float memRequested0) {}
  virtual void setKeepMinBandWidth(bool keepMinBandWidth0) {}
  virtual void setMaxLWRatio(float maxLWRatio0) {}
  virtual void setMinLWRatio(float minLWRatio0) {}
  virtual void setBubbleSize(int bubbleSize0) {}
  virtual void setBubbleOverlapRatio(float bubbleOverlapRatio0) {}
  virtual void setShrinkByAlignProb(bool shrinkByAlignProb0) {}
  virtual void setShrinkByAlignProbCutoff(float cutoff0) {}
  virtual void setTimeFactor(float timeFactor0) {}
  virtual void setMaxTime(float maxTime0) {}
  virtual void setSkipSizeExponent(float skipSizeExponent0) {}

protected:
  enum {
    STRIP_PHMM,
    STRIP_PSCFG,
    CHUNK_PHMM,
    CHUNK_PSCFG,
    SCFG,
    MCCASKILL,
    ALIFOLD,
    PROBCONS
  };
  Aligner(int type0) : type(type0) {}

private:
  const int type;

  Aligner(const Aligner& aligner);
  Aligner& operator=(const Aligner& aligner);
  void errorAbort() const {Die("method undefined");}
};

class AlignerStrip : public Aligner {
  typedef Aligner Super;
public:
  AlignerStrip(int type0, DPDeck& dpDeck0);
  ~AlignerStrip();

  float getScore() const;
  const DPPath& getDPPath() const;
  Vector<float> computeFeatureVector() const;
  Vector<float> getParamVector() const;
  Vector<float> getParamVectorFromCountVector(const Vector<float>& countVector) const;
  int getParamVectorSize() const;
  const char* getParamNameFromOffset(int offset) const;
  void setParamVector(const Vector<float>& pv);
  string featureVect2String(const Vector<float>& fv, const int& indentLevel) const;
  string getLogString() const {return dpDeck.getLogString();}

  int getLengthX() const;
  int getLengthY() const;
  int getMinBandWidth() const;
  int getMaxBandWidth() const;
  int getMinStripWidth() const;
  int getMaxStripWidth() const;
  void setDPPath(const DPPath& dpPath1);
  void setSkipSize(int skipSize0);
  int getSkipSize() const;
  //void setLockFlag(bool lockFlag0);
  void setPurpose(const string& purpose0);
  void setScoreMode(const string& modeName);
  void setCYKMode(const string& cyk_mode);
  void setDoRefine(bool doRefine0);
  void setRefinementStrategy(const string& strategy);
  void setSequenceSimilarity(const float& z);
  void setDPAlongPath(bool dpAlongPath0);
  void setMemRequested(float memRequested0);
  void setMaxStripWidth(int maxStripWidth0);
  void setKeepMinBandWidth(bool keepMinBandWidth0);
  void setMaxLWRatio(float maxLWRatio0);
  void setMinLWRatio(float minLWRatio0);
  void setBubbleSize(int bubbleSize0) {bubbleSize = bubbleSize0;}
  void setBubbleOverlapRatio(float bubbleOverlapRatio0) {bubbleOverlapRatio = bubbleOverlapRatio0;}
  void setScoreParam(ScoreParam& scoreParam);
  void setScoreParamList(Vector<ScoreParam>& scoreParamList);
  void setGuideTree(GuideTree& guideTree);
  void setWeight(Weight& weights);
  void setShrinkByAlignProb(bool shrinkByAlignProb0) {shrinkByAlignProb = shrinkByAlignProb0;}
  void setShrinkByAlignProbCutoff(float cutoff0){alignProbCutoff = cutoff0;}
  void setAlignProbCutoff(float alignProbCutoff0){alignProbCutoff = alignProbCutoff0;}
  void setTimeFactor(float timeFactor0){timeFactor = timeFactor0;}
  void setMaxTime(float maxTime0){maxTime = maxTime0;}
  void setSkipSizeExponent(float skipSizeExponent0){skipSizeExponent = skipSizeExponent0;}
  void setAlignProb(const NMatrix<float>& alignProb);
  void setPairProbData(const PairProbData& pairProbData);
  void setAlignProbData(const AlignProbData& AlignProbData);
  void getAlignProb(NMatrix<float>& alignProb) const;
  void getPairProb(UpTriMtx<float>& pairProb) const;
  float getDEDistance() const;
  void setMSAs(const MSA& msaX, const MSA& msaY);
  void setMSAs(const MSA& msaX, const MSA& msaY,
	       const UpTriMtx<float>& pairProbX, 
	       const UpTriMtx<float>& pairProbY);
  void run();

  float computeWorkspaceSizeInMB();
  float computeDPIterationCountInUnitsOfMegaTimes();
  void printWorkspaceSize(ostream& fo = cout);
  void printDPIterationCounts(ostream& fo = cout);
  float getTimeEstimate() const;

  struct GetStripArea {
    const AlignerStrip& aligner;
    const DPPath& dpPath;
    DPRegion dpRegion;

    GetStripArea(const AlignerStrip& aligner)
      : aligner(aligner), dpPath(aligner.dpPath) {}
    float operator()(int width) {// assuming dpPath is set
      dpPath.getStripDPRegion(width, dpRegion);
      return dpRegion.getArea();
    }
  };
  struct GetMemSize {
    AlignerStrip& aligner;
    DPDeck& dpDeck;
    DPRegion& dpRegion;
    DPPath& dpPath;

    GetMemSize(AlignerStrip& aligner)
      : aligner(aligner), dpDeck(aligner.dpDeck),
	dpRegion(aligner.dpRegion), dpPath(aligner.dpPath) {}
    float operator()(int width) {// assuming dpPath is set
      //dpPath.getStripDPRegion(width, dpRegion);
      const Rect& bbox = dpPath.getRect0();
      aligner.setStripDPRegion(dpRegion, dpPath, bbox, width);
      dpDeck.setDPRegion(dpRegion);
      return dpDeck.computeWorkspaceSizeInMB();
    }
  };
  struct GetMemSizeBBox {
    AlignerStrip& aligner;
    DPDeck& dpDeck;
    DPRegion& dpRegion;
    DPPath& dpPath;
    Rect bbox;
    GetMemSizeBBox(AlignerStrip& aligner, const Rect& bbox)
      : aligner(aligner), dpDeck(aligner.dpDeck),
	dpRegion(aligner.dpRegion), dpPath(aligner.dpPath) ,bbox(bbox) {}
    float operator()(int width) {// assuming dpPath is set
      //dpPath.getStripDPRegionBBox(width, bbox, dpRegion);
      aligner.setStripDPRegion(dpRegion, dpPath, bbox, width);
      dpDeck.setDPRegion(dpRegion);
      return dpDeck.computeWorkspaceSizeInMB();
    }
  };
  struct GetTimeEstimateBBox {
    AlignerStrip& aligner;
    DPDeck& dpDeck;
    DPRegion& dpRegion;
    DPPath& dpPath;
    Rect bbox;
    GetTimeEstimateBBox(AlignerStrip& aligner, const Rect& bbox)
      : aligner(aligner), dpDeck(aligner.dpDeck),
	dpRegion(aligner.dpRegion), dpPath(aligner.dpPath) ,bbox(bbox) {}
    float operator()(int width) {// assuming dpPath is set
      //dpPath.getStripDPRegionBBox(width, bbox, dpRegion);
      aligner.setStripDPRegion(dpRegion, dpPath, bbox, width);
      dpDeck.setDPRegion(dpRegion);
      return aligner.getTimeEstimate();
    }
  };


private:
  DPDeck& dpDeck;
  DPRegion& dpRegion;
  DPPath& dpPath;
  int maxStripWidth;
  bool initialDPPathIsSet;
  int stripWidth;
  int skipSize;

  static const int DEFAULT_MEM_SIZE_IN_MB = 1000;
  const MSA* pMSAX;
  const MSA* pMSAY;
  float memRequested;
  bool keepMinBandWidth;
  float maxLWRatio;
  float minLWRatio;
  int bubbleSize;
  float bubbleOverlapRatio;
  bool shrinkByAlignProb;
  const NMatrix<float>* pAlignProb;
  float alignProbCutoff;
  float timeFactor;
  float maxTime;
  float skipSizeExponent;

  void setUpDPRegion();
  void setStripDPRegion(DPRegion& dpRegion1, 
			const DPPath& dpPath1, const Rect& bbox1, const int stripWidth1);
  void shrinkDPRegion(DPRegion& dpRegion1, const DPPath& dpPath1);
  int findOptimalStripWidth();
  void runOnBubbles();
  void runDPRegion(const DPRegion& dpRegion1);
  Rect findOptimalBBox(int x_beg, DPPath& dpPath1);
  int findOptimalStripWidthBBox(Rect bbox);
  int findMaximalStripWidthBBox(Rect bbox);
  void findOptimalSkipSizeStripWidthBBox(int& skipSize1, int& stripWidth1, Rect bbox);
};

class AlignerFold : public Aligner {
  typedef Aligner Super;
public:
  AlignerFold(int type, DPDeck& dpDeck);
  ~AlignerFold();

  float getScore() const;
  const DPPath& getDPPath() const;
  Vector<float> computeFeatureVector() const;
  Vector<float> getParamVector() const;
  Vector<float> getParamVectorFromCountVector(const Vector<float>& countVector) const;
  int getParamVectorSize() const;
  const char* getParamNameFromOffset(int offset) const;
  void setParamVector(const Vector<float>& pv);
  string featureVect2String(const Vector<float>& fv, const int& indentLevel) const;
  string getLogString() const {return dpDeck.getLogString();}

  void setDPPath(const DPPath& dpPath1);
  void setSkipSize(int skipSize0);
  int getSkipSize() const;
  void setPurpose(const string& purpose0);
  void setCYKMode(const string& cyk_mode);
  void setScoreMode(const string& modeName);
  void setDoRefine(bool doRefine0);
  void setRefinementStrategy(const string& strategy);
  void setSequenceSimilarity(const float& z);
  void setDPAlongPath(bool dpAlongPath0);
  void setMemRequested(float memRequested0);
  void setScoreParam(ScoreParam& scoreParam);
  void setScoreParamList(Vector<ScoreParam>& scoreParamList);

  void setGuideTree(GuideTree& guideTree);
  void setWeight(Weight& weights);
  //void setAlignProb(const NMatrix<float>& alignProb);
  void setPairProbData(const PairProbData& pairProbData);
  void setAlignProbData(const AlignProbData& AlignProbData);
  //void getAlignProb(NMatrix<float>& alignProb) const;
  void getPairProb(UpTriMtx<float>& pairProb) const;
  //float getDEDistance() const;

  void setMSAs(const MSA& msaX, const MSA& msaY);
  void setMSAs(const MSA& msaX, const MSA& msaY,
	       const UpTriMtx<float>& pairProbX, 
	       const UpTriMtx<float>& pairProbY);
  void setUpDPRegion();
  void run();

  float computeWorkspaceSizeInMB();
  float computeDPIterationCountInUnitsOfMegaTimes();
  void printWorkspaceSize(ostream& fo = cout);
  void printDPIterationCounts(ostream& fo = cout);

private:
  DPDeck& dpDeck;
  DPPath& dpPath;
  DPRegion& dpRegion;
  bool initialDPPathIsSet;
};

class AlignerChunk : public Aligner {
  typedef Aligner Super;
public:
  AlignerChunk(int type, AlignerStrip& aligner);
  ~AlignerChunk();

  int getChunkSize() const {return chunkSize;}
   float getScore() const;
  const DPPath& getDPPath() const;
 
  Vector<float> computeFeatureVector() const;
  Vector<float> getParamVector() const;
  int getParamVectorSize() const;
  const char* getParamNameFromOffset(int offset) const;
  void setParamVector(const Vector<float>& pv);
  void getAlignProbMtx(AlignProbData& alignProbData);

  void setDPPath(const DPPath& dpPath1) {} //do nothing
  void setScoreParam(ScoreParam& scoreParam);
  //void setScoreParamList(Vector<ScoreParam>& scoreParamList);
  void setGuideTree(GuideTree& guideTree);
  void setWeight(Weight& weights);
  void setPairProbData(const PairProbData& pairProbData);
  void setAlignProbData(const AlignProbData& AlignProbData);
  void getAlignProb(NMatrix<float>& alignProb) const;
  void getPairProb(UpTriMtx<float>& pairProb) const;
  
  void setChunkSize(int chunkSize0) {chunkSize = chunkSize0;}
  void setSkipSize(int skipSize0) {skipSize = skipSize0;}
  //void setLockFlag(bool lockFlag0);
  void setPurpose(const string& purpose0);
  void setDPAlongPath(bool dpAlongPath0);
  //void setDoRefine(bool doRefine0);
  void setRefinementStrategy(const string& strategy);
  void setCYKMode(const string& cyk_mode);
  void run();
  float computeWorkspaceSizeInMB();
  float computeDPIterationCountInUnitsOfMegaTimes();
  void printWorkspaceSize(ostream& fo = cout);
  void printDPIterationCounts(ostream& fo = cout);

private:
  int chunkSize;
  int skipSize;
  int nTailX;
  int nTailY;

  static const int DEFAULT_MEM_SIZE_IN_MB = 1000;
  const MSA* pMSAX;
  const MSA* pMSAY;
  float memRequested;
  bool keepMinBandWidth;
  float maxLWRatio;

  DPPath dpPath;
  AlignerStrip& aligner;

  void setUpAligner();
  int getSkipSizeChunk() const {
    return (skipSize + chunkSize - 1) / chunkSize;
  }
  float getMaxLWRatioChunk() const {
    return (maxLWRatio >= 0.0
	    ? (maxLWRatio /(float)chunkSize)
	    : maxLWRatio);
  }
  int computeNTail(int nCol, int chunkSize) const {
    int n = (nCol + chunkSize - 1) / chunkSize;
    return (n * chunkSize - nCol);
  }
};
#endif

