/*
 * Cluster.hpp
 *
 */
#ifndef CLUSTER_HPP
#define CLUSTER_HPP

class GuideTree;
class Cluster {
public:
  Cluster();
  ~Cluster();

  void setDmtx(Dmtx& dmtx);
  void setDistType(const string& distType);
  void setUseSimilarity(bool use_similarity0) {use_similarity = use_similarity0;}
  const string& getDistTypeName() const;
  void run();
  const GuideTree& getGuideTree() const {return guideTree;}
  string toString() const;
  void print() const;

private:
  static const int NULL_INDEX = (-1);
  enum DistType {
    MIN_DIST,
    AVG_DIST,
    MAX_DIST,
    WAVG_DIST,
    AVG2_DIST
  };
  Dmtx* pDmtx;
  int distType;
  GuideTree guideTree;

  Vector<int> nexts;
  Vector<int> prevs;
  Vector<int> nnbrs;
  Vector<int> ndidx;
  Vector<float> minDists;
  Vector<float> wgts;
  Vector<float> hgts;

  bool use_similarity;

  Cluster(const Cluster& cluster);
  Cluster& operator=(const Cluster& cluster);

  bool isLess(float a, float b) const;
  void buildTreeInit();
  void buildTreeRun();
  float calcDist(float d1, float d2, float d12, float wgt1, float wgt2);
  const string& getDistTypeName(int distType) const;
};


class Cluster_NJ {
public:
  Cluster_NJ();
  ~Cluster_NJ();

  void setDmtx(Dmtx& dmtx);
  void run();
  const GuideTree& getGuideTree() const {return guideTree;}
  string toString() const;
  void print() const;

private:
  Dmtx* pDmtx;
  GuideTree guideTree;

  Cluster_NJ(const Cluster_NJ& cluster);
  Cluster_NJ& operator=(const Cluster_NJ& cluster);
  const string& getDistTypeName(int distType) const;
};

#endif
