/*
 * DPDeck.hpp
 *
 */
#ifndef DP_DECK_HPP
#define DP_DECK_HPP

#include "Util.hpp"
class MSA;
class DPRegion;
class DPPath;
class ScoreParam;
class GuideTree;
class Weight;
class PairProbData;
class AlignProbData;

class DPDeck {
public:
  enum Purpose {
    CYK,
    INSIDE_OUTSIDE,
    LIKELIHOOD
  };
  enum CYKMode {
    CYK_SMALL_CYK_MODE,
    CYK_MODE,
    SMALL_CYK_MODE
  };
  enum ScoreMode {
    SCORE_MODE_SIM_0,
    SCORE_MODE_SIM_1,
    SCORE_MODE_SIM_2,
    SCORE_MODE_SIM_3,
    SCORE_MODE_GUES_BY_SIM,
    SCORE_MODE_MEA
  };

  virtual ~DPDeck();
  virtual string toString() const;
  virtual void print() const;
  virtual string getLogString() const {return "";}

  void setPurpose(const string& purpose0);
  void setCYKMode(const string& cyk_mode);
  void setScoreMode(const string& modeName);

  void setMaxMemory(const float& z);
  void setDPAlongPath(const bool& dpAlongPath0);
  void setDoRefine(const bool& doRefine);

  Purpose getPurpose() const {return purpose;}
  CYKMode getCYKMode() const {return cyk_mode;}

  virtual void setScoreParam(ScoreParam& scoreParam);
  virtual void setScoreParamList(Vector<ScoreParam>& scoreParamList) = 0;
  virtual void setRefinementStrategy(const string& strategy);
  virtual void setSequenceSimilarity(const float& z) {}

  virtual void setSkipSize(int skipSize0);
  virtual int  getSkipSize() const;
  virtual void setGuideTree(GuideTree& guideTree);
  virtual void setWeight(Weight& weights);
  virtual void setPairProbData(const PairProbData& pairProbData);
  virtual void setAlignProbData(const AlignProbData& AlignProbData);

  virtual void setMSAs(const MSA& msaX, const MSA& msaY) = 0;
  virtual void setMSAs(const MSA& msaX, const MSA& msaY,
		       const UpTriMtx<float>& pairProbX,
		       const UpTriMtx<float>& pairProbY);
  virtual void setDPRegion(const DPRegion& dpRegion) = 0;
  virtual void setDPPath(const DPPath& dpPath0) = 0;
  virtual void run() = 0;
  virtual float getScore() const = 0;
  virtual const DPPath& getDPPath() const = 0;

  virtual Vector<float> computeFeatureVector() const;
  virtual Vector<float> getParamVector() const;
  virtual int getParamVectorSize() const;
  virtual Vector<float> getParamVectorFromCountVector(const Vector<float>& countVector) const;
  virtual void setParamVector(const Vector<float>& pv);
  virtual const char* getParamNameFromOffset(int offset) const = 0;
  virtual string featureVect2String(const Vector<float>& fv, const int& indentLevel) const;
  virtual void getPairProb(UpTriMtx<float>& pairProb) const;
  virtual void getAlignProb(NMatrix<float>& alignProb) const;
  virtual void setAlignProb(const NMatrix<float>& alignProb);
  virtual float getDEDistance() const;
  virtual float computeWorkspaceSizeInMB();
  virtual float computeDPIterationCountInUnitsOfMegaTimes() const;
  virtual map<string, float> computeDPIterationCount() const {return map<string, float>();}
  virtual void printWorkspaceSize(ostream& fo);
  virtual void printDPIterationCounts(ostream& fo);
  virtual float getTimeEstimate() const {return 0;}

protected:
  Purpose purpose;
  CYKMode cyk_mode;
  ScoreMode score_mode;
  float maxMemory;
  bool dpAlongPath;
  bool doRefine;

  DPDeck();
private:
  DPDeck(const DPDeck& dpDeck);
  DPDeck& operator=(const DPDeck& dpDeck);
};

#endif
