/*
 * Dmtx.cpp
 *
 */
#include "EScript.hpp"
// #include "UIntMap.hpp"
#include "SeqData.hpp"
#include "MSA.hpp"
// #include "WordTable.hpp"
// #include "SeqIndexes.hpp"
#include "Dmtx.hpp"
#include "Scheduler.hpp"

Dmtx::
Dmtx()
  : Super(), seqIds(*new SeqIds)
{
}
Dmtx::
~Dmtx()
{
  delete &seqIds;
}
int
Dmtx::
length() const
{
  return Super::length();
}
const SeqIds&
Dmtx::
getSeqIds() const
{
  return seqIds;
}
float&
Dmtx::
ref(int i, int j)
{
  return (i <= j ? Super::ref(i, j) : Super::ref(j, i));
}
const float&
Dmtx::
ref(int i, int j) const
{
  return (i <= j ? Super::ref(i, j) : Super::ref(j, i));
}
void
Dmtx::
setLength(int length)
{
  Super::setLength(length);
}
float
Dmtx::
kimuraCorrect(float identity)
{
  const float dist = ((0.2 < identity) 
		      ? (identity < 0.95 ? (-log(identity)) : (-log(0.95))) 
		      : (-log(0.2)));
  Check(dist > 0.0);
  return dist;
}
float
Dmtx::
jukes_cantor_dist(float identity)
{
  const float threshold_low = 0.28;
  const float threshold_high = 0.97;
  const float z = ((identity <= threshold_low)
		   ? threshold_low
		   : ((threshold_high <= identity) ? threshold_high : identity));
  const float dist = (-3.0 / 4.0) * log((4.0 / 3.0) * (z - (1.0 / 4.0)));
  Check(dist > 0.0);
  return dist;
}
// float
// Dmtx::
// computeDistFromTupleCount(SeqIndexes& si1, SeqIndexes& si2)
// {
//   int n = min(si1.swt.nVal, si2.swt.nVal);
//   int c = si1.swt.countCommonKeys(si2.swt);
//   //int n2 = min(si1.pwt.nVal, si2.pwt.nVal);
//   //int c2 = si1.pwt.countCommonKeys(si2.pwt);
//   const double threshold_low = 0.05;
//   float dist = 0.0;
//   if(n == 0) {
//     Warn("sequence with no word found.");
//     dist = 1.0;
//   } else {
//     Check(n >= c);
//     dist = max(threshold_low, (1.0 - c / (float) n));
//   }
//   Check(dist > 0.0);
//   return dist;
// }
// void
// Dmtx::
// setFromTupleCount(const SeqData& seqData, const SeqIds& seqIds0)
// {
//   seqIds = seqIds0;
//   Super::setLength(seqIds.size());
// 
//   Vector<SeqIndexes*> vSeqIndexes(seqIds.size());
//   for (int i = 0; i < (int) vSeqIndexes.size(); i++) {
//     vSeqIndexes[i] = new SeqIndexes(seqData[seqIds[i]]);
//   }
//   for (int i = (int) vSeqIndexes.size() - 1; i >= 0; i--) {
//     SeqIndexes& sii = *vSeqIndexes[i];
//     for (int j = i + 1; j < (int) vSeqIndexes.size(); j++) {
//       SeqIndexes& sij = *vSeqIndexes[j];
//       float dist = computeDistFromTupleCount(sii, sij);
//       Check(dist > 0.0);
//       ref(i, j) =  dist;
//     }
//   }
//   for (int i = 0; i < (int) vSeqIndexes.size(); i++) {
//     delete vSeqIndexes[i];
//   }
// }
void
Dmtx::
setFromSeqIdentity(const MSA& msa)
{
  setLength(msa.size());
  seqIds = msa.getSeqIds();
  for (int i = (int)msa.size() - 1; i >= 0; i--) {
    for (int j = i + 1; j < (int) msa.size(); j++) {
      float seqIdentity = msa.computeSeqIdentity(i, j);
      //float dist = kimuraCorrect(seqIdentity);
      float dist = jukes_cantor_dist(seqIdentity);
      Check(dist > 0.0);
      ref(i, j) = dist;
      //ref(i, j) = max((double)0.05, (double)(1.0 - seqIdentity));
    }
  }
}
void
Dmtx::
setFromMLDistance(SeqData& seqData, 
		  const SeqIds& seqIds0,
		  Scheduler& scheduler)
{
  int clientID = scheduler.createClient();

  seqIds = seqIds0;
  setLength(seqIds.size());
  Job job;
  MSA msaX(seqData);
  MSA msaY(seqData);
  DPPath dpPath;
  float score;
  ::Vector<float> params;
  for (int i = (int)seqIds.size() - 1; i >= 0; i--) {
    for (int j = i + 1; j < (int)seqIds.size(); j++) {
      msaX.setSeq(seqIds[i]);
      msaY.setSeq(seqIds[j]);
      job.set_CalcMLDistance(msaX, msaY, dpPath, score, params);
      scheduler.addJob(clientID, job);
    }
  }
  for (int i = (int)seqIds.size() - 1; i >= 0; i--) {
    for (int j = (i + 1); j < (int)seqIds.size(); j++) {
      scheduler.getResult(clientID, job);
      job.get_CalcMLDistance(msaX, msaY, dpPath, score, params);
      const int& seqId1 = msaX.getSeqId(0);
      const int& seqId2 = msaY.getSeqId(0);
      const int& i1 = seqIds.getIndex(seqId1);
      const int& i2 = seqIds.getIndex(seqId2);
      const float& distance = params[0];
      Check(distance > 0.0);
      ref(i1, i2) = distance;
    }
  }
  scheduler.finish(clientID);
  scheduler.removeClient(clientID);
}
void
Dmtx::
setFromMLDistance(MSA& msa, Scheduler& scheduler)
{
  int clientID = scheduler.createClient();
  seqIds = msa.getSeqIds();
  setLength(seqIds.size());
  Job job;
  MSA msaX(msa);
  MSA msaY(msa);
  DPPath dpPath;
  float score;
  ::Vector<float> params;
  for (int i = (int)seqIds.size() - 1; i >= 0; i--) {
    for (int j = i + 1; j < (int)seqIds.size(); j++) {
      msaX.setSeq(seqIds[i]);
      msaY.setSeq(seqIds[j]);
      dpPath.setFromMSAs(msaX, msaY);
      job.set_CalcMLDistance(msaX, msaY, dpPath, score, params);
      scheduler.addJob(clientID, job);
    }
  }
  for (int i = (int)seqIds.size() - 1; i >= 0; i--) {
    for (int j = (i + 1); j < (int)seqIds.size(); j++) {
      scheduler.getResult(clientID, job);
      job.get_CalcMLDistance(msaX, msaY, dpPath, score, params);
      const int& seqId1 = msaX.getSeqId(0);
      const int& seqId2 = msaY.getSeqId(0);
      const int& i1 = seqIds.getIndex(seqId1);
      const int& i2 = seqIds.getIndex(seqId2);
      const float& distance = params[0];
      Check(distance >= 0.0);
      ref(i1, i2) = distance;
    }
  }
#ifdef DEBUG
  print();
#endif

  scheduler.finish(clientID);
  scheduler.removeClient(clientID);
}
void
Dmtx::
setFromAlignDistance(SeqData& seqData, 
		     const SeqIds& seqIds0,
		     Scheduler& scheduler)
{
  int clientID = scheduler.createClient();
  seqIds = seqIds0;
  setLength(seqIds0.size());
  Job job;
  MSA msaX(seqData);
  MSA msaY(seqData);
  DPPath dpPath;
  float score;
  ::Vector<float> params;
  for (int i = (int)seqIds.size() - 1; i >= 0; i--) {
    for (int j = i + 1; j < (int)seqIds.size(); j++) {
      msaX.setSeq(seqIds[i]);
      msaY.setSeq(seqIds[j]);
      job.set_CalcAlignDistance(msaX, msaY, dpPath, score, params);
      scheduler.addJob(clientID, job);
    }
  }
  for (int i = (int)seqIds.size() - 1; i >= 0; i--) {
    for (int j = (i + 1); j < (int)seqIds.size(); j++) {
      scheduler.getResult(clientID, job);
      job.get_CalcAlignDistance(msaX, msaY, dpPath, score, params);
      const int& seqId1 = msaX.getSeqId(0);
      const int& seqId2 = msaY.getSeqId(0);
      const int& i1 = seqIds.getIndex(seqId1);
      const int& i2 = seqIds.getIndex(seqId2);
      //const float& seqIdentity = params[0];
      //float dist = jukes_cantor_dist(seqIdentity);
      //float score1 = params[1];//phmm score
      //float dist = score1 / (float)min(msaX[0].size(), msaY[0].size());
      float dist = score / (float)min(msaX[0].size(), msaY[0].size()); // maximal expected accuracy score <- "similarity" measure (it can be larger than 1)
      //cout << "[ " << i1 << ", " << i2 << " ]: " << distance << endl;
      ref(i1, i2) = dist;
    }
  }
#ifdef DEBUG
  print();
#endif
  scheduler.finish(clientID);
  scheduler.removeClient(clientID);
}
void
Dmtx::
setFromMSA(const MSA& msa, Scheduler& scheduler)
{
  int clientID = scheduler.createClient();
  seqIds = msa.getSeqIds();
  setLength(seqIds.size());
  Job job;
  MSA msaX(msa);
  MSA msaY(msa);
  DPPath dpPath;
  float score;
  ::Vector<float> params;
  for (int i = (msa.size() - 1); i >= 0; i--) {
    for (int j = (i + 1); j < (int)msa.size(); j++) {
      msaX.setSeq(msa.getSeqId(i));
      msaY.setSeq(msa.getSeqId(j));
      dpPath.setFromMSAs(msaX, msaY);
      job.set_CalcAlignDistance(msaX, msaY, dpPath, score, params);
      scheduler.addJob(clientID, job);
    }
  }
  for (int i = (msa.size() - 1); i >= 0; i--) {
    for (int j = (i + 1); j < (int)msa.size(); j++) {
      scheduler.getResult(clientID, job);
      job.get_CalcAlignDistance(msaX, msaY, dpPath, score, params);
      const int& seqId1 = msaX.getSeqId(0);
      const int& seqId2 = msaY.getSeqId(0);
      const int& i1 = seqIds.getIndex(seqId1);
      const int& i2 = seqIds.getIndex(seqId2);
      //const float& seqIdentity = params[0];
      //float dist = jukes_cantor_dist(seqIdentity);
      // maximal expected accuracy score <- "similarity" measure
      float dist = score / (float)min(msaX[0].size(), msaY[0].size());
      // float score1 = params[1]//phmm score
      //cout << "[ " << i1 << ", " << i2 << " ]: " << distance << endl;
      ref(i1, i2) = dist;
    }
  }
#ifdef DEBUG
  print();
#endif
  scheduler.finish(clientID);
  scheduler.removeClient(clientID);
}
float
Dmtx::
getMeanDistance() const
{
  if (empty()) return 0.0;

  float z = 0.0;
  for (int i = 0; i < length(); i++) {
    for (int j = (i + 1); j < length(); j++) {
      z += ref(i, j);
    }
  }
  z /= (float)(0.5 * length() * (length() - 1));
  Check(z > 0.0);
  return z;
}
string
Dmtx::
toString() const 
{
  ostringstream oss;
  oss << setprecision(1);
  oss << "length:" << length() << '\n';
  for (int i = 0; i < length(); i++) {
    for (int j = 0; j < length(); j++) {
      if (j <= i) {
	oss << "xxxx";
      } else {
	oss << setw(4) << ref(i, j);
      }
      oss << (j == length() - 1 ? '\n' : ' ');
    }
  }
  return oss.str();
}
void
Dmtx::
print() const
{
  cout << toString() << flush;
}



