/*
 * JobRunner.hpp
 *
 */
#ifndef JOB_RUNNER_HPP
#define JOB_RUNNER_HPP


// #include "EvalAlign.hpp"
#include "Job.hpp"
#include "DPPath.hpp"
#include "MSA.hpp"

class ScoreParam;
class Aligner;
class GuideTree;
class SeqData;
class Weight;
class PairProbData;
class AlignProbData;

class JobRunner {
public:
  JobRunner(SeqData& seqData, 
	    Vector<ScoreParam>& scoreParamList,
	    GuideTree& guideTree,
	    Weight& weights,
	    Aligner& aligner,
	    Aligner& aligner2,
	    Aligner& folder,
	    Aligner& folder2,
	    PairProbData& pairProbData,
	    AlignProbData& alignProbData,
	    bool pair_prob_from_msa = false);
  JobRunner(const JobRunner& jobRunner);
  ~JobRunner();
  string toString() const;
  void print() const;
  void setJob(const Job& job);
  Job getJob() const;
  bool hasJob() const;
  void run();

private:
  SeqData& seqData;
  Vector<ScoreParam>& scoreParamList;
  GuideTree& guideTree;
  Weight& weights;
  Aligner& aligner;
  Aligner& aligner2;
  Aligner& folder;
  Aligner& folder2;
  PairProbData& pairProbData;
  AlignProbData& alignProbData;
  
  Job job;

  MSA msaX;
  MSA msaY;
  DPPath dpPath;
  float score;
  Vector<float> params;

  // EvalAlign evalAlign;
  SeqData seqDataSbj;
  MSA msaXSbj;
  MSA msaYSbj;
  MSA msaRef;

  bool use_pair_prob;
  bool pair_prob_from_msa;

  JobRunner();
  JobRunner& operator=(const JobRunner& jobRunner);

  void align();
  void align1();
  void align2();
  void align3();
  void refine();
  void refine2();
  void merge();
  void calcLL();
  void calcSPS();
  void calcFDF();
  void calcEvalScores();
  void fold();
  void fold2Step();
  void testParam();
  void calcPairProb();
  void bcastPairProb();
  void calcAlignProb();
  void bcastAlignProb();
  void smoothAlignProb();
  void calcMLDistance();
  void calcAlignDistance();
  void calcFeatureVect();
  void getPairProbs(UpTriMtx<float>& pairProbX, UpTriMtx<float>& pairProbY,
		    const MSA& msaX, const MSA& msaY);
  float calcSCI(const MSA& msaX1, const MSA& msaY1, const DPPath& dpPath1);
};

#endif
