/*
 * MSA.hpp
 *
 */
#ifndef MSA_HPP
#define MSA_HPP

#include "SeqData.hpp"
class EScript;
class PosMap;
class DPPath;

class MSA : private SeqIds {
  typedef SeqIds Super;
public:
  MSA(SeqData& seqData);
  MSA(SeqData& seqData, const SeqIds& seqIds);
  MSA(const MSA& msa);
  ~MSA();
  MSA& operator=(const MSA& msa);
  using Super::const_iterator;
  using Super::size;
  using Super::empty;
  const_iterator begin() const {return Super::begin();}
  const_iterator end() const {return Super::end();}
  void clear();
  void push_back(SeqId id0);
  template <typename InputIterator>
  void assign(InputIterator begin0, InputIterator end0);

  void setSeq(SeqId id0);
  void setAllSeqs();
  void setByNames(const MSA& msa);
  void setSeqs(const SeqIds& seqIds);
  const string& getName(int index) const;
  Seq& getSeq(int index);
  const Seq& getSeq(int index) const;
  const Seq& operator[](int index) const {return getSeq(index);}
  SeqId getSeqId(int index) const {return Super::operator[](index);}
  const Seq& getSeqFromSeqId(SeqId id) const;
  const SeqIds& getSeqIds() const {return *this;}
  const Vector<char>& getUngappedSeq(int i) const;
  void merge(const MSA& msa);
  
  string getAC() const {return seqData.getAC(*this);}
  string getID() const {return seqData.getID(*this);}
  int getNCol() const;
  bool isGapCol(int i) const;
  bool isBaseCol(int i) const;
  int getUngappedLength() const;
  const PosMap& getPosMap() const {return posMap;}
  void merge(MSA& msa, const DPPath& dpPath);
  void edit(const EScript& eScript);

  void writeStockholm(ostream& fo, bool withSSCons = false) const;
  void writeStockholm(ostream& fo, const string& sscons) const;
  void writeStockholm(ostream& fo, const string& sscons, const float& score) const;
  void writeClustal(ostream& fo) const;
  void writeFasta(ostream& fo, const string& sscons) const;

  float computeSeqIdentity(int index1 , int index2) const;
  float computeMeanSeqIdentity() const;
  float computeMeanSeqLength() const;
  float computeSeqIdentity(const MSA& msa) const;
  float computeTupleCountSimilarity(const MSA& msa) const;
  float computeGCFraction() const;
  void removeGapCols();
  string getSSCons() const;

  string toString() const;
  void print() const;
  void testConsistency() const;
  void resetPosMap();
  void sortSeqId();
  
private:
  SeqData& seqData;
  PosMap& posMap;

  MSA();
};

template <typename InputIterator>
void
MSA::
assign(InputIterator begin0, InputIterator end0)
{
  Super::assign(begin0, end0);
#ifdef DEBUG
  if (!seqData.nColsAreEqual(*this)) {
    cout << "error: nColsAreNotEqual" << endl;
    seqData.writeStockholm(cout, getSeqIds());
    Die("nColsAreNotEqual size=", size());
  }
#endif
  resetPosMap();
}

#endif

