/*
 * MultiFamAligner.cpp
 *
 */
#include "Util.hpp"
#include "Alpha.hpp"
#include "Vector.hpp"
#include "Rect.hpp"
#include "SeqData.hpp"
#include "Profile.hpp"
#include "MSA.hpp"
#include "ScoreParam.hpp"
#include "EScript.hpp"
#include "Dmtx.hpp"
#include "Weight.hpp"
#include "GuideTree.hpp"
#include "Cluster.hpp"
#include "Trapezoid.hpp"
#include "DPPath.hpp"
#include "DPRegion.hpp"
#include "DPDeck.hpp"
#include "Aligner.hpp"
#include "Job.hpp"
// #include "EvalAlign.hpp"
#include "Scheduler.hpp"
#include "SeqFamAligner.hpp"
#include "ProgressiveAligner.hpp"
// #include "IterativeAligner.hpp"
#include "MultipleAligner.hpp"
#include "MultiFamAligner.hpp"

//pthread_mutex_t MultiFamAligner::mutex = PTHREAD_MUTEX_INITIALIZER;
static MultiFamAligner* pMultiFamAligner = NULL;
// static struct sigaction act;
// static void sigint_handler(int sig)
// {
//   cout << "msg: reporting intermediate results" << endl;
//   pMultiFamAligner->printReport();
//   cout << "msg: progrram interrupted by SIGINT" << endl;
//   _exit(EXIT_FAILURE);// segfault if exit() is used
// }
// static void set_sigaction()
// {
//   act.sa_handler = sigint_handler;
//   sigemptyset(&act.sa_mask);
//   sigaddset(&act.sa_mask, SIGINT);
//   act.sa_flags = SA_RESETHAND;
//   sigaction(SIGINT, &act, NULL);
// }

void*
MultiFamAligner_Client_run(void* pMultiFamAligner)
{
  MultiFamAligner& multiFamAligner = *(MultiFamAligner*) pMultiFamAligner;
  MultiFamAligner::Client client(multiFamAligner);
  client.run();
  return NULL;
}

MultiFamAligner::
MultiFamAligner(SeqData& seqData, SeqFamAligner& seqFamAligner,
		int clientSize, bool pairwise, ostream& fo, ostream& fl)
  : seqData(seqData),
    seqFamAligner(seqFamAligner),
    //clientSize(clientSize),
    familyIterator(seqData),
    souts(),
    slogs(),
    pairwise(pairwise),
    fo(fo),
    fl(fl)
{
  pMultiFamAligner = this;
  // set_sigaction();

  SeqIds seqIds;
  SeqIds seqIds1;
  SeqData::FamilyIterator fit(seqData);
  while (fit.getNext(seqIds)) {
    Check(seqIds.size() >= 2);
    if (pairwise) {
      seqIds1.clear();
      seqIds1.push_back(seqIds[0]);
      seqIds1.push_back(seqIds[1]);
    } else {
      seqIds1 = seqIds;
    }
    souts[seqIds1];
    slogs[seqIds1];
  }
}

MultiFamAligner::
~MultiFamAligner()
{
}

void
MultiFamAligner::
run()
{
  //Vector<pthread_t> vThread(clientSize);
  //for (int i = 0; i < clientSize; i++) {
  //  int err = pthread_create(&vThread[i], NULL,
  //			     MultiFamAligner_Client_run, (void*)this);
  //  if (err) {
  //    //pthread_join(vThread[i], (void**)0);
  //    vThread.resize(i);
  //    break;
  //  }
  //}
  //if (vThread.empty()) Die("cannot create clients");

#ifdef DEBUG
  //cout << vThread.size() << " clients started" << endl;
#endif
  //for_each(vThread.begin(), vThread.end(),
  //	   bind2nd(ptr_fun(pthread_join), (void**)0));

  MultiFamAligner_Client_run((void*)this);
  printReport();
}

void
MultiFamAligner::
printReport() const
{
  SeqIds seqIds;
  SeqIds seqIds1;
  SeqData::FamilyIterator fit(seqData);
  while (fit.getNext(seqIds)) {
    if (pairwise) {
      seqIds1.clear();
      seqIds1.push_back(seqIds[0]);
      seqIds1.push_back(seqIds[1]);
    } else {
      seqIds1 = seqIds;
    }
    fo << souts.find(seqIds1)->second;
    fl << slogs.find(seqIds1)->second;
  }
  fo << flush;
  fl << flush;
  // cout << "printing report end" << endl;
}

MultiFamAligner::Client::
Client(MultiFamAligner& multiFamAligner)
  : multiFamAligner(multiFamAligner)
{
}

MultiFamAligner::Client::
~Client()
{
}

void
MultiFamAligner::Client::
run()
{
  SeqIds seqIds;
  SeqIds seqIds1;
  SeqFamAligner& seqFamAligner = multiFamAligner.seqFamAligner.dup();
  //pthread_mutex_lock(&multiFamAligner.mutex);
  while (multiFamAligner.familyIterator.getNext(seqIds)) {
    if (multiFamAligner.pairwise) {
      seqIds1.clear();
      seqIds1.push_back(seqIds[0]);
      seqIds1.push_back(seqIds[1]);
    } else {
      seqIds1 = seqIds;
    }
    seqFamAligner.setSeqIds(seqIds1);
    seqFamAligner.setSOut(multiFamAligner.souts[seqIds1]);
    seqFamAligner.setSLog(multiFamAligner.slogs[seqIds1]);
    //pthread_mutex_unlock(&multiFamAligner.mutex);

    seqFamAligner.run();

    //pthread_mutex_lock(&multiFamAligner.mutex);
  }
  //pthread_mutex_unlock(&multiFamAligner.mutex);
  delete &seqFamAligner;
}
