/*
 * Opt.hpp
 *
 */
#ifndef OPT_HPP
#define OPT_HPP

class Opt {
public:
  enum ArgumentType {
    NO_ARGUMENT       = 0,
    REQUIRED_ARGUMENT = 1
  };
  struct Entry {
    const char* name;
    int type;
    const char* dflt;
  };
  Opt(const Entry* const optTable);
  ~Opt();
  void setValue(const string& key, const string& value);
  void setFromFile(const string& filename);
  void setFromArgs(int& argc, char**& argv);
  const string& getStr(const string& name) const;
  int getInt(const string& name) const;
  double getDbl(const string& name) const;
  bool getBool(const string& name) const;
  void checkIntLowerBound(const string& name, int minValue) const;
  void checkIntUpperBound(const string& name, int maxValue) const;
  void checkIntRange(const string& name, int minValue, int maxValue) const;
  void checkDblLowerBound(const string& name, double minValue) const;
  void checkDblUpperBound(const string& name, double maxValue) const;
  void checkDblRange(const string& name,
			double minValue, double maxValue) const;
  void checkStrEntry(const string& name, const string& entries) const;
  void printHeader(ostream& fo) const;
  void printYAML(ostream& fo) const;
  void setFromYAML(istream& fi);
  string toString() const;
  void print() const;

private:
  struct EntryData {
    EntryData(); 
    EntryData(const Entry& entry);
    EntryData(const EntryData& entryData);
    ~EntryData();
    EntryData& operator=(const EntryData& entryData);

    bool requiredArgument() const {return type == Opt::REQUIRED_ARGUMENT;}
    bool valueIsSet() const {return valIsSet;}
    void setValue(const string& value0);
    const string& getName() const {return name;}
    const string& getValue() const;
    string toString() const;
    void print() const;
  private:
    string name;
    int type;
    string dflt;
    string value;
    bool valIsSet;
  };
  typedef map<string, EntryData> Table;
  string fname;
  Table table;

  bool hasKey(const string& name) const;
};

#endif
