/*
 * BasePairData.hpp
 *
 */
#ifndef PAIR_PROB_DATA_HPP
#define PAIR_PROB_DATA_HPP

#include "Util.hpp"
#include "SparseProb.hpp"

class ScoreParam;
class SeqIds;
class SeqData;
class McCaskill;
class GuideTree;
class Weight;
class Aligner;
class Scheduler;

class 
PairProbData 
{
  typedef SparseProb ValueT ;
  typedef Array<ValueT> Data;
public:
  PairProbData(SeqData& seqData);
  ~PairProbData();

  bool empty() const {return data.empty();}
  int size() const {return data.size();}
  ValueT& ref(int id) {return data[seqIds.getIndex(id)];}
  const ValueT& ref(int id) const {return data[seqIds.getIndex(id)];}
  void run(Scheduler& scheduler);
  void bcastPairProbs(Scheduler& scheduler);
  void getPairProb(UpTriMtx<float>& pairProb, 
		   const Weight& weights1, const MSA& msa) const;

private:
  SeqData& seqData;
  SeqIds& seqIds;
  GuideTree& guideTree;
  Weight& weights;
  Data data;
  float cutoff;
 
  PairProbData();
  PairProbData(const PairProbData& pairProbData);
  PairProbData& operator=(const PairProbData& pairProbData);
};

#endif
