/*
 * Profile.hpp
 *
 */
#ifndef PROFILE_HPP
#define PROFILE_HPP

class DPPath;
class EScript;
class MSA;
class Weight;
class ScoreParam;
class BasePairData;

class Profile {
public:
  struct FCol;
  struct FPairFull;
  struct FPairReduced;
  struct FStack;
  struct FDimer;

#if defined(FPAIR_FULL)
  typedef FPairFull FPair;
#elif defined(FPAIR_REDUCED)
  typedef FPairReduced FPair;
#else
  typedef FStack FPair;
#endif
  
  struct VFCol;
  struct VFPair;
  struct VFDimer;

  VFCol& vFCol;
  VFPair& vFPair;
  VFDimer& vFDimer;
  Matrix<char> seqs;

  Profile();
  Profile(const Profile& profile);
  ~Profile();
  Profile& operator=(const Profile& profile);
  string toString() const;
  void print() const;

  void resize(int newLength);
  void clear();
  void clean();
  
  int length() const;
  void setFromMSA(const MSA& msa, const Weight& weights);
  void setLODScores();
  void priorify();

  const MSA& getMSA() const {Assert(pMSA); return *pMSA;}
  float getBaseWeightAt(int i) const;
  float getGapWeightAt(int i) const;
  float getWeight() const;
  float getLGOccupancy(int i) const;
  float getGLOccupancy(int i) const;
  void chunk(int chunkSize);
  void normalize();
  int getNSeq() const {return seqs.getRowLength();}

  float getCanFrac(const int& b, const int& e) const;
  
  typedef Vector<const float*> VPc;
  typedef Vector<const float*> VPp;
  static const float* getLogPc(int baseCode);
  static const float* getLogPp(int reducedPairCode);
  VPc getVPc(int i) const;
  VPp getVPp(int i, int j) const;

  //const Weight* getWeights() const {return pWeights;}
  //void editFreqs(const EScript& es); 
  //void mergeFreqs(const Profile& profile);
  //void merge(Profile& profile, const DPPath& dpPath);
  void testConsistency() const;
private:
  const MSA* pMSA;
  const Weight* pWeights;
};

#include "ProfileFreqs.hpp"
#include "ProfileVectors.hpp"

#endif


