/*
 * ProgressiveAligner.hpp
 *
 */
#ifndef PROGRESSIVE_ALIGNER_HPP
#define PROGRESSIVE_ALIGNER_HPP

class ProgressiveAligner 
  : public SeqFamAligner 
{
public:
  ProgressiveAligner(SeqData& seqData,
		     GuideTree& guideTree,
		     Weight& weights,
		     Scheduler& scheduler,
		     int clientSize,
		     string outfile_format = "STOCKHOLM");
  ~ProgressiveAligner();
  ProgressiveAligner& dup() const;
  void run();

private:
  typedef SeqFamAligner Super;
  enum Status {
    TOBEDONE = 0,
    OCCUPIED = 1,
    FINISHED = 2
  };
  class Client;
  //int clientSize;

  GuideTree guideTree;
  Weight weights;
  Vector<Status> vStatus;
  Vector<MSA> vMSA;
  //Vector<pthread_t> vThread;
  //static pthread_mutex_t mutex;
  string outfile_format;

  ProgressiveAligner();
  ProgressiveAligner(const ProgressiveAligner& progressiveAligner);
  ProgressiveAligner&
  operator=(const ProgressiveAligner& progressiveAligner);
  friend void* ProgressiveAligner_Client_run(void* pProgressiveAligner);
  friend class MultipleAligner;

  string get_ncrna_web_server_format_string(const GuideTree& guideTree,
					    const MSA& msa, const CT& ct, 
					    const float& score0) const;
};

class 
ProgressiveAligner::Client
{
public:
  Client(ProgressiveAligner& progressiveAligner);
  ~Client();
  void run();

private:
  ProgressiveAligner& progressiveAligner;
  int clientID;
  MSA msaX;
  MSA msaY;
  DPPath dpPath;
  float score;
  Vector<float> params;

  Client();
  Client(const Client& client);
  Client& operator=(const Client& client);
  friend class ProgressiveAligner;
};

#endif
