/*
 * ScoreParam.hpp
 *
 */
#ifndef SCORE_PARAM_HPP
#define SCORE_PARAM_HPP

#include "Util.hpp"
#include "Alpha.hpp"
// #include "EnergyParam.hpp"
#include "pscfg/PSCFG_Energy.hpp"
// #include "scfg/SCFG_Energy.hpp"
// #include "phmm/PHMM_Domain.hpp"

// class EnergyParam;
class ScoreParam {
public:
  enum Type {
    FLOAT = 0,
    VECTOR = 1,
    MATRIX = 2,
    SYMMTX = 3,
    ASMMTX = 4,
    STACK  = 5,
    INVALID_TYPE = -1
  };
  enum Param {
    //PSCFG
    PSCFG_ESC0,              
    PSCFG_TSC0,        
      
    PSCFG_PAIR_PROB_CUTOFF,  
    PSCFG_PAIR_PROB_SCALE,   
    PSCFG_PAIR_PROB_COEFF,
    PSCFG_PAIR_PROB_NSEQ_COEFF,
    PSCFG_LOOP_PROB_SCALE,
    PSCFG_LOOP_PROB_COEFF,
    PSCFG_ALN_PROB_CUTOFF,   
    PSCFG_ALN_PROB_SCALE, 
    PSCFG_ALN_PROB_COEFF,
    PSCFG_ALN_PROB_STEM_SCALE, 
    PSCFG_ALN_PROB_STEM_COEFF,
    PSCFG_GAP_PROB_SCALE,
    PSCFG_GAP_PROB_COEFF,

    PSCFG_ESC_STEM,   
    PSCFG_ESC_FCOL,
    PSCFG_ESC_FPAIR,
    PSCFG_ESC_LOD_SCORE,
    PSCFG_ESC_LOOP,          
    PSCFG_GAP_EXTEND,        
    PSCFG_GAP_EXTEND_OCC,    
    PSCFG_PI_LOOP,           
    PSCFG_PI_STEM,           
    PSCFG_RATE_PARAM_LOOP,     
    PSCFG_RATE_PARAM_STEM,     
    PSCFG_PHY_LOOP_SCALE,
    PSCFG_PHY_STEM_SCALE,
    PSCFG_SMTX_LOOP,         
    PSCFG_PAIR_SMTX,    
    PSCFG_SMTX_LOOP_SIM,         
    PSCFG_PAIR_SMTX_SIM,    
    PSCFG_SMTX_LOOP_NSEQ,         
    PSCFG_PAIR_SMTX_NSEQ,    
    PSCFG_SMTX_LOOP_ASYM,         
    PSCFG_PAIR_SMTX_ASYM,    
    PSCFG_SMTX_LOOP_GC,         
    PSCFG_PAIR_SMTX_GC,    
    PSCFG_STACK_MTX,         
    PSCFG_GAP_OPEN,          
    PSCFG_GAP_OPEN_OCC,      
    PSCFG_GAP_OPEN_OCC_CROSS,
    PSCFG_TSC_MP_LOOP,       
    PSCFG_TERM_STEM,         
    PSCFG_MIN_BASE_PAIR,
    PSCFG_MIN_CAN_FRAC,
    PSCFG_SEQ_SIM_BASE,

    PSCFG_TURNER_STACK_SCALE,
    PSCFG_TURNER_STACK_SCALE_SIM,
    PSCFG_TURNER_TSTACKH_SCALE,
    PSCFG_TURNER_TSTACKI_SCALE,
    PSCFG_TURNER_TSTACKM_SCALE,
    PSCFG_TURNER_TERM_AU_SCALE,
    PSCFG_TURNER_LOOP_LEN_SCALE,
    PSCFG_TURNER_LOOP_LEN_C0,
    PSCFG_TURNER_LOOP_LEN_C1,

    PSCFG_SMTX_LOOP_SCALE      ,
    PSCFG_PAIR_SMTX_SCALE      ,
    PSCFG_SMTX_LOOP_SIM_SCALE  ,
    PSCFG_PAIR_SMTX_SIM_SCALE  ,
    PSCFG_STACK_MTX_SCALE      ,
    PSCFG_RATE_PARAM_LOOP_SCALE,
    PSCFG_RATE_PARAM_STEM_SCALE,

#if false
    //SCFG
    SCFG_ESC0,             
    SCFG_TSC0,         
    
    SCFG_PAIR_PROB_CUTOFF, 
    SCFG_PAIR_PROB_SCALE,
    SCFG_PAIR_PROB_COEFF,
    SCFG_LOOP_PROB_SCALE,
    SCFG_LOOP_PROB_COEFF,
    
    SCFG_ESC_STEM,  
    SCFG_ESC_FCOL,
    SCFG_ESC_FPAIR,
    SCFG_ESC_LOD_SCORE,  
    SCFG_ESC_LOOP,         
    SCFG_PI_LOOP,          
    SCFG_PI_STEM,          
    SCFG_RATE_PARAM_LOOP,    
    SCFG_RATE_PARAM_STEM,  
    SCFG_PHY_LOOP_SCALE,
    SCFG_PHY_STEM_SCALE,
    SCFG_SMTX_LOOP,        
    SCFG_PAIR_SMTX,       
    SCFG_SMTX_LOOP_SIM,        
    SCFG_PAIR_SMTX_SIM,       
    SCFG_SMTX_LOOP_GC,        
    SCFG_PAIR_SMTX_GC,       
    SCFG_STACK_MTX,        
    SCFG_TSC_MP_LOOP,      
    SCFG_TERM_STEM,        
    SCFG_MIN_BASE_PAIR,
    SCFG_MIN_CAN_FRAC,
    SCFG_MIN_PSCORE,
    SCFG_PSCORE_SCALE,
 
    SCFG_TURNER_STACK_SCALE,
    SCFG_TURNER_TSTACKH_SCALE,
    SCFG_TURNER_TSTACKI_SCALE,
    SCFG_TURNER_TSTACKM_SCALE,
    SCFG_TURNER_TERM_AU_SCALE,
    SCFG_TURNER_LOOP_LEN_SCALE,
    SCFG_TURNER_LOOP_LEN_C0,
    SCFG_TURNER_LOOP_LEN_C1,

    SCFG_SMTX_LOOP_SCALE      ,
    SCFG_PAIR_SMTX_SCALE      ,
    SCFG_SMTX_LOOP_SIM_SCALE  ,
    SCFG_PAIR_SMTX_SIM_SCALE  ,
    SCFG_STACK_MTX_SCALE      ,
    SCFG_RATE_PARAM_LOOP_SCALE,
    SCFG_RATE_PARAM_STEM_SCALE,

    //PHMM  add Stem part
    PHMM_ESC0,               
    PHMM_TSC0,  

    PHMM_PAIR_PROB_CUTOFF,    
    PHMM_PAIR_PROB_SCALE,  
    PHMM_PAIR_PROB_COEFF,
    PHMM_LOOP_PROB_SCALE,
    PHMM_LOOP_PROB_COEFF,
    PHMM_ALN_PROB_CUTOFF,    
    PHMM_ALN_PROB_SCALE,     
    PHMM_ALN_PROB_COEFF,
    PHMM_GAP_PROB_SCALE,
    PHMM_GAP_PROB_COEFF,

    PHMM_ESC_LOOP,           
    PHMM_GAP_EXTEND,         
    PHMM_GAP_EXTEND_OCC,     
    PHMM_PI_LOOP,            
    PHMM_RATE_PARAM_LOOP,   
    PHMM_PHY_LOOP_SCALE,
    PHMM_PI_STEM,
    PHMM_RATE_PARAM_STEM,
    PHMM_PHY_STEM_SCALE,
    PHMM_SMTX_LOOP,
    PHMM_SMTX_STEM,
    PHMM_SMTX_LOOP_SIM,
    PHMM_SMTX_STEM_SIM,
    PHMM_SMTX_LOOP_GC,
    PHMM_SMTX_STEM_GC,
    PHMM_SMTX_LOOP_DINUC,
    PHMM_SMTX_STEM_DINUC,
    PHMM_SMTX_LOOP_SCALE,
    PHMM_SMTX_STEM_SCALE,
    PHMM_SMTX_LOOP_SIM_SCALE,
    PHMM_SMTX_STEM_SIM_SCALE,
    PHMM_GAP_OPEN,           
    PHMM_GAP_OPEN_OCC,       
    PHMM_GAP_OPEN_OCC_CROSS,
#endif

    SEQ_SIM_RANGE,

    NPARAM,

    INVALID_PARAM = (-1)
  };
private:
  struct Info {
    Param param;
    Type type;
    int nrow;
    int ncol;
    int size;
    char name[50];
  };
  static const Info& getParamInfo(const Param& param) {
    enum {
      np = Alpha::REDUCED_NPBASE,
      nc = Alpha::NBASE
    };
    typedef CArray<Info, NPARAM> Table;
    static const Table table = {{
#define Item(param0, type0, nrow0, ncol0) \
      {(param0), (type0), (nrow0), (ncol0), \
      (((type0)==SYMMTX || (type0)==STACK) ? (((nrow0)*((nrow0)+1))/2) : ((type0)==ASMMTX ? ((nrow0 * (nrow0 - 1))/2) : ((nrow0)*(ncol0)))), #param0}
   
      //PSCFG
      Item(PSCFG_ESC0,               VECTOR, 1,  PSCFG_Energy::getModel().getNESC0()),
      Item(PSCFG_TSC0,               VECTOR, 1,  PSCFG_Energy::getModel().getNTSC0()),

      Item(PSCFG_PAIR_PROB_CUTOFF,   FLOAT,  1,  1),
      Item(PSCFG_PAIR_PROB_SCALE,    FLOAT,  1,  1),
      Item(PSCFG_PAIR_PROB_COEFF,    FLOAT,  1,  1),
      Item(PSCFG_PAIR_PROB_NSEQ_COEFF,    FLOAT,  1,  1),
      Item(PSCFG_LOOP_PROB_SCALE,    FLOAT,  1,  1),
      Item(PSCFG_LOOP_PROB_COEFF,    FLOAT,  1,  1),
      Item(PSCFG_ALN_PROB_CUTOFF,    FLOAT,  1,  1),
      Item(PSCFG_ALN_PROB_SCALE,     FLOAT,  1,  1), 
      Item(PSCFG_ALN_PROB_COEFF,     FLOAT,  1,  1),
      Item(PSCFG_ALN_PROB_STEM_SCALE,     FLOAT,  1,  1), 
      Item(PSCFG_ALN_PROB_STEM_COEFF,     FLOAT,  1,  1),
      Item(PSCFG_GAP_PROB_SCALE,     FLOAT,  1,  1), 
      Item(PSCFG_GAP_PROB_COEFF,     FLOAT,  1,  1),

      Item(PSCFG_ESC_STEM,           FLOAT,  1,  1), 
      Item(PSCFG_ESC_FCOL,           VECTOR, 1,  nc), 
      Item(PSCFG_ESC_FPAIR,          VECTOR, 1,  np),  
      Item(PSCFG_ESC_LOD_SCORE,      FLOAT,  1,  1),  
      Item(PSCFG_ESC_LOOP,           VECTOR, 1,  PSCFG_Energy::getNLoop()),
      Item(PSCFG_GAP_EXTEND,         VECTOR, 1,  PSCFG_Energy::getNLoop()),
      Item(PSCFG_GAP_EXTEND_OCC,     VECTOR, 1,  PSCFG_Energy::getNLoop()),
      Item(PSCFG_PI_LOOP,            VECTOR, 1,  nc),
      Item(PSCFG_PI_STEM,            VECTOR, 1,  np),    
      Item(PSCFG_RATE_PARAM_LOOP,    VECTOR, 1,  ((nc*(nc-1)) / 2)),
      Item(PSCFG_RATE_PARAM_STEM,    VECTOR, 1,  ((np*(np-1)) / 2)),
      Item(PSCFG_PHY_LOOP_SCALE,     FLOAT,  1,  1),
      Item(PSCFG_PHY_STEM_SCALE,     FLOAT,  1,  1),
      Item(PSCFG_SMTX_LOOP,          SYMMTX, nc, nc),      
      Item(PSCFG_PAIR_SMTX,          SYMMTX, np, np),  
      Item(PSCFG_SMTX_LOOP_SIM,      SYMMTX, nc, nc),      
      Item(PSCFG_PAIR_SMTX_SIM,      SYMMTX, np, np),  
      Item(PSCFG_SMTX_LOOP_NSEQ,     SYMMTX, nc, nc),      
      Item(PSCFG_PAIR_SMTX_NSEQ,     SYMMTX, np, np),  
      Item(PSCFG_SMTX_LOOP_ASYM,     ASMMTX, nc, nc),      
      Item(PSCFG_PAIR_SMTX_ASYM,     ASMMTX, np, np),  
      Item(PSCFG_SMTX_LOOP_GC,       SYMMTX, nc, nc),      
      Item(PSCFG_PAIR_SMTX_GC,       SYMMTX, np, np),  
      Item(PSCFG_STACK_MTX,          STACK,  np, np),      
      Item(PSCFG_GAP_OPEN,           VECTOR, 1,  PSCFG_Energy::getNLoop()),
      Item(PSCFG_GAP_OPEN_OCC,       VECTOR, 1,  PSCFG_Energy::getNLoop()),
      Item(PSCFG_GAP_OPEN_OCC_CROSS, VECTOR, 1,  PSCFG_Energy::getNLoop()),
      Item(PSCFG_TSC_MP_LOOP,        VECTOR, 1,  PSCFG_Energy::getNLoop()),
      Item(PSCFG_TERM_STEM,          VECTOR, 1,  np),
      Item(PSCFG_MIN_BASE_PAIR,      FLOAT,  1,  1),
      Item(PSCFG_MIN_CAN_FRAC,       FLOAT,  1,  1),
      Item(PSCFG_SEQ_SIM_BASE,       FLOAT,  1,  1),

      Item(PSCFG_TURNER_STACK_SCALE,    FLOAT,   1,  1),
      Item(PSCFG_TURNER_STACK_SCALE_SIM, FLOAT,   1,  1),
      Item(PSCFG_TURNER_TSTACKH_SCALE,  FLOAT,   1,  1),
      Item(PSCFG_TURNER_TSTACKI_SCALE,  FLOAT,   1,  1),
      Item(PSCFG_TURNER_TSTACKM_SCALE,  FLOAT,   1,  1),
      Item(PSCFG_TURNER_TERM_AU_SCALE,  FLOAT,   1,  1),
      Item(PSCFG_TURNER_LOOP_LEN_SCALE, FLOAT,   1,  1),
      Item(PSCFG_TURNER_LOOP_LEN_C0,    VECTOR,  1,  PSCFG_Energy::getNLoop()),
      Item(PSCFG_TURNER_LOOP_LEN_C1,    VECTOR,  1,  PSCFG_Energy::getNLoop()),

      Item(PSCFG_SMTX_LOOP_SCALE      , FLOAT, 1,  1),
      Item(PSCFG_PAIR_SMTX_SCALE      , FLOAT, 1,  1),
      Item(PSCFG_SMTX_LOOP_SIM_SCALE  , FLOAT, 1,  1),
      Item(PSCFG_PAIR_SMTX_SIM_SCALE  , FLOAT, 1,  1),
      Item(PSCFG_STACK_MTX_SCALE      , FLOAT, 1,  1),
      Item(PSCFG_RATE_PARAM_LOOP_SCALE, FLOAT, 1,  1),
      Item(PSCFG_RATE_PARAM_STEM_SCALE , FLOAT, 1,  1),

#if false
      //SCFG
      Item(SCFG_ESC0,                VECTOR,  1, SCFG_Energy::getNESC0()),
      Item(SCFG_TSC0,                VECTOR,  1, SCFG_Energy::getNTSC0()),

      Item(SCFG_PAIR_PROB_CUTOFF,    FLOAT,   1, 1),
      Item(SCFG_PAIR_PROB_SCALE,     FLOAT,   1, 1),
      Item(SCFG_PAIR_PROB_COEFF,     FLOAT,   1, 1),
      Item(SCFG_LOOP_PROB_SCALE,     FLOAT,   1, 1),
      Item(SCFG_LOOP_PROB_COEFF,     FLOAT,   1, 1),
      
      Item(SCFG_ESC_STEM,            FLOAT,   1, 1),
      Item(SCFG_ESC_FCOL,            VECTOR,  1, nc),  
      Item(SCFG_ESC_FPAIR,           VECTOR,  1, np),  
      Item(SCFG_ESC_LOD_SCORE,       FLOAT,   1, 1),  
      Item(SCFG_ESC_LOOP,            VECTOR,  1, SCFG_Energy::getNLoop()),
      Item(SCFG_PI_LOOP,             VECTOR,  1, nc),
      Item(SCFG_PI_STEM,             VECTOR,  1, np),
      Item(SCFG_RATE_PARAM_LOOP,     VECTOR,  1, (nc * (nc - 1)) / 2),
      Item(SCFG_RATE_PARAM_STEM,     VECTOR,  1, (np * (np - 1)) / 2),
      Item(SCFG_PHY_LOOP_SCALE,      FLOAT,   1,  1),
      Item(SCFG_PHY_STEM_SCALE,      FLOAT,   1,  1),
      Item(SCFG_SMTX_LOOP,           SYMMTX, nc, nc),
      Item(SCFG_PAIR_SMTX,           SYMMTX, np, np),
      Item(SCFG_SMTX_LOOP_SIM,       SYMMTX, nc, nc),
      Item(SCFG_PAIR_SMTX_SIM,       SYMMTX, np, np),
      Item(SCFG_SMTX_LOOP_GC,        SYMMTX, nc, nc),
      Item(SCFG_PAIR_SMTX_GC,        SYMMTX, np, np),
      Item(SCFG_STACK_MTX,           STACK,  np, np),
      Item(SCFG_TSC_MP_LOOP,         VECTOR,  1, SCFG_Energy::getNLoop()),
      Item(SCFG_TERM_STEM,           VECTOR,  1, np),
      Item(SCFG_MIN_BASE_PAIR,       FLOAT,   1, 1),
      Item(SCFG_MIN_CAN_FRAC,        FLOAT,   1, 1),
      Item(SCFG_MIN_PSCORE,          FLOAT,   1, 1),
      Item(SCFG_PSCORE_SCALE,        FLOAT,   1, 1),

      Item(SCFG_TURNER_STACK_SCALE,    FLOAT,   1, 1),
      Item(SCFG_TURNER_TSTACKH_SCALE,  FLOAT,   1, 1),
      Item(SCFG_TURNER_TSTACKI_SCALE,  FLOAT,   1, 1),
      Item(SCFG_TURNER_TSTACKM_SCALE,  FLOAT,   1, 1),
      Item(SCFG_TURNER_TERM_AU_SCALE,  FLOAT,   1, 1),
      Item(SCFG_TURNER_LOOP_LEN_SCALE, FLOAT,   1, 1),
      Item(SCFG_TURNER_LOOP_LEN_C0,    VECTOR,  1, SCFG_Energy::getNLoop()),
      Item(SCFG_TURNER_LOOP_LEN_C1,    VECTOR,  1, SCFG_Energy::getNLoop()),

      Item(SCFG_SMTX_LOOP_SCALE      , FLOAT, 1,  1),
      Item(SCFG_PAIR_SMTX_SCALE      , FLOAT, 1,  1),
      Item(SCFG_SMTX_LOOP_SIM_SCALE  , FLOAT, 1,  1),
      Item(SCFG_PAIR_SMTX_SIM_SCALE  , FLOAT, 1,  1),
      Item(SCFG_STACK_MTX_SCALE      , FLOAT, 1,  1),
      Item(SCFG_RATE_PARAM_LOOP_SCALE, FLOAT, 1,  1),
      Item(SCFG_RATE_PARAM_STEM_SCALE, FLOAT, 1,  1),

      //PHMM
      Item(PHMM_ESC0,                VECTOR,  1, PHMM_Domain::getNESC0()),
      Item(PHMM_TSC0,                VECTOR,  1, PHMM_Domain::getNTSC0()),

      Item(PHMM_PAIR_PROB_CUTOFF,    FLOAT,   1, 1), 
      Item(PHMM_PAIR_PROB_SCALE,     FLOAT,   1, 1),  
      Item(PHMM_PAIR_PROB_COEFF,     FLOAT,   1, 1),  
      Item(PHMM_LOOP_PROB_SCALE,     FLOAT,   1, 1),  
      Item(PHMM_LOOP_PROB_COEFF,     FLOAT,   1, 1),  
      Item(PHMM_ALN_PROB_CUTOFF,     FLOAT,   1, 1), 
      Item(PHMM_ALN_PROB_SCALE,      FLOAT,   1, 1),   
      Item(PHMM_ALN_PROB_COEFF,      FLOAT,   1, 1),
      Item(PHMM_GAP_PROB_SCALE,      FLOAT,   1, 1),   
      Item(PHMM_GAP_PROB_COEFF,      FLOAT,   1, 1),

      Item(PHMM_ESC_LOOP,            VECTOR,  1, PHMM_Domain::getNLoop()), 
      Item(PHMM_GAP_EXTEND,          VECTOR,  1, PHMM_Domain::getNLoop()), 
      Item(PHMM_GAP_EXTEND_OCC,      VECTOR,  1, PHMM_Domain::getNLoop()),
      Item(PHMM_PI_LOOP,             VECTOR,  1, nc),
      Item(PHMM_RATE_PARAM_LOOP,     VECTOR,  1, ((nc*(nc-1)) / 2)),
      Item(PHMM_PHY_LOOP_SCALE,      FLOAT,   1,  1),
      Item(PHMM_PI_STEM,             VECTOR,  1, nc),
      Item(PHMM_RATE_PARAM_STEM,     VECTOR,  1, ((nc*(nc-1)) / 2)),
      Item(PHMM_PHY_STEM_SCALE,      FLOAT,   1,  1),
      Item(PHMM_SMTX_LOOP,           SYMMTX, nc, nc),
      Item(PHMM_SMTX_STEM,           SYMMTX, nc, nc),
      Item(PHMM_SMTX_LOOP_SIM,       SYMMTX, nc, nc),
      Item(PHMM_SMTX_STEM_SIM,       SYMMTX, nc, nc),
      Item(PHMM_SMTX_LOOP_GC,        SYMMTX, nc, nc),
      Item(PHMM_SMTX_STEM_GC,        SYMMTX, nc, nc),
      Item(PHMM_SMTX_LOOP_DINUC,     SYMMTX, (nc*nc), (nc*nc)),
      Item(PHMM_SMTX_STEM_DINUC,     SYMMTX, (nc*nc), (nc*nc)),
      Item(PHMM_SMTX_LOOP_SCALE,     FLOAT,   1,  1),
      Item(PHMM_SMTX_STEM_SCALE,     FLOAT,   1,  1),
      Item(PHMM_SMTX_LOOP_SIM_SCALE, FLOAT,   1,  1),
      Item(PHMM_SMTX_STEM_SIM_SCALE, FLOAT,   1,  1),
      Item(PHMM_GAP_OPEN,            VECTOR,  1, PHMM_Domain::getNLoop()), 
      Item(PHMM_GAP_OPEN_OCC,        VECTOR,  1, PHMM_Domain::getNLoop()), 
      Item(PHMM_GAP_OPEN_OCC_CROSS,  VECTOR,  1, PHMM_Domain::getNLoop()),
#endif

      Item(SEQ_SIM_RANGE,            VECTOR,  1,  2),
#undef Item
    }};
    return table[param2index(param)];
  }
public:
  static int getNParamInfo() {return NPARAM;}
  inline static int param2index(const Param& param) {
    Assert(hasParam(param));
    return param;
  }
  inline static Param index2param(const int& index) {
    return Param(index);
  }
  static bool hasParam(int param) {
    return (0 <= param && param < NPARAM);
  }
  static Vector<ScoreParam> readScoreParamList(const string& fname);
  static Vector<ScoreParam> readScoreParamList(istream& is, const string& name);

  ScoreParam();
  ScoreParam(const string& fname);
  ScoreParam(const ScoreParam& scoreParam);
  ~ScoreParam();
  ScoreParam& operator=(const ScoreParam& scoreParam);
  string toString() const;
  void print() const;

  void clean();
  string getAC() const {return ac;}
  bool readEntry(istream& fi, const string& name);
  void setFromFile(const string& infile);
  string getName(int param) const;
  int getParamSize(int param) const;
  const float& getVal(int param) const;
  //Vector<float> getVector(int param) const;
  const Array<float>& getArray(int param) const;
  const NMatrix<float>& getMatrix(int param) const;
  //void setFromVector(int param, const Vector<float>& v);
  template <typename InpIt>
  void setParam(int param, InpIt first, InpIt last) {
    Datum& datum = getDatum(param);
    Assert((int)std::distance(first, last) <= (int)datum.size());
    const Info& info = getParamInfo(Param(param));
    switch (info.type) {
    case FLOAT:
      std::copy(first, last, datum.begin());
      break;
    case VECTOR:
      std::copy(first, last, datum.begin());
      break;
    case MATRIX:
      Assert((int)std::distance(first, last) == info.size);
      std::copy(first, last, datum.begin());
      break;
    case SYMMTX:
      Assert((int)std::distance(first, last) == (info.nrow * (info.nrow + 1)) / 2);
      for (int i = 0; i < info.nrow; i++) {
	for (int j = i; j < info.ncol; j++) {
	  datum.ref(i, j) = datum.ref(j, i) = (*first++);
	}
      }
      break;
    case ASMMTX:
      Assert((int)std::distance(first, last) == (info.nrow * (info.nrow - 1)) / 2);
      for (int i = 0; i < info.nrow; i++) {
	datum.ref(i, i) = 0.0;
	for (int j = (i + 1); j < info.ncol; j++) {
	  datum.ref(i, j) = (*first++);
	  datum.ref(j, i) = -datum.ref(i, j);
	}
      }
      break;
    case STACK:
      {
	Assert((int)std::distance(first, last) == getNStack());
	Array<float> v(first, last);
	for (int i = 0; i < info.nrow; i++) {
	  for (int j = 0; j < info.nrow; j++) {
	    datum.ref(i, j) = v[getStackIndex(i, j)];
	  }
	}
      }
      break;
    default:
      Die("bad info.type=%d", info.type);
      break;
    }
  }
  Array<float> getParam(int param) const {
    const Datum& datum =  getDatum(param);
    const Info& info = getParamInfo(Param(param));
    switch (info.type) {
    case FLOAT:
      return datum.toArray();;
    case VECTOR:
      return datum.toArray();
    case MATRIX:
      return datum.toArray();
    case SYMMTX:
      {
	Array<float> pv((info.nrow * (info.nrow + 1)) / 2);
	int k = 0;
	for (int i = 0; i < info.nrow; i++) {
	  for (int j = i; j < info.nrow; j++) {
	    pv[k++] = datum.ref(i, j);
	  }
	}
	return pv;
      }
    case ASMMTX:
      {
	Array<float> pv((info.nrow * (info.nrow - 1)) / 2);
	int k = 0;
	for (int i = 0; i < info.nrow; i++) {
	  for (int j = (i + 1); j < info.nrow; j++) {
	    pv[k++] = datum.ref(i, j);
	  }
	}
	return pv;
      }
    case STACK:
      {
	Array<float> pv(getNStack());
	for (int i = 0; i < info.nrow; i++) {
	  for (int j = 0; j < info.nrow; j++) {
	    pv[getStackIndex(i, j)] = datum.ref(i, j);
	  }
	}
	return pv;
      }
    default:
      Die("bad info.type=%d", info.type);
      return Array<float>();
    }
  }
  //void setEnergyParam(const EnergyParam& energyParam) {
  //  pEnergyParam = &energyParam;
  //}
  int getStackIndex(int i, int j) const;
  int getNStack() const;
  // const EnergyParam* getEnergyParam() const {return pEnergyParam;}

private:
  typedef map<string, Param> NameTable;
  typedef NMatrix<float> Datum;
  typedef CArray<Datum, NPARAM> Data;
  string ac;
  string id;
  NameTable nameTable;
  Data data;
  // const EnergyParam* pEnergyParam;

  void createNameTable();
  void createData();
  Datum& getDatum(int param);
  const Datum& getDatum(int param) const;
  void readArray(Datum& datum, string name, istream& fi);
  void readArray_varlen(Datum& datum, string name, istream& fi);
  void readMatrix(Datum& datum, string name, istream& fi);
  void writeArray(const Datum& datum, int indentLevel, ostream& fo) const;
  void writeMatrix(const Datum& datum, int indentLevel, ostream& fo) const;
};

#endif
