/*
 * SeqFamAligner.hpp
 *
 */
#ifndef SEQ_FAM_ALIGNER_HPP
#define SEQ_FAM_ALIGNER_HPP

#include "Util.hpp"
#include "SeqData.hpp"
#include "GuideTree.hpp"
#include "Weight.hpp"
#include "Scheduler.hpp"

class SeqIds;
class SeqFamAligner {
public:
  virtual ~SeqFamAligner() {delete &seqIds;}
  virtual SeqFamAligner& dup() const = 0;
  virtual void run() = 0;
  void setSeqIds(const SeqIds& seqIds0) {seqIds = seqIds0;}
  void setSOut(string& sout) {pSOut = &sout;}
  void setSLog(string& slog) {pSLog = &slog;}
    
protected:
  SeqData& seqData;
  Scheduler& scheduler;
  SeqIds& seqIds;
  string* pSOut;
  string* pSLog;
  
  SeqFamAligner(SeqData& seqData, 
		Scheduler& scheduler)
    : seqData(seqData), 
      scheduler(scheduler),
      seqIds(*new SeqIds),
      pSOut(NULL),
      pSLog(NULL)
  {}
  SeqFamAligner(const SeqFamAligner& seqFamAligner)
    : seqData(seqFamAligner.seqData),
      scheduler(seqFamAligner.scheduler),
      seqIds(*new SeqIds(seqFamAligner.seqIds)),
      pSOut(seqFamAligner.pSOut),
      pSLog(seqFamAligner.pSLog)
  {}
private:
  SeqFamAligner();
  SeqFamAligner& operator=(const SeqFamAligner& seqFamAligner);
};
  

#endif
