/*
 * StripRegionIterators.cpp
 *
 */
#ifndef STRIP_REGION_ITERATORS_CPP
#define STRIP_REGION_ITERATORS_CPP
#include "StripRegion.hpp"

template <bool fwd>
StripRegion::RowIterator<fwd>::
RowIterator(const StripRegion& sr)
  : x(-1),
    ly(0),
    y_beg(-1),
    y_last(-2),
    sr(sr),
    bbox(sr.rect0),
    idx_lim(-1),
    idx(-1),
    x_lim(-1),
    next_x(-1)
{
  reset();
}
template <bool fwd>
StripRegion::RowIterator<fwd>::
RowIterator(const StripRegion& sr, const Rect& bbox)
  : x(-1),
    ly(0),
    y_beg(-1),
    y_last(-2),
    sr(sr),
    bbox(bbox),
    idx_lim(-1),
    idx(-1),
    x_lim(-1),
    next_x(-1)
{
  reset();
}
template <bool fwd>
StripRegion::RowIterator<fwd>::
~RowIterator()
{
}
template <bool fwd>
void
StripRegion::RowIterator<fwd>::
reset(const Rect& rect0)
{
  bbox = rect0;
  reset();
}
template <bool fwd>
void
StripRegion::RowIterator<fwd>::
reset()
{
  x = -1;
  ly = 0;
  y_beg = -1;
  y_last = -2;
  if (!bbox.isNull()) {
    const int& x_beg = max(sr.rect0.bx, bbox.bx);
    const int& x_last = min(sr.rect0.ex, bbox.ex);
    if (0 <= x_beg && x_beg <= x_last) {
      x_lim = (fwd ? (x_last + 1) : (x_beg - 1));
      next_x = (fwd ? x_beg : x_last);
      idx_lim = sr.getXTIndex(fwd ? x_last : x_beg);
      idx = sr.getXTIndex(next_x);
      return;
    }
  }
  x_lim = -1;
  next_x = -1;
  idx_lim = -1;
  idx = -1;
}
template <bool fwd>
bool
StripRegion::RowIterator<fwd>::
getNext()
{
  if (next_x == x_lim) return false;

  x = next_x;
  y_beg = max(sr[idx].getBY(x), bbox.by);
  y_last = min(sr[idx].getEY(x), bbox.ey);
  ly = (y_last - y_beg + 1);
  if (ly <= 0) {
    ly = 0;
    y_beg = -1;
    y_last = -2;
  }

  (fwd ? next_x++ : next_x--);
  while (fwd ? (sr[idx].ex < next_x) : (next_x < sr[idx].bx)) {
    if (idx == idx_lim) break;

    (fwd ? idx++ : idx--);
  }
  return true;
}

template <bool fwd>
StripRegion::Iterator<fwd>::
Iterator(const StripRegion& sr)
  : pt(Point::null()), rit(sr), y_lim(-1), next_y(-1)
{
  reset();
}
template <bool fwd>
StripRegion::Iterator<fwd>::
Iterator(const StripRegion& sr, const Rect& bbox)
  : pt(Point::null()), rit(sr, bbox), y_lim(-1), next_y(-1)
{
  reset();
}
template <bool fwd>
StripRegion::Iterator<fwd>::
~Iterator()
{
}
template <bool fwd>
void
StripRegion::Iterator<fwd>::
reset(const Rect& rect0)
{
  rit.reset(rect0);
  reset0();
}
template <bool fwd>
void
StripRegion::Iterator<fwd>::
reset()
{
  rit.reset();
  reset0();
}
template <bool fwd>
void
StripRegion::Iterator<fwd>::
reset0()
{
  pt.setNull();
  while (rit.getNext()) {
    if (rit.ly == 0) continue;

    y_lim = (fwd ? (rit.y_last + 1) : (rit.y_beg - 1));
    next_y = (fwd ? rit.y_beg : rit.y_last);
    pt.x = rit.x;
    return;
  }
  y_lim = -1;
  next_y = -1;
}
template <bool fwd>
bool
StripRegion::Iterator<fwd>::
getNext()
{
  while (next_y == y_lim) {
    if (!rit.getNext()) return false;
    if (rit.ly == 0) continue;

    y_lim = (fwd ? (rit.y_last + 1) : (rit.y_beg - 1));
    next_y = (fwd ? rit.y_beg : rit.y_last);
    pt.x = rit.x;
  }
  pt.y = next_y;
  (fwd ? next_y++ : next_y--);
  return true;
}
#endif
