/*
 * DPDeckqPSCFGDPRegion.cpp
 *
 */
#include "DPDeckPSCFGDPRegion.hpp"

const char*
DPDeckPSCFG::DPRegion::
type2str(int type0) {
  static const char names[][20] = {
    "TRIANGLE",
    "CARET",
    "NOSE",
    "TRIANGLE_CUT",
    "CARET_CUT_L",
    "CARET_CUT_R",
    "INVALID_R_TYPE"
  };
  return names[(0 <= type0 && type0 < INVALID_R_TYPE)
	       ? type0 : INVALID_R_TYPE];
}
const char*
DPDeckPSCFG::DPRegion::
shape2str(int shape0) {
  static const char names[][20] = {
    "TRI_SHAPE",
    "RECT_VERT_SHAPE",
    "RECT_DIAG_SHAPE",
    "INVALID_SHAPE"
  };
  return names[(0 <= shape0 && shape0 < INVALID_SHAPE)
	       ? shape0 : INVALID_SHAPE];
}
const char*
DPDeckPSCFG::DPRegion::
regionNum2str(int regionNum0) {
  static const char names[][20] = {
    "REGION_EE",
    "REGION_BB",
    "REGION_BE",
    "REGION_ME",
    "REGION_BM",
    "REGION_MM",
    "INVALID_NUM"
  };
  return names[(0 <= regionNum0 && regionNum0 < INVALID_NUM)
	       ? regionNum0 : INVALID_NUM];
}
int
DPDeckPSCFG::DPRegion::
getNRegion(int type1)
{
  switch (type1) {
  case TRIANGLE: return 1;
  case CARET: return 3;
  case NOSE: return 6;
  case TRIANGLE_CUT: return 3;
  case CARET_CUT_L: return 6;
  case CARET_CUT_R: return 6;
  default:
    Die("bad type=%d", type1);
    return (-1);
  }
}
int
DPDeckPSCFG::DPRegion::
regionNum2Index(int regionNum)
{// 0 <= regionNum < getNRegion(type)
  return regionNum;
}

DPDeckPSCFG::DPRegion::
DPRegion()
  : rect0(Rect::null()),
    type(INVALID_R_TYPE), 
    pSkipSet(NULL),
    regions(*new VRegion())
{
}
DPDeckPSCFG::DPRegion::
DPRegion(const DPRegion& dpRegion)
  : rect0(dpRegion.rect0),
    type(dpRegion.type),
    pSkipSet(dpRegion.pSkipSet),
    regions(*new VRegion())
{
  for (int i = 0; i < (int)dpRegion.regions.size(); i++) {
    const Region& r = dpRegion.regions[i];
    regions.push_back(Region(*this, r));
  }
}
DPDeckPSCFG::DPRegion::
~DPRegion()
{
  delete &regions;
}
DPDeckPSCFG::DPRegion&
DPDeckPSCFG::DPRegion::
operator=(const DPRegion& dpRegion)
{
  rect0 = dpRegion.rect0;
  type = dpRegion.type;
  pSkipSet = dpRegion.pSkipSet;
  regions.clear();
  for (int i = 0; i < (int)dpRegion.regions.size(); i++) {
    const Region& r = dpRegion.regions[i];
    regions.push_back(Region(*this, r));
  }
  return *this;
}
string
DPDeckPSCFG::DPRegion::
toString() const
{
  ostringstream oss;
  oss << "rect0: " << rect0.toString() << "\n"
      << "type: " << type2str(type) << "\n"
      << "skipSize: " << pSkipSet->getSkipSize() << "\n"
      << "regions: \n";
  for (int i = 0; i < (int)regions.size(); i++) {
    oss << "--- \n"
	<< regions[i].toString();
  }
  return oss.str();
}
void
DPDeckPSCFG::DPRegion::
print() const
{
  cout << toString() << flush;
}

bool
DPDeckPSCFG::DPRegion::
isType(const int& type1) const
{
  return (type == type1);
}
int
DPDeckPSCFG::DPRegion::
getArea() const
{
  int n = 0;
  for (int i = 0; i < (int)regions.size(); i++) {
    n += regions[i].getArea();
  }
  return n;
}
bool
DPDeckPSCFG::DPRegion::
isRegionType(int regionType) const
{
  return (type == regionType);
}
bool
DPDeckPSCFG::DPRegion::
includes(const Rect& rect) const
{
  if (!rect0.includes(rect)) return false;

  for (int i = 0; i < (int)regions.size(); i++) {
    if (regions[i].includes(rect)) return true;
  }
  return false;
}
int
DPDeckPSCFG::DPRegion::
getNRegion() const
{
  return getNRegion(type);
}
const DPDeckPSCFG::DPRegion::Region&
DPDeckPSCFG::DPRegion::
getSubRegion(int regionNum) const
{
  return regions[regionNum2Index(regionNum)];
}
void
DPDeckPSCFG::DPRegion::
setTriangle(const StripRegion& sr)
{
  Check(sr.rect0.isNormal());

  type = TRIANGLE;
  rect0 = sr.rect0;
  regions.clear();
  regions.push_back(Region(*this, REGION_EE, TRI_SHAPE, sr, sr));
}
void
DPDeckPSCFG::DPRegion::
setCaret(const StripRegion& srb, const StripRegion& sre)
{
  Check(srb.rect0.isNormal()
	&& sre.rect0.isNormal()
	&& (srb.ex0 < sre.bx0) 
	&& (srb.ey0 <= sre.getBXEY())
	&& (srb.getEXBY() <= sre.by0));

  type = CARET;
  rect0.setVal(srb.b0, sre.e0);
  regions.clear();
  regions.push_back(Region(*this, REGION_EE, TRI_SHAPE, sre, sre));
  regions.push_back(Region(*this, REGION_BB, TRI_SHAPE, srb, srb));
  regions.push_back(Region(*this, REGION_BE, RECT_DIAG_SHAPE, srb,sre));
}
void
DPDeckPSCFG::DPRegion::
setNose(const StripRegion& srb, const StripRegion& sre,
	const StripRegion& srm)
{
  Check(srb.rect0.isNormal()
	&& sre.rect0.isNormal()
	&& srm.rect0.isNormal()
	&& (srb.ex0 < srm.bx0) 
	&& (srb.ey0 <= srm.getBXEY())
	&& (srb.getEXBY() <= srm.by0)
	&& (srm.ex0 < sre.bx0)
	&& (srm.ey0 <= sre.getBXEY())
	&& (srm.getEXBY() <= sre.by0));

  type = NOSE;
  rect0.setVal(srb.b0, sre.e0);
  regions.clear();
  regions.push_back(Region(*this, REGION_EE, TRI_SHAPE, sre, sre));
  regions.push_back(Region(*this, REGION_BB, TRI_SHAPE, srb, srb));
  regions.push_back(Region(*this, REGION_BE,RECT_DIAG_SHAPE,srb,sre,srm));
  regions.push_back(Region(*this, REGION_ME, RECT_DIAG_SHAPE, srm,sre));
  regions.push_back(Region(*this, REGION_BM, RECT_DIAG_SHAPE, srb,srm));
  regions.push_back(Region(*this, REGION_MM, TRI_SHAPE, srm, srm));
}
bool
DPDeckPSCFG::DPRegion::
isSplittable() const
{
  if (type == TRIANGLE) {
    const int& lx = rect0.lengthX();
    return (lx > 3);// (mx + 1) == rect0.ex is disallowed
  } else if (type == CARET) {
    const Rect& rectl = getSubRegion(REGION_BB).srb.rect0;
    const Rect& rectr = getSubRegion(REGION_EE).srb.rect0;
    const int& l = rectl.lengthX();
    const int& r = rectr.lengthX();
    if (l >= r) {
      return (l >= 3);// (mx + 1) == rectl.ex is allowed
    } else {
      return (r > 3);// (mx + 1) == rectr.ex is disallowed
    }
  } else if (type == NOSE) {
    return true;// always has different shapes after split
  } else {
    Die("method undefined for type=%s", type2str(type));
    return false;
  }
}
DPDeckPSCFG::DPRegion::RType
DPDeckPSCFG::DPRegion::
getCutType() const
{
  if (type == TRIANGLE) {
    return TRIANGLE_CUT;
  } else if (type == CARET) {
    const Rect& rectl = getSubRegion(REGION_BB).srb.rect0;
    const Rect& rectr = getSubRegion(REGION_EE).srb.rect0;
    const int& l = rectl.lengthX();
    const int& r = rectr.lengthX();
    if (l >= r) {
      return CARET_CUT_L;
    } else {
      return CARET_CUT_R;
    }
  } else {
    Die("method undefined for type=%s", type2str(type));
    return INVALID_R_TYPE;
  }
}
int
DPDeckPSCFG::DPRegion::
getCutPoint() const
{
  const int& cutType = getCutType();
  if (cutType == TRIANGLE_CUT) {
    const int& m = ((rect0.bx + 1) + (rect0.ex - 2)) / 2;
    Assert(rect0.bx < m && m < (rect0.ex - 1));
    return m;
  } else if (cutType == CARET_CUT_L) {
    const Rect& rectl = getSubRegion(REGION_BB).srb.rect0;
    const int& m = ((rectl.bx + 1) + (rectl.ex - 1)) / 2;
    Assert(rectl.bx < m && m < rectl.ex); // (m+1 == rect0.ex) is allowed
    return m;
  } else if (cutType == CARET_CUT_R) {
    const Rect& rectr = getSubRegion(REGION_EE).srb.rect0;
    const int& m = ((rectr.bx + 1) + (rectr.ex - 2)) / 2;
    Assert(rectr.bx < m && m < (rectr.ex - 1));
    return m;
  } else {
    Die("method undefined for cutType=%s", type2str(cutType));
    return (-1);
  }
}
void
DPDeckPSCFG::DPRegion::
insertCut()
{
  Rect rect, rectb, recte;
  StripRegion sr0, srb0, sre0, sr, srb, sre;

  Assert(isSplittable());
  int mx = getCutPoint();//using current type
  type = getCutType();
  if (type == TRIANGLE_CUT) {
    sr0 = regions[0].srb; // have a local copy
    regions.clear();

    rect.setVal(mx, sr0.by0, sr0.ex0, sr0.ey0);
    sr = sr0.getBoundedRegion(rect);
    regions.push_back(Region(*this, REGION_EE, TRI_SHAPE, sr, sr));

    rect.setVal(sr0.bx0, sr0.by0, mx + 1, sr0.ey0);
    sr = sr0.getBoundedRegion(rect);
    regions.push_back(Region(*this, REGION_BB, TRI_SHAPE, sr, sr));
 
    rectb.setVal(sr0.bx0, sr0.by0, mx, sr0.ey0);
    recte.setVal(mx + 1, sr0.by0, sr0.ex0, sr0.ey0);
    srb = sr0.getBoundedRegion(rectb);
    sre = sr0.getBoundedRegion(recte);
    regions.push_back(Region(*this, REGION_BE, RECT_VERT_SHAPE, srb,sre));

#ifdef DEBUG
    const Region& ree = regions[0];
    const Region& rbb = regions[1];
    const Region& rbe = regions[2];
    Assert(rbb.sre.ex0 == mx + 1);
    Assert(ree.srb.bx0 == mx);
    Assert(rbb.sre.ex0 == rbe.sre.bx0);
    Assert(rbe.srb.ex0 == ree.srb.bx0);
#endif
  } else if (type == CARET_CUT_L) {
    // have local copy
    srb0 = getSubRegion(REGION_BB).srb;
    sre0 = getSubRegion(REGION_EE).sre; 
    regions.clear();

    regions.push_back(Region(*this, REGION_EE, TRI_SHAPE, sre0, sre0));

    rect.setVal(srb0.bx0, srb0.by0, mx + 1, srb0.ey0);
    sr = srb0.getBoundedRegion(rect);
    regions.push_back(Region(*this, REGION_BB, TRI_SHAPE, sr, sr));

    rectb.setVal(srb0.bx0, srb0.by0, mx, srb0.ey0);
    srb = srb0.getBoundedRegion(rectb);
    regions.push_back(Region(*this, REGION_BE, RECT_DIAG_SHAPE,srb,sre0));

    rectb.setVal(mx, srb0.by0, srb0.ex0, srb0.ey0);
    srb = srb0.getBoundedRegion(rectb);
    regions.push_back(Region(*this, REGION_ME, RECT_DIAG_SHAPE,srb,sre0));

    rectb.setVal(srb0.bx0, srb0.by0, mx, srb0.ey0);
    recte.setVal(mx + 1, srb0.by0, srb0.ex0, srb0.ey0);
    srb = srb0.getBoundedRegion(rectb);
    sre = srb0.getBoundedRegion(recte);
    regions.push_back(Region(*this,REGION_BM,RECT_VERT_SHAPE,srb,sre,sre0));

    rect.setVal(mx, srb0.by0, srb0.ex0, srb0.ey0);
    sr = srb0.getBoundedRegion(rect);
    regions.push_back(Region(*this, REGION_MM, TRI_SHAPE, sr, sr));

#ifdef DEBUG
    //const Region& ree = regions[0];
    const Region& rbb = regions[1];
    const Region& rbe = regions[2];
    const Region& rme = regions[3];
    const Region& rbm = regions[4];
    const Region& rmm = regions[5];
    Assert(rbb.sre.ex0 == rbm.sre.bx0);
    Assert(rbm.srb.ex0 == rmm.srb.bx0);
    Assert(rbe.srb.ex0 == rme.srb.bx0);
    Assert(rbe.sre == rme.sre);
#endif

  } else if (type == CARET_CUT_R) {
    srb0 = getSubRegion(REGION_BB).srb;
    sre0 = getSubRegion(REGION_EE).sre; 
    regions.clear();

    rect.setVal(mx, sre0.by0, sre0.ex0, sre0.ey0);
    sr = sre0.getBoundedRegion(rect);
    regions.push_back(Region(*this, REGION_EE, TRI_SHAPE, sr, sr));

    regions.push_back(Region(*this, REGION_BB, TRI_SHAPE, srb0, srb0));

    recte.setVal(mx + 1, sre0.by0, sre0.ex0, sre0.ey0);
    sre = sre0.getBoundedRegion(recte);
    regions.push_back(Region(*this, REGION_BE, RECT_DIAG_SHAPE,srb0,sre));

    rectb.setVal(sre0.bx0, sre0.by0, mx, sre0.ey0);
    recte.setVal(mx + 1, sre0.by0, sre0.ex0, sre0.ey0);
    srb = sre0.getBoundedRegion(rectb);
    sre = sre0.getBoundedRegion(recte);
    regions.push_back(Region(*this,REGION_ME,RECT_VERT_SHAPE,srb,sre,srb0));

    recte.setVal(sre0.bx0, sre0.by0, mx + 1, sre0.ey0);
    sre = sre0.getBoundedRegion(recte);
    regions.push_back(Region(*this, REGION_BM, RECT_DIAG_SHAPE,srb0,sre));

    rect.setVal(sre0.bx0, sre0.by0, mx + 1, sre0.ey0);
    sr = sre0.getBoundedRegion(rect);
    regions.push_back(Region(*this, REGION_MM, TRI_SHAPE, sr, sr));

#ifdef DEBUG
    const Region& ree = regions[0];
    //const Region& rbb = regions[1];
    const Region& rbe = regions[2];
    const Region& rme = regions[3];
    const Region& rbm = regions[4];
    const Region& rmm = regions[5];
    Assert(rbm.sre.ex0 == rbe.sre.bx0);
    Assert(rmm.sre.ex0 == rme.sre.bx0);
    Assert(rme.srb.ex0 == ree.srb.bx0);
    Assert(rbe.srb == rbm.srb);
#endif
  } else {
    Die("method undefined for type=%s", type2str(type));
  }
}
Vector<DPDeckPSCFG::DPRegion>
DPDeckPSCFG::DPRegion::
split(const Rect& rect, const Point& mpt, SplitType splitType) const
{
  Vector<DPRegion> vdpr;
  DPRegion dpRegion0;
  DPRegion dpRegion1;
  DPRegion dpRegion2;
  dpRegion0.setSkipSet(*pSkipSet);
  dpRegion1.setSkipSet(*pSkipSet);
  dpRegion2.setSkipSet(*pSkipSet);

  switch (type) {
  case TRIANGLE_CUT:
    {
      Rect rectb = {{rect0.b}, {rect.b}};
      Rect recte = {{rect.e}, {rect0.e}};
      const StripRegion& srb0 = getSubRegion(REGION_BB).srb;
      const StripRegion& sre0 = getSubRegion(REGION_EE).srb;
      const StripRegion& srb = srb0.getBoundedRegion(rectb);
      const StripRegion& sre = sre0.getBoundedRegion(recte);
      dpRegion0.setCaret(srb, sre);

      switch (splitType) {
      case SPLIT_TYPE_B: 
	dpRegion1.setTriangle(srb0.getBoundedRegion(rect));
	break;
      case SPLIT_TYPE_E:
	dpRegion1.setTriangle(sre0.getBoundedRegion(rect));
	break;
      default:
	Die("TRIANGLE_CUT:bad splitType=%d,rect[%d,%d,%d,%d],mpt[%d,%d]", 
	    splitType, rect.bx,rect.by,rect.ex,rect.ey,mpt.x,mpt.y);
      }

      vdpr.push_back(dpRegion0);
      vdpr.push_back(dpRegion1);
#ifdef DEBUG
      Assert(dpRegion0.getLeftChildRect() == dpRegion1.rect0);
#endif
    }
    break;
  case CARET_CUT_L:
    {
      const Point& pt0 = rect0.b;
      const Point& pt1 = getSubRegion(REGION_MM).srb.e0;
      const Point& pt2 = getSubRegion(REGION_EE).sre.b0;
      const Point& pt3 = rect0.e;
      const StripRegion& srb0 = getSubRegion(REGION_BB).srb;
      const StripRegion& srm0 = getSubRegion(REGION_MM).srb;
      const StripRegion& sre0 = getSubRegion(REGION_EE).sre;
      switch (splitType) {
      case SPLIT_TYPE_B: case SPLIT_TYPE_M: // CL -> N + T
	{
	  Rect rectb = {{pt0}, {rect.b}};
	  Rect rectm = {{rect.e}, {pt1}};
	  //Rect recte = {{pt2}, {pt3}};
	  const StripRegion& srb = srb0.getBoundedRegion(rectb);
	  const StripRegion& srm = srm0.getBoundedRegion(rectm);
	  const StripRegion& sre = sre0;
	  dpRegion0.setNose(srb, sre, srm);
	  if (splitType == SPLIT_TYPE_B) {
	    const StripRegion& sr = srb0.getBoundedRegion(rect);
	    dpRegion1.setTriangle(sr);
	  } else {// SPLIT_TYPE_M
	    const StripRegion& sr = srm0.getBoundedRegion(rect);
	    dpRegion1.setTriangle(sr);
	  }
	  Assert(dpRegion0.getLeftChildRect() == dpRegion1.rect0);
	  Assert(dpRegion0.getRightChildRect() == getLeftChildRect());
	}
	break;
      case SPLIT_TYPE_E: // CL -> C + C
	{
	  Rect rectob = {{pt0}, {rect.b}};
	  Rect rectoe = {{rect.e}, {pt3}};
	  Rect rectib = {{rect.b}, {pt1}};
	  Rect rectie = {{pt2}, {rect.e}};
	  const StripRegion& srob = srb0.getBoundedRegion(rectob);
	  const StripRegion& sroe = sre0.getBoundedRegion(rectoe);
	  const StripRegion& srib = srm0.getBoundedRegion(rectib);
	  const StripRegion& srie = sre0.getBoundedRegion(rectie);
	  dpRegion0.setCaret(srob, sroe);
	  dpRegion1.setCaret(srib, srie);
	  Assert(dpRegion0.getLeftChildRect() == dpRegion1.rect0);
	  Assert(dpRegion1.getLeftChildRect() == getLeftChildRect());
	}
	break;
      default:
	Die("CARET_CUT_L:bad splitType=%d,rect[%d,%d,%d,%d],mpt[%d,%d]", 
	    splitType, rect.bx,rect.by,rect.ex,rect.ey,mpt.x,mpt.y);
	break;
      }
      vdpr.push_back(dpRegion0);
      vdpr.push_back(dpRegion1);
    }
    break;
  case CARET_CUT_R:
    {
      const Point& pt0 = rect0.b;
      const Point& pt1 = getSubRegion(REGION_BB).srb.e0;
      const Point& pt2 = getSubRegion(REGION_MM).srb.b0;
      const Point& pt3 = rect0.e;
      const StripRegion& srb0 = getSubRegion(REGION_BB).srb;
      const StripRegion& srm0 = getSubRegion(REGION_MM).srb;
      const StripRegion& sre0 = getSubRegion(REGION_EE).sre;
      switch (splitType) {
      case SPLIT_TYPE_B:  // CR -> C + C
	{
	  Rect rectob = {{pt0}, {rect.b}};
	  Rect rectoe = {{rect.e}, {pt3}};
	  Rect rectib = {{rect.b}, {pt1}};
	  Rect rectie = {{pt2}, {rect.e}};
	  const StripRegion& srob = srb0.getBoundedRegion(rectob);
	  const StripRegion& sroe = sre0.getBoundedRegion(rectoe);
	  const StripRegion& srib = srb0.getBoundedRegion(rectib);
	  const StripRegion& srie = srm0.getBoundedRegion(rectie);
	  dpRegion0.setCaret(srob, sroe);
	  dpRegion1.setCaret(srib, srie);

	  Assert(dpRegion0.getLeftChildRect() == dpRegion1.rect0);
	  Assert(dpRegion1.getLeftChildRect() == getLeftChildRect());
	} 
	break;
      case SPLIT_TYPE_M: case SPLIT_TYPE_E: // CR -> N + T
	{
	  //Rect rectb = {{pt0}, {pt1}};
	  Rect rectm = {{pt2}, {rect.b}};
	  Rect recte = {{rect.e}, {pt3}};
	  const StripRegion& srb = srb0;
	  const StripRegion& srm = srm0.getBoundedRegion(rectm);
	  const StripRegion& sre = sre0.getBoundedRegion(recte);
	  dpRegion0.setNose(srb, sre, srm);
	  if (splitType == SPLIT_TYPE_M) {
	    const StripRegion& sr = srm0.getBoundedRegion(rect);
	    dpRegion1.setTriangle(sr);
	  } else {
	    const StripRegion& sr = sre0.getBoundedRegion(rect);
	    dpRegion1.setTriangle(sr);
	  }
	  Assert(dpRegion0.getLeftChildRect() == getLeftChildRect());
	  Assert(dpRegion0.getRightChildRect() == dpRegion1.rect0);
	}
	break;
      default:
	Die("CARET_CUT_R:bad splitType=%d,rect[%d,%d,%d,%d],mpt[%d,%d]", 
	    splitType, rect.bx,rect.by,rect.ex,rect.ey,mpt.x,mpt.y);
	break;
      }
      vdpr.push_back(dpRegion0);
      vdpr.push_back(dpRegion1);
    }
    break;
  case NOSE:
    {
      const Point& pt0 = rect0.b;
      const Point& pt1 = getSubRegion(REGION_BB).srb.e0;
      const Point& pt2 = getSubRegion(REGION_MM).srb.b0;
      const Point& pt3 = getSubRegion(REGION_MM).srb.e0;
      const Point& pt4 = getSubRegion(REGION_EE).srb.b0;
      const Point& pt5 = rect0.e;
      const StripRegion& srb0 = getSubRegion(REGION_BB).srb;
      const StripRegion& srm0 = getSubRegion(REGION_MM).srb;
      const StripRegion& sre0 = getSubRegion(REGION_EE).sre;

      Rect rectob = {{pt0}, {rect.b}};
      Rect rectoe = {{rect.e}, {pt5}};
      const StripRegion& srob = srb0.getBoundedRegion(rectob);
      const StripRegion& sroe = sre0.getBoundedRegion(rectoe);
      dpRegion0.setCaret(srob, sroe);

      Rect recti1b = {{rect.b}, {pt1}};
      Rect recti1e = {{pt2}, {mpt}};
      const StripRegion& sri1b = srb0.getBoundedRegion(recti1b);
      const StripRegion& sri1e = srm0.getBoundedRegion(recti1e);
      dpRegion1.setCaret(sri1b, sri1e);

      Rect recti2b = {{mpt}, {pt3}};
      Rect recti2e = {{pt4}, {rect.e}};
      const StripRegion& sri2b = srm0.getBoundedRegion(recti2b);
      const StripRegion& sri2e = sre0.getBoundedRegion(recti2e);
      dpRegion2.setCaret(sri2b, sri2e);

      Assert(dpRegion0.getLeftChildRect().b == dpRegion1.rect0.b);
      Assert(dpRegion0.getLeftChildRect().e == dpRegion2.rect0.e);
      Assert(dpRegion1.getLeftChildRect() == getLeftChildRect());
      Assert(dpRegion2.getLeftChildRect() == getRightChildRect());

      vdpr.push_back(dpRegion0);
      vdpr.push_back(dpRegion1);
      vdpr.push_back(dpRegion2);
    }
    break;
  default:
    Die("NOSE: bad type=%d", type);
    break;
  }
  return vdpr;
}
void
DPDeckPSCFG::DPRegion::
setSkipSet(SkipSet& skipSet)
{
  pSkipSet = &skipSet;
}
const SkipSet&
DPDeckPSCFG::DPRegion::
getSkipSet() const
{
  return *pSkipSet;
}
void
DPDeckPSCFG::DPRegion::
setSkipSize(int skipSize0)
{
  pSkipSet->setSkipSize(skipSize0);
}
int
DPDeckPSCFG::DPRegion::
getSkipSize() const
{
  return pSkipSet->getSkipSize();
}
bool
DPDeckPSCFG::DPRegion::
isSkipPoint(const Point& pt) const
{
  return pSkipSet->isSkipPoint(pt);
}
bool
DPDeckPSCFG::DPRegion::
isSkipRect(const Rect& rect) const
{
  return pSkipSet->isSkipRect(rect);
}
bool
DPDeckPSCFG::DPRegion::
hasLeftChildRect() const
{
  switch (type) {
  case CARET: return true;
  case CARET_CUT_L: return true;
  case CARET_CUT_R: return true;
  case NOSE: return true;
  default: return false;
  }
}
bool
DPDeckPSCFG::DPRegion::
hasRightChildRect() const
{
  switch (type) {
  case NOSE: return true;
  default: return false;
  }
}
Rect
DPDeckPSCFG::DPRegion::
getLeftChildRect() const
{
  Rect rect(Rect::null());
  switch (type) {
  case CARET:
    {
      const Region& rbe = getSubRegion(REGION_BE);
      rect.setVal(rbe.srb.e0, rbe.sre.b0);
    }
    break;
  case CARET_CUT_L:
    {
      const Region& rme = getSubRegion(REGION_ME);
      rect.setVal(rme.srb.e0, rme.sre.b0);
    }
    break;
  case CARET_CUT_R:
    {
      const Region& rbm = getSubRegion(REGION_BM);
      rect.setVal(rbm.srb.e0, rbm.sre.b0);
    }
    break;
  case NOSE:
    {
      const Region& rbm = getSubRegion(REGION_BM);
      rect.setVal(rbm.srb.e0, rbm.sre.b0);
    }
    break;
  default:
    Die("bad dpRegion.type=%d", type);
    break;
  }
  return rect;
}
Rect
DPDeckPSCFG::DPRegion::
getRightChildRect() const
{
  Rect rect(Rect::null());
  switch (type) {
  case NOSE:
    {
      const Region& rme = getSubRegion(REGION_ME);
      rect.setVal(rme.srb.e0, rme.sre.b0);
    }
    break;
  default:
    Die("bad dpRegion.type=%d", type);
    break;
  }
  return rect;
}
const DPDeckPSCFG::DPRegion::Region&
DPDeckPSCFG::DPRegion::
findRegion(const Rect& rect) const
{
  Assert(isRegionType(TRIANGLE)
  	 || isRegionType(CARET)
  	 || isRegionType(NOSE));

  for (int i = 0; i < (int)regions.size(); i++) {
    if (regions[i].includes(rect)) return regions[i];
  }
  rect.print();
  print();
  for (int i = 0; i < (int)regions.size(); i++) {
    regions[i].draw();
  }
  Die("region not found");
  return *new Region(*this);
}
const DPDeckPSCFG::DPRegion::Region&
DPDeckPSCFG::DPRegion::
findRegion(const Region& region) const
{// not unique
  for (int i = 0; i < (int)regions.size(); i++) {
    if (regions[i].includes(region)) return regions[i];
  }
  region.print();
  print();
  for (int i = 0; i < (int)regions.size(); i++) {
    regions[i].draw();
  }
  Die("region not found");
  return *new Region(*this);
}
void
DPDeckPSCFG::DPRegion::
testConsistency() const
{
#ifdef DEBUG

#endif
}


