/*
 * PSCFGDPRegion.hpp
 *
 */
#ifndef DP_DECK_PSCFG_DP_REGION_HPP
#define DP_DECK_PSCFG_DP_REGION_HPP

#include "Util.hpp"
#include "Rect.hpp"
#include "SkipSet.hpp"
#include "DPDeckPSCFG.hpp"

class StripRegion;

class 
DPDeckPSCFG::
DPRegion 
{
public:
  class Region;
  enum RType {
    TRIANGLE = 0,
    CARET = 1,
    NOSE = 2,
    TRIANGLE_CUT = 3,
    CARET_CUT_L = 4,
    CARET_CUT_R = 5,
    INVALID_R_TYPE = 6,
    NTYPE = 6
  };
  enum RShape {
    TRI_SHAPE = 0,
    RECT_VERT_SHAPE = 1,
    RECT_DIAG_SHAPE = 2,
    INVALID_SHAPE = 3
  };
  enum RNum {
    REGION_EE = 0,
    REGION_BB = 1,
    REGION_BE = 2,
    REGION_ME = 3,
    REGION_BM = 4,
    REGION_MM = 5,
    INVALID_NUM = 6,
    RMAX = 6
  };
  static const char* type2str(int type0);
  static const char* shape2str(int shape0);
  static const char* regionNum2str(int regionNum0);
  static int getNRegion(int type1);
  static int regionNum2Index(int regionNum);

  Rect rect0;

  DPRegion();
  DPRegion(const DPRegion& dpRegion);
  ~DPRegion();
  DPRegion& operator=(const DPRegion& dpRegion);
  string toString() const;
  void print() const;

  bool isType(const int& type0) const;
  int getType() const {return type;}
  int getArea() const;
  bool isRegionType(int regionType) const;
  bool includes(const Rect& rect) const;
  bool includes(const Region& region) const;
  int getNRegion() const;
  const Region& getSubRegion(int regionNum) const;
  void setTriangle(const StripRegion& sr);
  void setCaret(const StripRegion& srb, const StripRegion& sre);
  void setNose(const StripRegion& srb, const StripRegion& sre,
	       const StripRegion& srm);
  bool isSplittable() const;
  void insertCut();
  Vector<DPRegion> split(const Rect& rect, const Point& mpt,
			 SplitType splitType) const;

  void setSkipSet(SkipSet& skipSet);
  void setSkipSize(int skipSize0);
  int getSkipSize() const;
  const SkipSet& getSkipSet() const;
  bool isSkipPoint(const Point& pt) const;
  bool isSkipRect(const Rect& rect) const;

  bool hasLeftChildRect() const;
  bool hasRightChildRect() const;
  Rect getLeftChildRect() const;
  Rect getRightChildRect() const;
  const Region& findRegion(const Rect& rect) const;// not unique
  const Region& findRegion(const Region& region) const;
  void testConsistency() const;

  template <bool forward0> class Iterator;
  template <bool forward0> class SkipRowIterator;
  template <bool forward0> class SkipIterator;

private:
  friend class DPTask;
  friend class Region;
  class VRegion;
  RType type;
  SkipSet* pSkipSet;
  VRegion& regions;
  
  RType getCutType() const;
  int getCutPoint() const;
};

#include "DPDeckPSCFGDPRegionRegion.hpp"
#endif
