/*
 * DPDeckPSCFGDPRegionRegion.cpp
 *
 */
#include "DPDeckPSCFGDPRegionRegion.hpp"

DPDeckPSCFG::DPRegion::Region::
Region(const DPRegion& dpRegion)
  : dpRegion(dpRegion),
    skipSet(dpRegion.getSkipSet()),
    regionNum(INVALID_NUM),
    shape(INVALID_SHAPE),
    srb(),
    sre(),
    srm(),
    srkb(srb, dpRegion.getSkipSet()),
    srke(sre, dpRegion.getSkipSet()),
    srkm(srm, dpRegion.getSkipSet()),
    bxk_beg(srkb.getFirstSkipRow()),
    bxk_last(srkb.getLastSkipRow()),
    exk_beg(srke.getFirstSkipRow()),
    exk_last(srke.getLastSkipRow()),
    mxk_beg(srkm.getFirstSkipRow()),
    mxk_last(srkm.getLastSkipRow()),
    bxn_beg(convertXK2XN(bxk_beg)),
    bxn_last(convertXK2XN(bxk_last)),
    exn_beg(convertXK2XN(exk_beg)),
    exn_last(convertXK2XN(exk_last)),
    mxn_beg(convertXK2XN(mxk_beg)),
    mxn_last(convertXK2XN(mxk_last))
{
}
DPDeckPSCFG::DPRegion::Region::
Region(const DPRegion& dpRegion, const Region& region)
  : dpRegion(dpRegion),
    skipSet(dpRegion.getSkipSet()),
    regionNum(region.regionNum),
    shape(region.shape),
    srb(region.srb),
    sre(region.sre),
    srm(region.srm),
    srkb(srb, dpRegion.getSkipSet()),
    srke(sre, dpRegion.getSkipSet()),
    srkm(srm, dpRegion.getSkipSet()),
    bxk_beg(srkb.getFirstSkipRow()),
    bxk_last(srkb.getLastSkipRow()),
    exk_beg(srke.getFirstSkipRow()),
    exk_last(srke.getLastSkipRow()),
    mxk_beg(srkm.getFirstSkipRow()),
    mxk_last(srkm.getLastSkipRow()),
    bxn_beg(convertXK2XN(bxk_beg)),
    bxn_last(convertXK2XN(bxk_last)),
    exn_beg(convertXK2XN(exk_beg)),
    exn_last(convertXK2XN(exk_last)),
    mxn_beg(convertXK2XN(mxk_beg)),
    mxn_last(convertXK2XN(mxk_last))
{
}
DPDeckPSCFG::DPRegion::Region::
Region(const DPRegion& dpRegion, int regionNum, int shape,
       const StripRegion& srb, const StripRegion& sre,
       const StripRegion& srm)
  : dpRegion(dpRegion),
    skipSet(dpRegion.getSkipSet()),
    regionNum(RNum(regionNum)),
    shape(RShape(shape)),
    srb(srb),
    sre(sre),
    srm(srm),
    srkb(srb, dpRegion.getSkipSet()),
    srke(sre, dpRegion.getSkipSet()),
    srkm(srm, dpRegion.getSkipSet()),
    bxk_beg(srkb.getFirstSkipRow()),
    bxk_last(srkb.getLastSkipRow()),
    exk_beg(srke.getFirstSkipRow()),
    exk_last(srke.getLastSkipRow()),
    mxk_beg(srkm.getFirstSkipRow()),
    mxk_last(srkm.getLastSkipRow()),
    bxn_beg(convertXK2XN(bxk_beg)),
    bxn_last(convertXK2XN(bxk_last)),
    exn_beg(convertXK2XN(exk_beg)),
    exn_last(convertXK2XN(exk_last)),
    mxn_beg(convertXK2XN(mxk_beg)),
    mxn_last(convertXK2XN(mxk_last))
{
}
DPDeckPSCFG::DPRegion::Region::
Region(const Region& region)
  : dpRegion(region.dpRegion),
    skipSet(region.skipSet),
    regionNum(RNum(region.regionNum)),
    shape(RShape(region.shape)),
    srb(region.srb),
    sre(region.sre),
    srm(region.srm),
    srkb(srb, region.skipSet),
    srke(sre, region.skipSet),
    srkm(srm, region.skipSet),
    bxk_beg(srkb.getFirstSkipRow()),
    bxk_last(srkb.getLastSkipRow()),
    exk_beg(srke.getFirstSkipRow()),
    exk_last(srke.getLastSkipRow()),
    mxk_beg(srkm.getFirstSkipRow()),
    mxk_last(srkm.getLastSkipRow()),
    bxn_beg(convertXK2XN(bxk_beg)),
    bxn_last(convertXK2XN(bxk_last)),
    exn_beg(convertXK2XN(exk_beg)),
    exn_last(convertXK2XN(exk_last)),
    mxn_beg(convertXK2XN(mxk_beg)),
    mxn_last(convertXK2XN(mxk_last))
{
}
DPDeckPSCFG::DPRegion::Region::
~Region()
{
}
DPDeckPSCFG::DPRegion::Region&
DPDeckPSCFG::DPRegion::Region::
operator=(const Region& region)
{
  Assert(&dpRegion == &region.dpRegion);
  Assert(&skipSet == &region.skipSet);
  regionNum = region.regionNum;
  shape = region.shape;
  srb = region.srb;
  sre = region.sre;
  srm = region.srm;
  srkb = region.srkb;
  srke = region.srke;
  srkm = region.srkm;
  bxk_beg = region.bxk_beg;
  bxk_last = region.bxk_last;
  exk_beg = region.exk_beg;
  exk_last = region.exk_last;
  mxk_beg = region.mxk_beg;
  mxk_last = region.mxk_last;
  bxn_beg = region.bxn_beg;
  bxn_last = region.bxn_last;
  exn_beg = region.exn_beg;
  exn_last = region.exn_last;
  mxn_beg = region.mxn_beg;
  mxn_last = region.mxn_last;
  return *this;
}
string
DPDeckPSCFG::DPRegion::Region::
toString() const
{
  ostringstream oss;
  oss << "regionType: " << dpRegion.type2str(dpRegion.getType()) << "\n";
  oss << "regionNum: " << dpRegion.regionNum2str(regionNum) << "\n";
  oss << "shape: "     << dpRegion.shape2str(shape) << "\n";
  oss << "srb:\n  "   << srb.toString();
  oss << "sre:\n  "   << sre.toString();
  oss << "srm:\n  "   << srm.toString();
  //oss << "bxk_beg: "  << bxk_beg << "\n";
  //oss << "bxk_last: " << bxk_last << "\n";
  //oss << "exk_beg: "  << exk_beg << "\n";
  //oss << "exk_last: " << exk_last << "\n";
  //oss << "mxk_beg: "  << mxk_beg << "\n";
  //oss << "mxk_last: " << mxk_last << "\n";
  //oss << "bxn_beg: "  << bxn_beg << "\n";
  //oss << "bxn_last: " << bxn_last << "\n";
  //oss << "exn_beg: "  << exn_beg << "\n";
  //oss << "exn_last: " << exn_last << "\n";
  //oss << "mxn_beg: "  << mxn_beg << "\n";
  //oss << "mxn_last: " << mxn_last << "\n";
  oss << "bxk : [ " << bxk_beg << ", " << bxk_last << " ], ";
  oss << "bxn : [ " << bxn_beg << ", " << bxn_last << " ]\n";
  oss << "exk : [ " << exk_beg << ", " << exk_last << " ], ";
  oss << "exn : [ " << exn_beg << ", " << exn_last << " ]\n";
  oss << "mxk : [ " << mxk_beg << ", " << mxk_last << " ], ";
  oss << "mxn : [ " << mxn_beg << ", " << mxn_last << " ]\n";
  return oss.str();
}
void
DPDeckPSCFG::DPRegion::Region::
print() const
{
  cout << toString() << flush;
}
void 
DPDeckPSCFG::DPRegion::Region::
draw(ostream& fo) const
{
  fo << "--- \n";
  fo << "srb: \n";
  srb.draw(fo);
  fo << "sre: \n";
  sre.draw(fo);
  fo << "srm: \n";
  srm.draw(fo);
}
void 
DPDeckPSCFG::DPRegion::Region::
draw() const
{
  draw(cout);
}
bool
DPDeckPSCFG::DPRegion::Region::
isRegionNum(int regionNum0) const
{
  return (regionNum == regionNum0);
}
bool
DPDeckPSCFG::DPRegion::Region::
isTriShape() const
{
  return (shape == TRI_SHAPE);
}
bool
DPDeckPSCFG::DPRegion::Region::
isRectVertShape() const
{
  return (shape == RECT_VERT_SHAPE);
}
bool
DPDeckPSCFG::DPRegion::Region::
isRectDiagShape() const
{
  return (shape == RECT_DIAG_SHAPE);
}
bool
DPDeckPSCFG::DPRegion::Region::
isRectShape() const
{
  return (isRectVertShape() || isRectDiagShape());
}

bool
DPDeckPSCFG::DPRegion::Region::
includes(const Rect& rect) const
{
  if (!rect.isNormal()) return false;

  return (srb.includes(rect.b) && sre.includes(rect.e));
}
bool
DPDeckPSCFG::DPRegion::Region::
includes(const Region& region) const
{
  if (!srb.rect0.includes(region.srb.rect0)
      || !sre.rect0.includes(region.sre.rect0)) return false;

  Rect bbox_b = {
    {{region.srb.bx0, srb.by0}},
    {{region.srb.ex0, srb.ey0}}
  };
  StripRegion::RowIterator<true> itb(srb, bbox_b);
  StripRegion::RowIterator<true> itb1(region.srb);
  while (itb.getNext()) {
    itb1.getNext();
    Assert(itb.ly > 0 && itb1.ly > 0 && itb.x == itb1.x);
    if (itb1.y_beg < itb.y_beg || itb.y_last < itb1.y_last) return false;
  }
  
  Rect bbox_e = {
    {{region.sre.bx0, sre.by0}},
    {{region.sre.ex0, sre.ey0}}
  };
  StripRegion::RowIterator<true> ite(sre, bbox_e);
  StripRegion::RowIterator<true> ite1(region.sre);
  while (ite.getNext()) {
    ite1.getNext();
    Assert(ite.ly > 0 && ite1.ly > 0 && ite.x == ite1.x);
    if (ite1.y_beg < ite.y_beg || ite.y_last < ite1.y_last) return false;
  }
  return true;
}
int
DPDeckPSCFG::DPRegion::Region::
getArea() const
{
  return (srb.getArea() + (isTriShape() ? 0 : sre.getArea()));
}
int
DPDeckPSCFG::DPRegion::Region::
getB0Size() const
{
  if (bxn_beg < 0) return 0;

  return (bxn_last - bxn_beg + 1);
}
int
DPDeckPSCFG::DPRegion::Region::
getE0Size() const
{
  if (exn_beg < 0) return 0;
  if (bxn_beg == exn_beg) return 0;

  return (exn_last - exn_beg + 1);
}
int
DPDeckPSCFG::DPRegion::Region::
getM0Size() const
{
  if (mxn_beg < 0) return 0;

  return (mxn_last - mxn_beg + 1);
}
int
DPDeckPSCFG::DPRegion::Region::
getVTBSizeBE(const Trapezoid& t)
{// should test against N^2 procedures
  const int& lx = (t.ex - t.bx + 1);
  const int& ly = (t.ey - t.by + 1);
  const int& j = (t.ey - t.getEXBY() + 1);
  int n = 0;
  switch (t.type) {
  case Trapezoid::LV_RV:
    n += (lx * (lx+1) * ly * (ly+1)) / 4;
    break;
  case Trapezoid::LV_RD:
    n += (lx * (lx+1) * (ly+1) * (ly+1)) * 3;
    n += -(lx * (lx+1) * (2*lx+1) * (ly+1));
    n /= 12;
    break;
  case Trapezoid::LD_RV:
    n += (lx * lx * (lx+1) * (lx+1)) * 3;
    n += (lx * (lx+1) * (2*lx+1) * (2*j-1)) * 2;
    n += (lx * (lx+1) * j * (j-1)) * 6;
    n += -((lx-1) * lx * (lx+1) * (lx+2));
    n /= 24;
    break;
  case Trapezoid::LD_RD:
    n += (lx * (lx+1) * j * (j-1)) * 6;
    n += (lx * (lx+1) * (lx+2) * j) * 4;
    n += -((lx-1) * lx * (lx+1) * (lx+2));
    if (lx > j) {
      n += ((lx-j-1) * (lx-j) * (lx-j+1) * (lx-j+2));
    }
    n /= 24;
    break;
  default:
    Die("bad type");
  }
  return n;
}
int
DPDeckPSCFG::DPRegion::Region::
getVTBSizeProdSum(int trapezoidType, int width, int blx, int elx,
		  int areaE)
{
  const int& w = width;
  const int& b = blx;
  const int& e = elx;
  const int& a = areaE;
  const int& s0 = +b * a * w;
  const int& s1 = -b * e * ((w * (w - 1)) / 2);
  const int& s2 = -a * ((w * (w - 1)) / 2);
  const int& s3 = -b * (((w - 2) * (w - 1) * w) / 6);
  const int& s4 = +e * (((w - 1) * w * (2 * w - 1)) / 6);
  const int& s5 = +((w - 2) * (w - 1) * w * (3 * w - 1)) / 24;
  // trapezoid type is L for et R for bt
  switch (trapezoidType) {
  case Trapezoid::LV_RV :
    return s0 + s1;
  case Trapezoid::LV_RD :
    return s0 + s1 + s2 + s4;
  case Trapezoid::LD_RV :
    return s0 + s1 + s3;
  case Trapezoid::LD_RD :
    return s0 + s1 + s2 + s3 + s4 + s5;
  default:
    Die("bad type");
    return (-1);
  }
}
int
DPDeckPSCFG::DPRegion::Region::
getVTBSize(const StripRegion& srb, const StripRegion& sre) const
{
  const bool isTriShape = (srb.rect0 == sre.rect0);
  int n = 0;
  for (int ei = 0; ei < (int) sre.size(); ei++) {
    const Trapezoid& et = sre[ei];

    if (isTriShape) {
      n += getVTBSizeBE(et);
    }

    int bi = 0;
    const int& bi_last = ((isTriShape ? ei : (int)srb.size()) - 1);
    while (bi <= bi_last) {
      const Trapezoid& bt = srb[bi];
      n += bt.getAreaLeft(et.by - 1) * et.getArea();
      bi++;
    }

    bi = 0;
    int by = et.by;
    int ld_last = ((et.bxey < et.exby) ? et.bxey : (et.exby - 1));
    int elx = (et.getEX(et.by) - et.bx + 1);
    int etAreaRight = et.getArea();  
    while (bi <= bi_last) {
      const Trapezoid& bt = srb[bi];
      const Trapezoid& bt_last = srb[bi_last];
      if (bt.ey < by) {
	bi++;
	continue;
      }

      if (by <= ld_last) {//LD
	if (by <= bt.bxey) {//RV
	  const int& y_last = min(bt.bxey, ld_last);
	  const int& width = (y_last - by + 1);
	  const int& blx = (bt_last.ex - bt.getBX(by) + 1);

	  n += getVTBSizeProdSum(Trapezoid::LD_RV,
				 width, blx, elx, etAreaRight);

	  etAreaRight -= Trapezoid::getArea(width, elx, elx + width - 1);
	  elx += width;
	  by = (y_last + 1);
	} else {//RD
	  const int& y_last = min(bt.ey, ld_last);
	  const int& width = (y_last - by + 1);
	  const int& blx = (bt_last.ex - bt.getBX(by) + 1);

	  n += getVTBSizeProdSum(Trapezoid::LD_RD,
				 width, blx, elx, etAreaRight);

	  etAreaRight -= Trapezoid::getArea(width, elx, elx +width - 1);
	  elx += width;
	  by = (y_last + 1);
	}
      } else {//LV
	if (by <= bt.bxey) {//RV
	  const int& y_last = bt.bxey;
	  const int& width = y_last - by + 1;
	  const int& blx = bt_last.ex - bt.getBX(by) + 1;

	  n += getVTBSizeProdSum(Trapezoid::LV_RV,
				 width, blx, elx, etAreaRight);

	  etAreaRight -= Trapezoid::getArea(width, elx, elx);
	  // elx unchanged
	  by = (y_last + 1);
	} else {//RD
	  const int& y_last = bt.ey;
	  const int& width = (y_last - by + 1);
	  const int& blx = bt_last.ex - bt.getBX(by) + 1;

	  n += getVTBSizeProdSum(Trapezoid::LV_RD,
				 width, blx, elx, etAreaRight);

	  etAreaRight -= Trapezoid::getArea(width, elx, elx);
	  // elx unchanged
	  by = (y_last + 1);
	}
      }
    }
  }
  return n;
}
int
DPDeckPSCFG::DPRegion::Region::
getVEYRangeSize() const
{
  if (isTriShape()) {
    int n = 0;
    int lx = sre.rect0.lengthX();
    StripRegion::RowIterator<true> it(srb);
    while (it.getNext()) {
      n += (it.ly * (lx--));
    }
    return n;
  } else {
    return (srb.getArea() * sre.rect0.lengthX());
  }
}
int
DPDeckPSCFG::DPRegion::Region::
getVSTBVEYRangeSize() const
{
  if (isTriShape()) {
    int n = 0;
    int lx = srke.getNSkipRow();
    StripRegionSkip::RowIterator<true> it(srkb);
    while (it.getNext()) {
      n += (it.ly * (lx--));
    }
    return n;
  } else {
    return (srkb.getArea() * srke.getNSkipRow());
  }
}
int
DPDeckPSCFG::DPRegion::Region::
getVSTEVBYRangeSize() const
{
  if (isTriShape()) {
    int n = 0;
    int lx = srkb.getNSkipRow();
    StripRegionSkip::RowIterator<false> it(srke);
    while (it.getNext()) {
      n += (it.ly * (lx--));
    }
    return n;
  } else {
    return (srke.getArea() * srkb.getNSkipRow());
  }
}
int
DPDeckPSCFG::DPRegion::Region::
getVTBSize() const
{
  return getVTBSize(srb, sre);
}
int
DPDeckPSCFG::DPRegion::Region::
getVTBIndex(const Rect& rect) const
{
  int index = 0;
  StripRegion::Iterator<false> itb(srb);
  Rect rect1 = {{rect.b}, {sre.e0}};
  while (itb.getNext()) {
    if ((rect.bx < itb.x)
	|| (rect.bx == itb.x && rect.by < itb.y)) {
      index += sre.getAreaBottomRight(itb.pt);
      continue;
    } else if (rect.b == itb.pt) {
      index += sre.getBoundedRegion(rect1).getAreaBefore(rect.e);
      return index;
    }
    rect.print();
    print();
    draw();
    Die("never reach here");
  }
  rect.print();
  print();
  draw();
  Die("never reach here");
  return (-1);
}
int
DPDeckPSCFG::DPRegion::Region::
getVSkipArea() const
{
  enum MarginSize {
    margin_b = 1,
    margin_e = 1
  };
  int n = 0;
  for (int y = srb.by0; y <= sre.ey0; y++) {
    if (srb.ey0 < y && y < sre.by0) continue;

    n += margin_b;
    const int bxn = getBXN(y);
    const int exn = getEXN(y);
    if (bxn >= 0) {
      n += (exn - bxn + 1);
    }
    n += margin_e;
  }
  return n;
}
int
DPDeckPSCFG::DPRegion::Region::
getVSkipIndex(const Point& pt) const
{
  enum MarginSize {
    margin_b = 1,
    margin_e = 1
  };
  Assert(srb.includes(pt) || sre.includes(pt));
  Assert(skipSet.isSkipRow(pt.x));
  int n = 0;
  for (int y = srb.by0; y < pt.y; y++) {
    if (srb.ey0 < y && y < sre.by0) continue;

    n += margin_b;
    const int bxn = getBXN(y);
    const int exn = getEXN(y);
    if (bxn >= 0) {
      n += (exn - bxn + 1);
    }
    n += margin_e;
  }
  n += margin_b;
  const int bxn = getBXN(pt.y);
  const int exn = convertXK2XN(pt.x);
  Assert(bxn >= 0 && exn >= 0);
  n += (exn - bxn);
  return n;
}
int
DPDeckPSCFG::DPRegion::Region::
getVSTSize() const
{
  StripRegionSkip::RowIterator<true> it(srkb);
  Rect bbox = srb.rect0;
  int n = 0;
  int skipAreaTopLeft;

  for (bbox.ey = sre.by0; bbox.ey <= srb.ey0; bbox.ey++) {
    const int& ex_beg = sre.getBX(bbox.ey);
    const int& ex_last = sre.getEX(bbox.ey);
    skipAreaTopLeft = 0;
    it.reset(bbox);
    while (it.getNext()) {
      if (it.ly <= 0) continue;

      skipAreaTopLeft += it.ly;
      if (isTriShape()) {
	if ((it.y_last == bbox.ey) 
	    && (ex_beg <= it.x && it.x <= ex_last)) {
	  n += skipAreaTopLeft;
	}
      }
    }
    if (isRectShape()) {
      n += (skipAreaTopLeft
	    * skipSet.getNSkipPointX(ex_beg, ex_last, bbox.ey));
    }
  }
  if (isRectShape()) {
    n += srkb.getArea() * srke.getAreaRight(srb.ey0 + 1);
  }

#ifdef DEBUG
  {
    StripRegionSkip::Iterator0<false> itb0(srkb);
    StripRegionSkip::Iterator0<true> ite0(srke);
    Rect bbox0 = sre.rect0;
    int n0 = 0;
    while (itb0.getNext()) {
      bbox0.b = itb0.pt;
      ite0.reset(bbox0);
      while (ite0.getNext()) {
	n0++;
      }
    }
    Assert(n == n0);
  }
#endif

  return n;
}
int
DPDeckPSCFG::DPRegion::Region::
getVSTBIndex(const Rect& rect) const
{
  Assert(dpRegion.isSkipRect(rect));
  StripRegionSkip::Iterator<false> it(srkb);
  int index = 0;
  while (it.getNext()) {
    if (it.pt == rect.b) break;

    index += srke.getAreaBottomRight(it.pt);
  }
  Rect rect1 = {{rect.b}, {sre.e0}};
  index += srke.getBoundedRegion(rect1).getAreaBefore(rect.e);
  return index;
}
int
DPDeckPSCFG::DPRegion::Region::
getVSTEIndex(const Rect& rect) const
{
  Assert(dpRegion.isSkipRect(rect));
  StripRegionSkip::Iterator<true> it(srke);
  int index = 0;
  while (it.getNext()) {
    if (it.pt == rect.e) break;

    index += srkb.getAreaTopLeft(it.pt);
  }
  Rect rect1 = {{srb.b0}, {rect.e}};
  index += srkb.getBoundedRegion(rect1).getAreaBefore(rect.b);
  return index;
}
int
DPDeckPSCFG::DPRegion::Region::
getVSFBWBSize(const int& bx) const
{
  int n = 0;
  Assert(srb.bx0 <= bx && bx <= srb.ex0);
  const int& by_beg = srb.getBY(bx);
  const int& by_last = srb.getEY(bx);
  Point pt = {bx, by_beg};
  for (pt.y = by_beg; pt.y <= by_last; pt.y++) {
    n += sre.getAreaBottomRight(pt);
  }
#ifdef DEBUG
  Rect rect = {{{bx, srb.getBY(bx)}}, {{sre.ex0, sre.ey0}}};
  Assert(n == getVSFBWBIndex(rect) + 1);
#endif
  return n;
}
int
DPDeckPSCFG::DPRegion::Region::
getVSFBWESize(const int& ex) const
{
  int n = 0;
  Assert(sre.bx0 <= ex && ex <= sre.ex0);
  const int& ey_beg = sre.getBY(ex);
  const int& ey_last = sre.getEY(ex);
  Point pt = {ex, ey_beg};
  for (pt.y = ey_beg; pt.y <= ey_last; pt.y++) {
    n += srb.getAreaTopLeft(pt);
  }
#ifdef DEBUG
  Rect rect = {{{srb.bx0, srb.by0}}, {{ex, sre.getEY(ex)}}};
  Assert(n == getVSFBWEIndex(rect) + 1);
#endif
  return n;
}
int 
DPDeckPSCFG::DPRegion::Region::
getVSFBWBSizeI() const
{
  return getVSFBWBSize(srb.bx0);
}
int
DPDeckPSCFG::DPRegion::Region::
getVSFBWBSizeO() const
{
  Assert(!isTriShape());
  return getVSFBWBSize(srb.ex0);
}
int
DPDeckPSCFG::DPRegion::Region::
getVSFBWESizeI() const
{
  return getVSFBWESize(sre.ex0);
}
int
DPDeckPSCFG::DPRegion::Region::
getVSFBWESizeO() const
{
  Assert(!isTriShape());
  return getVSFBWESize(sre.bx0);
}
int
DPDeckPSCFG::DPRegion::Region::
getVSFBWBIndex(const Rect& rect) const
{
  int n = 0;
  Assert(includes(rect));
  //const int& by_beg = srb.getBY(rect.bx);
  const int& by_last = srb.getEY(rect.bx);
  Assert(srb.getBY(rect.bx) <= rect.by && rect.by <= by_last);
  Point pt = {rect.bx, by_last};
  for (pt.y = by_last; pt.y > rect.by; pt.y--) {
    n += sre.getAreaBottomRight(pt);
  }
  const Rect bbox = {{{rect.bx, rect.by}}, {{sre.ex0, sre.ey0}}};
  const StripRegion& sr = sre.getBoundedRegion(bbox);
  n += sr.getAreaBefore(rect.e);
  return n;
}
int
DPDeckPSCFG::DPRegion::Region::
getVSFBWEIndex(const Rect& rect) const
{
  int n = 0;
  Assert(includes(rect));
  const int& ey_beg0 = sre.getBY(rect.ex);
  const int& ey_last = sre.getEY(rect.ex);
  StripRegion::Iterator<false> itb(srb);
  while (itb.getNext()) {
    const int& ey_beg = max(itb.pt.y, ey_beg0);
    if (itb.pt == rect.b) {
      n += (rect.ey - ey_beg);
      break;
    } else {
      n += (ey_last - ey_beg + 1);
    }
  }
  return n;
}
int
DPDeckPSCFG::DPRegion::Region::
getVSFBWBIndexI(const int& by, const int& ex, const int& ey) const
{
  Rect rect = {{{srb.bx0, by}}, {{ex, ey}}};
  return getVSFBWBIndex(rect);
}
int
DPDeckPSCFG::DPRegion::Region::
getVSFBWBIndexO(const int& by, const int& ex, const int& ey) const
{
  Assert(!isTriShape());
  Rect rect = {{{srb.ex0, by}}, {{ex, ey}}};
  return getVSFBWBIndex(rect);
}
int
DPDeckPSCFG::DPRegion::Region::
getVSFBWEIndexI(const int& bx, const int& by, const int& ey) const
{
  Rect rect = {{{bx, by}}, {{sre.ex0, ey}}};
  return getVSFBWEIndex(rect);
}
int
DPDeckPSCFG::DPRegion::Region::
getVSFBWEIndexO(const int& bx, const int& by, const int& ey) const
{
  Assert(!isTriShape());
  Rect rect = {{{bx, by}}, {{sre.bx0, ey}}};
  return getVSFBWEIndex(rect);
}
Rect
DPDeckPSCFG::DPRegion::Region::
getRectFromVSTBArea(const int& areaSkip) const
{
  Assert(areaSkip > 0);
  StripRegionSkip::Iterator<false> it(srkb);
  int i = (areaSkip - 1);
  while (it.getNext()) {
    const int& n1 = srke.getAreaBottomRight(it.pt);
    if (i < n1) {
      Rect rect = {{it.pt}, {sre.e0}};
      const StripRegionSkip& srk = srke.getBoundedRegion(rect);
      const Rect& rectk = srk.getRectK();
      rect.e = srk.advance(rectk.b, i);

      Assert(dpRegion.isSkipRect(rect));
      return rect;
    }
    i -= n1;
  }
  print();
  Die("areaSkip=%d too large", areaSkip);
  return Rect::null();
}
int
DPDeckPSCFG::DPRegion::Region::
getVBESize() const
{
  if (isTriShape()) {
    return srb.getAreaMargin();
  } else {
    return (srb.getAreaMargin() + sre.getAreaMargin());
  }
}
int
DPDeckPSCFG::DPRegion::Region::
getBXK(int y) const
{
  int bxk = srkb.getFirstSkipRow(y);
  if (bxk >= 0) return bxk;

  return srke.getFirstSkipRow(y);
}
int
DPDeckPSCFG::DPRegion::Region::
getEXK(int y) const
{
  int exk = srke.getLastSkipRow(y);
  if (exk >= 0) return exk;

  return srkb.getLastSkipRow(y);
}
int
DPDeckPSCFG::DPRegion::Region::
getBXN(int y) const
{
  const int& bxk = getBXK(y);
  const int& bxn = convertXK2XN(bxk);
  return bxn;
}
int
DPDeckPSCFG::DPRegion::Region::
getEXN(int y) const
{
  const int& exk = getEXK(y);
  const int& exn = convertXK2XN(exk);
  return exn;
}
int
DPDeckPSCFG::DPRegion::Region::
convertXK2XN(int xk) const
{
  if (xk < 0) return (-1);

  Assert(skipSet.isSkipRow(xk));
  if (bxk_beg <= xk && xk <= bxk_last) {
    return skipSet.getNSkipRow(bxk_beg, xk - 1);
  } else if (exk_beg <= xk && xk <= exk_last) {
    return (srkb.getNSkipRow()
	    + skipSet.getNSkipRow(exk_beg, xk - 1));
  } else if (mxk_beg <= xk && xk <= mxk_last) {
    return (srkb.getNSkipRow() 
	    + srke.getNSkipRow()
	    + skipSet.getNSkipRow(mxk_beg, xk - 1));
  } else {
    return (-1);
  }
}
int
DPDeckPSCFG::DPRegion::Region::
convertXN2XK(int xn) const
{
  const int& dxk = skipSet.getSkipRowPeriod();
  if (xn < 0) return (-1);// take into account bxn_last < 0 etc.

  int xk = (-1);
  if (bxn_beg <= xn && xn <= bxn_last) {
    xk = (bxk_beg + (xn - bxn_beg) * dxk);
    Assert(xn == convertXK2XN(xk));
  } else if (exn_beg <= xn && xn <= exn_last) {
    xk = (exk_beg + (xn - exn_beg) * dxk);
     Assert(xn == convertXK2XN(xk));
  } else if (mxn_beg <= xn && xn <= mxn_last) {
    xk = (mxk_beg + (xn - mxn_beg) * dxk);
    Assert(xn == convertXK2XN(xk));
  }
  return xk;
}
