/*
 * DPDeckPSCFGDPRegionRegionIterators.cpp
 *
 */
#ifndef DP_DECK_PSCFG_DP_REGION_REGION_ITERATORS_CPP
#define DP_DECK_PSCFG_DP_REGION_REGION_ITERATORS_CPP
#include "DPDeckPSCFGDPRegionRegion.hpp"

template <bool inside0>
DPDeckPSCFG::DPRegion::Region::Iterator<inside0>::
Iterator(const Region& region)
  : rect(Rect::null()),
    region(region),
    itb(region.srb),
    ite(region.sre)
{
  reset();
}
template <bool inside0>
DPDeckPSCFG::DPRegion::Region::Iterator<inside0>::
Iterator(const Iterator& it)
  : rect(it.rect),
    region(it.region),
    itb(it.itb),
    ite(it.ite)
{
}
template <bool inside0>
DPDeckPSCFG::DPRegion::Region::Iterator<inside0>::
~Iterator()
{
}
template <bool inside0>
void
DPDeckPSCFG::DPRegion::Region::Iterator<inside0>::
reset()
{
  itb.reset();
  if (itb.getNext()) {
    rect.setVal(itb.pt, region.sre.e0);
    ite.reset(rect);
  } else {
    rect.setNull();
    ite.reset(rect);
  }
}
template <bool inside0>
bool
DPDeckPSCFG::DPRegion::Region::Iterator<inside0>::
getNext()
{
  while (!ite.getNext()) {
    if (!itb.getNext()) return false;

    rect.setVal(itb.pt, region.sre.e0);
    ite.reset(rect);
  }
  rect.e = ite.pt;
  return true;
}

template <bool inside0>
DPDeckPSCFG::DPRegion::Region::BFIterator<inside0>::
BFIterator(const Region& region)
  : rect1(Rect::null()),
    rect2(Rect::null()),
    mpt(Point::null()),
    dpRegion(region.dpRegion),
    skipSet(region.skipSet),
    region(region),
    pitm(NULL),
    pitm_b(NULL),
    bfType(-1),
    rect(Rect::null())
{

}
template <bool inside0>
DPDeckPSCFG::DPRegion::Region::BFIterator<inside0>::
~BFIterator()
{
  if (pitm) {
    delete pitm;
    pitm = NULL;
  }
  if (pitm_b) {
    delete pitm_b;
    pitm_b = NULL;
  }
}

template <bool inside0>
string
DPDeckPSCFG::DPRegion::Region::BFIterator<inside0>::
toString() const
{
  ostringstream oss;
  oss << "bfType: " << bfType << "\n";
  oss << "rect: " << rect.toString() << "\n";
  oss << "rect1: " << rect1.toString() << "\n";
  oss << "rect2: " << rect2.toString() << "\n";
  oss << "mpt: " << mpt.toString() << "\n";
  return oss.str();
}
template <bool inside0>
void
DPDeckPSCFG::DPRegion::Region::BFIterator<inside0>::
print() const
{
  cout << toString() << flush;
}

template <bool inside0>
void
DPDeckPSCFG::DPRegion::Region::BFIterator<inside0>::
reset(const Rect& rect1, const int& bfType1)
{
  Assert(region.includes(rect1));
  Assert(skipSet.isSkipRect(rect1));
  rect = rect1;
  bfType = bfType1;
  reset();
}
template <bool inside0>
void
DPDeckPSCFG::DPRegion::Region::BFIterator<inside0>::
reset()
{
  if (pitm) {
    delete pitm;
    pitm = NULL;
  }
  if (pitm_b) {
    delete pitm_b;
    pitm_b = NULL;
  }
  if (inside0) {
    rect1.b = rect.b;
    rect2.e = rect.e;
    if (region.isTriShape()) {
      Assert(bfType == BF_TYPE_0);
      pitm = new StripRegionSkip::Iterator<true>(region.srke);
    } else if (region.isRectShape()) {
      switch (bfType) {
      case BF_TYPE_0:
	pitm = new StripRegionSkip::Iterator<true>(region.srkb);
	break;  
      case BF_TYPE_1:
	pitm_b = new StripRegionSkip::Iterator<false>(region.srke);
	break;
      case BF_TYPE_2:
	pitm = new StripRegionSkip::Iterator<true>(region.srkm);
	break;
      default:
	Die("bad bfType=%d", bfType);
	break;
      }
    }
  } else { //outside
    Rect rect3 = rect;
    switch (bfType) {
    case BF_TYPE_0:
      rect1.b = rect.b;
      rect2.b = rect.e;
      pitm = new StripRegionSkip::Iterator<true>(region.srke);
      rect.b = rect3.e;
      rect.e = region.sre.rect0.e;
      break;
    case BF_TYPE_1:
      rect1.e = rect.e;
      rect2.e = rect.b;
      pitm_b = new StripRegionSkip::Iterator<false>(region.srkb);
      rect.b = region.srb.rect0.b;
      rect.e = rect3.b;
      break;
    case BF_TYPE_2:
      if (dpRegion.isRegionType(CARET_CUT_L)) {
	Assert(region.isRegionNum(REGION_BM));
	rect1.b = rect.b;
	rect2.b = rect.e;
	pitm = new StripRegionSkip::Iterator<true>(region.srkm);
	rect.b = rect3.e;
	rect.e = region.srm.rect0.e;
      } else if (dpRegion.isRegionType(CARET_CUT_R)) {
	Assert(region.isRegionNum(REGION_ME));
	rect1.e = rect.e;
	rect2.e = rect.b;
	pitm = new StripRegionSkip::Iterator<true>(region.srkm);
	rect.b = region.srm.rect0.b;
	rect.e = rect3.b;
      } else {
	Die("bad dpRegion.type=%d", dpRegion.type);
      }
      break;
    default:
      Die("bad bfType=%d", bfType);
      break;
    }
  }
  if (pitm) {
    pitm->reset(rect);
  } else if (pitm_b) {
    pitm_b->reset(rect);
  }
}
template <bool inside0>
bool
DPDeckPSCFG::DPRegion::Region::BFIterator<inside0>::
getNext()
{
  if (pitm_b) {
    if (!pitm_b->getNext()) return false;

    mpt = pitm_b->pt;
  } else if (pitm) {
    if (!pitm->getNext()) return false;

    mpt = pitm->pt;
  }

  if (inside0) {
    rect1.e = mpt;
    rect2.b = mpt;
  } else {//outsideType
    switch (bfType) {
    case BF_TYPE_0:
      rect1.e = mpt;
      rect2.e = mpt;
      break;
    case BF_TYPE_1:
      rect1.b = mpt;
      rect2.b = mpt;
      break;
    case BF_TYPE_2:
      if (dpRegion.isRegionType(CARET_CUT_L)) {
	rect1.e = mpt;
	rect2.e = mpt;
      } else if (dpRegion.isRegionType(CARET_CUT_R)) {
	rect1.b = mpt;
	rect2.b = mpt;
      }
      break;
    default:
      Die("bad bfType=%d", bfType);
      break;
    }
  }
  return true;
}


  

#endif
