/*
 * DPDeckPSCFGDPTask.hpp
 *
 */
#ifndef DP_DECK_PSCFG_DP_TASK_HPP
#define DP_DECK_PSCFG_DP_TASK_HPP

#include "DPDeckPSCFG.hpp"
#include "PSCFGEnergyScoreManager.hpp"
// #include "PSCFGBaseScoreManager.hpp"
#include "DPDeckPSCFGDPRegion.hpp"
#include "DPDeckPSCFGTBTree.hpp"

class DPDeckPSCFG::DPTask {
public:
  enum DPType {
    CYK             = 0,
    SMALL_INSIDE    = 1,
    SMALL_OUTSIDE   = 2,
    INSIDE          = 3,
    OUTSIDE         = 4,
    LIKELIHOOD      = 5,
    SMALL_CYK       = 6,
    INVALID_DP_TYPE = 7
  };
  enum RType {
    TRIANGLE     = DPRegion::TRIANGLE,  
    CARET        = DPRegion::CARET,
    NOSE         = DPRegion::NOSE,
    TRIANGLE_CUT = DPRegion::TRIANGLE_CUT,
    CARET_CUT_L  = DPRegion::CARET_CUT_L,
    CARET_CUT_R  = DPRegion::CARET_CUT_R,
    INVALID_R_TYPE = DPRegion::INVALID_R_TYPE
  };
  enum RShape {
    TRI_SHAPE       = DPRegion::TRI_SHAPE,
    RECT_VERT_SHAPE = DPRegion::RECT_VERT_SHAPE,
    RECT_DIAG_SHAPE = DPRegion::RECT_DIAG_SHAPE
  };
  enum RNum {
    REGION_EE    = DPRegion::REGION_EE,
    REGION_BB    = DPRegion::REGION_BB, 
    REGION_BE    = DPRegion::REGION_BE,
    REGION_ME    = DPRegion::REGION_ME, 
    REGION_BM    = DPRegion::REGION_BM, 
    REGION_MM    = DPRegion::REGION_MM
  };
  static const char* type2str(int dpType0);

  const Model& model;
  DPDeckPSCFG& dpDeck;
  ScoreManager& scoreManager;
  VTB& vtb;
  VTBBF& vtbbf;
  VSFBWB& vsfbwB;
  VSFBWE& vsfbwE;
  VFBW& vfw;
  VFBW& vbw;

  DPTask* parent;
  int num_unprocessed_children;
  

  DPType dpType;
  DPRegion dpRegion;
  TBTree::Node treeNode;
  Vector<State> states;
  VSTB& vstb;
  VSTE& vste;
  DPScoreT finalScore;
  State finalState;
  Rect split_rect;
  Point split_mpt;
  SplitType splitType;

  DPTask(DPDeckPSCFG& dpDeck,
	 DPTask* parent,
	 const DPRegion& dpRegion,
	 const Vector<State>& states);
  DPTask(const DPTask& dpTask);
  ~DPTask();
  string toString() const;
  void print() const;

  bool isDPType(int dpType0) const;
  Vector<TBTree::Link> getTreeNodes();
  Vector<DPTask*> cyk_small();
  Vector<TBTree::Link> cyk();
  DPScoreT inside();
  void outside();
  DPScoreT likelihood();
  float computeMemorySizeInMB(const int& dpType) const;
  float computeCYKMemorySizeInMB() const;
  float computeSmallCYKMemorySizeInMB() const;
  float computeInsideOutsideMemorySizeInMB() const;
  float computeLikelihoodMemorySizeInMB() const;
  Vector<State> getStates() const {return states;}
  void setParent(DPTask*const& parent0) {parent = parent0;}
  Vector<TBTree::Link> traceback(const Rect& rect0, 
				 const State& state0) const;
  int getChild(Rect& rect, State& state) const;
  Point getBifurcationPoint(const Rect& rect) const;
  Point getBifurcationPoint(const DPRegion::Region& region,
			    const Rect& rect, const int& bfType,
			    const int& areaSkip,
			    Rect& rect1, Rect& rect2) const;
  bool isLeftChild(const Rect& rect, const State& state) const;
  bool isRightChild(const Rect& rect, const State& state) const;
  void testConsistency() const;
#ifdef SLR_STATE
  void loadVSTB(const DPRegion::Region& region, const State& state);
  void loadVSTE(const DPRegion::Region& region, const State& state);
#else
  void loadVSTB(const DPRegion::Region& region);
  void loadVSTE(const DPRegion::Region& region);
#endif
  
#ifdef SLR_STATE
  const DPScoreT& vstb_ref(int regionNum,
			   int bx, int by, int ex, int ey, State s) const;
  const DPScoreT& vste_ref(int regionNum,
			   int bx, int by, int ex, int ey, State s) const;
  const DPScoreT& vstb_ref(int bx, int by, int ex, int ey, State s) const;
  const DPScoreT& vste_ref(int bx, int by, int ex, int ey, State s) const;
#else
  const DPScoreT& vstb_ref(int regionNum,
			   int bx, int by, int ex, int ey) const;
  const DPScoreT& vste_ref(int regionNum,
			   int bx, int by, int ex, int ey) const;
  const DPScoreT& vstb_ref(int bx, int by, int ex, int ey) const;
  const DPScoreT& vste_ref(int bx, int by, int ex, int ey) const;
#endif
  const TBNode& vtb_ref(int regionNum,
			int bx, int by, int ex, int ey) const;
  const TBBFNode& vtbbf_ref(int regionNum,
			    int bx, int by, int ex, int ey) const;
  const ScoreManager::SCNode& vsfbwB_refI(int regionNum,
					  int by, int ex, int ey) const;
  const ScoreManager::SCNode& vsfbwB_refO(int regionNum,
					  int by, int ex, int ey) const;
  const ScoreManager::SCNode& vsfbwE_refI(int regionNum,
					  int bx, int by, int ey) const;
  const ScoreManager::SCNode& vsfbwE_refO(int regionNum,
					  int bx, int by, int ey) const;

  float computeDPIterationCountVTB() const;
  float computeDPIterationCountVST() const;

private:
  typedef DPRegion::Region Region;
  template <DPType dpType0, RType regionType0, RNum regionNum0>
  class SubTask;
  enum {
    LEFT_CHILD,
    RIGHT_CHILD,
    END_STATE,
    BIFURCATION,
    GENERIC_STATE
  }; 

  DPScoreT findBestSplitPoint(Rect& rect, State& state,
			      Point& mpt, SplitType& splitType) const;
  void findBestSplitPointRect(const int& regionO,
			      const int& regionI,
			      const bool& emitB,
			      const SplitType& splitType1,
			      Rect& best_rect,
			      SplitType& best_splitType,
			      DPScoreT& best_score,
			      State& best_state) const;
  void findBestBifurcatePoint(Rect& best_rect, Point& best_ptm,
			      DPScoreT& best_score) const;
  Vector<DPTask*> split(const Rect& rect, const State& state,
			const Point& ptm, const SplitType& splitType);

  void cyk_dp();
  DPScoreT getScore() const;
  State getFinalState() const;
};

#endif
