/*
 * DPDeckPSCFGDPTaskSubTask.hpp
 *
 */
#ifndef DP_DECK_PSCFG_DP_TASK_SUB_TASK_CPP
#define DP_DECK_PSCFG_DP_TASK_SUB_TASK_CPP
#include "DPDeckPSCFGDPTaskSubTask.hpp"
#include "DPDeckPSCFGVectors.hpp"
#include "DPDeckPSCFGDPLocals.cpp"

#define TEMPLATE                                    \
template <DPDeckPSCFG::DPTask::DPType dpType0,      \
	  DPDeckPSCFG::DPTask::RType regionType0,   \
	  DPDeckPSCFG::DPTask::RNum regionNum0>
#define CLASS \
  DPDeckPSCFG::DPTask::SubTask<dpType0,regionType0,regionNum0>

TEMPLATE
bool
CLASS::
isDPType(int dpt1)
{
  return (dpType == dpt1);
}
TEMPLATE
bool
CLASS::
isRegionType(int rt1)
{
  return (regionType == rt1);
}
TEMPLATE
bool
CLASS::
isRegionNum(int rn1)
{
  return (regionNum == rn1);
}
TEMPLATE
bool
CLASS::
isRegionTypeNum(int rt1, int rn1)
{
  return (isRegionType(rt1) && isRegionNum(rn1));
}
TEMPLATE
bool
CLASS::
isTriShape()
{
  return (isRegionNum(REGION_EE) 
	  || isRegionNum(REGION_BB)
	  || isRegionNum(REGION_MM));
}
TEMPLATE
bool
CLASS::
isRectShape()
{
  return !isTriShape();
}
TEMPLATE
bool
CLASS::
isRectVertShape()
{
  return (isRegionTypeNum(TRIANGLE_CUT, REGION_BE)
	  || isRegionTypeNum(CARET_CUT_L, REGION_BM)
	  || isRegionTypeNum(CARET_CUT_R, REGION_ME));
}
TEMPLATE
bool
CLASS::
isRectDiagShape()
{
  return (isRectShape() && !isRectVertShape());
}
TEMPLATE
bool
CLASS::
storeB()
{
  if (!isSmallCYK()) return false;

  switch (regionType) {
  case TRIANGLE_CUT:
    switch (regionNum) {
    case REGION_EE: return true;
    case REGION_BE: return true;
    default:        return false;
    }
    break;
  case CARET_CUT_L:
    switch (regionNum) {
    case REGION_BE: return true;
    case REGION_ME: return true;
    case REGION_BM: return true;
    case REGION_MM: return true;
    default:        return false;
    }
    break;
  case CARET_CUT_R:
    switch (regionNum) {
    case REGION_EE: return true;
    case REGION_ME: return true;
    default:        return false;
    }
    break;
  default: return false;
  }
  return false;
}
TEMPLATE
bool
CLASS::
storeE()
{
  if (!isSmallCYK()) return false;

  switch (regionType) {
  case TRIANGLE_CUT:
    switch (regionNum) {
    case REGION_BB: return true;
    case REGION_BE: return true;
    default:        return false;
    }
    break;
  case CARET_CUT_L:
    switch (regionNum) {
    case REGION_BB: return true;
    case REGION_BM: return true;
    default:        return false;
    }
    break;
  case CARET_CUT_R:
    switch (regionNum) {
    case REGION_BE: return true;
    case REGION_ME: return true;
    case REGION_BM: return true;
    case REGION_MM: return true;
    default:        return false;
    }
    break;
  default: return false;
  }
  return false;
}
TEMPLATE
bool
CLASS::
isInsideType()
{
  return (!isDPType(OUTSIDE) && !isDPType(SMALL_OUTSIDE));
}
TEMPLATE
bool
CLASS::
isOutsideType()
{
  return !isInsideType();
}
TEMPLATE
bool
CLASS::
takesMaxScore()
{
  return (isDPType(CYK)
	  || isDPType(SMALL_INSIDE)
	  || isDPType(SMALL_OUTSIDE));
}
TEMPLATE
bool
CLASS::
isSmallCYK()
{
  return (isDPType(SMALL_INSIDE) || isDPType(SMALL_OUTSIDE));
}
TEMPLATE
bool
CLASS::
isCYK()
{
  return isDPType(CYK);
}
TEMPLATE
bool
CLASS::
isInside()
{
  return isDPType(INSIDE);
}
TEMPLATE
bool
CLASS::
isOutside()
{
  return isDPType(OUTSIDE);
}
TEMPLATE
bool
CLASS::
isLikelihood()
{
  return isDPType(LIKELIHOOD);
}
TEMPLATE
bool
CLASS::
isTriIn()
{
  return (isTriShape() && isInsideType());
}
TEMPLATE
bool
CLASS::
isTriOut()
{
  return (isTriShape() && !isInsideType());
}
TEMPLATE
bool
CLASS::
isRectIn()
{
  return (isRectShape() && isInsideType());
}
TEMPLATE
bool
CLASS::
isRectVertIn()
{
  return (isRectVertShape() && isInsideType());
}
TEMPLATE
bool
CLASS::
isRectDiagIn()
{
  return (isRectDiagShape() && isInsideType());
}
TEMPLATE
bool
CLASS::
isRectOut()
{
  return (isRectShape() && !isInsideType());
}
TEMPLATE
bool
CLASS::
isRectVertOut()
{
  return (isRectVertShape() && !isInsideType());
}
TEMPLATE
bool
CLASS::
isRectDiagOut()
{
  return (isRectDiagShape() && !isInsideType());
}
TEMPLATE
bool
CLASS::
hasBFType1()
{
  return (!isTriIn());
}
TEMPLATE
bool
CLASS::
hasBFType2()
{
  if (isInsideType()) {
    if (isRegionTypeNum(NOSE, REGION_BE)) {
      return true;
    }
  } else {// outside type
    if (isRegionTypeNum(CARET_CUT_L, REGION_BM)
	|| isRegionTypeNum(CARET_CUT_R, REGION_ME)) {
      return true;
    }
  }
  return false; 
}
TEMPLATE
CLASS::
SubTask(DPTask& dpTask)
 : dpDeck(dpTask.dpDeck),
   scoreManager(dpDeck.scoreManager),
   dpPath(dpDeck.dpPath),
   vt(dpDeck.vt),
   vx(dpDeck.vx),
   vy(dpDeck.vy),
   vpairX(dpDeck.vpairX),
   vpairY(dpDeck.vpairY),
   vbe(dpDeck.vbe),
   vsc(dpDeck.vsc),
   vskip(dpDeck.vskip),
   vtb(dpDeck.vtb),
   vtbbf(dpDeck.vtbbf),
   vsfbwB(dpDeck.vsfbwB),
   vsfbwE(dpDeck.vsfbwE),
   vfw(dpDeck.vfw),
   vbw(dpDeck.vbw),
   dpTask(dpTask),
   vstb(dpTask.vstb),
   vste(dpTask.vste),
   dpRegion(dpTask.dpRegion),
   skipSet(dpRegion.getSkipSet()),
   region(dpTask.dpRegion.getSubRegion(regionNum)),
   srb(region.srb),
   sre(region.sre),
   srm(region.srm),
   srkb(region.srkb),
   srke(region.srke),
   srkm(region.srkm),
   rectb(srb.rect0),
   recte(sre.rect0),
   rectm(sre.rect0),
   rectkb(),
   rectke(),
   rectkm()
{
  rectkb = srkb.getRectK();
  rectke = srke.getRectK();
  rectkm = srkm.getRectK();

  run();
}
TEMPLATE
CLASS::
SubTask(DPTask& dpTask, DPScoreT& finalScore0, State& finalState0)
 : dpDeck(dpTask.dpDeck),
   scoreManager(dpDeck.scoreManager),
   dpPath(dpDeck.dpPath),
   vt(dpDeck.vt),
   vx(dpDeck.vx),
   vy(dpDeck.vy),
   vpairX(dpDeck.vpairX),
   vpairY(dpDeck.vpairY),
   vbe(dpDeck.vbe),
   vsc(dpDeck.vsc),
   vskip(dpDeck.vskip),
   vtb(dpDeck.vtb),
   vtbbf(dpDeck.vtbbf),
   vsfbwB(dpDeck.vsfbwB),
   vsfbwE(dpDeck.vsfbwE),
   vfw(dpDeck.vfw),
   vbw(dpDeck.vbw),
   dpTask(dpTask),
   vstb(dpTask.vstb),
   vste(dpTask.vste),
   dpRegion(dpTask.dpRegion),
   skipSet(dpRegion.getSkipSet()),
   region(dpTask.dpRegion.getSubRegion(regionNum)),
   srb(region.srb),
   sre(region.sre),
   srm(region.srm),
   srkb(region.srkb),
   srke(region.srke),
   srkm(region.srkm),
   rectb(srb.rect0),
   recte(sre.rect0),
   rectm(srm.rect0),
   rectkb(),
   rectke(),
   rectkm()
{
  rectkb = srkb.getRectK();
  rectke = srke.getRectK();
  rectkm = srkm.getRectK();
  run();
  finalScore0 = finalScore;
  finalState0 = finalState;
}
TEMPLATE
CLASS::
~SubTask()
{
}
TEMPLATE
string
CLASS::
toString() const
{
  ostringstream oss;
  oss << "dpType: " << DPTask::type2str(dpType) << "\n";
  oss << "regionType: " << DPRegion::type2str(regionType) << "\n";
  oss << "region: " << region.toString() << "\n";
  return oss.str();
}
TEMPLATE
void
CLASS::
print() const
{
  cout << toString() << flush;
}
TEMPLATE
void
CLASS::
run()
{
  if (onlyToLoadVST()) {
    loadVST();
  } else {
    initData();
    DPLocals<SubTask> dpl(*this);
    dpl.run();
    //DPLocals dpl(dpTask, region);
    //dpl.run(*this);

    finalScore = dpl.getFinalScore();
    finalState = dpl.getFinalState();

#ifdef SLR_STATE
    //if (insideType) {
    //  vstb.copy(region, vste, Model::SR_st);
    //  vste.copy(region, vstb, Model::SL_st);
    //}
#endif
  }
}
TEMPLATE
void
CLASS::
initData()
{
  resizeVectors();
  clean();
  dpAreaMargin = region.sre.getAreaMargin();
  initState0();
  initStateLim();
  initSCStore();
  initTau();
  initTZNodes();
  initBENodes();
  initSkip();
  // can use bxn_beg etc.
  initVTZ();
  initVXY();
  initVPair();
  initVBE();
  initVSC();
  initVSkip();
}
TEMPLATE
void
CLASS::
resizeVectors()
{
  vt.resize(region);
  vx.resize(region);
  vy.resize(region);
  vpairX.resize(region);
  vpairY.resize(region);
  vbe.resize(region);
  vsc.resize(region);
  vskip.resize(region);

#ifdef SLR_STATE
  if (isInsideType()) {
    vstb.resize(region, Model::SL_st);
    vste.resize(region, Model::SR_st);
  } else {
    vstb.resize(region, Model::BF_st);
    vste.resize(region, Model::BF_st);
  }
#else
  vstb.resize(region, Model::ST_st);
  vste.resize(region, Model::ST_st);
#endif

  if (isCYK()) {
    vtb.resize(region);
    vtbbf.resize(region);
  } else if (isSmallCYK()) {
    if (storeB()) {
      if (isInsideType()) {
	vsfbwB.resizeI(region);
      } else {
	vsfbwB.resizeO(region);
      }
    }
    if (storeE()) {
      if (isInsideType()) {
	vsfbwE.resizeI(region);
      } else {
	vsfbwE.resizeO(region);
      }
    }
  } else if (isInside()) {
    vfw.resize(region);
  } else if (isOutside()) {
    vbw.resize(region);
  }
}
TEMPLATE
void
CLASS::
initVTZ()
{
  const StripRegion& srb = region.srb;
  for (int i = 0; i < (int)srb.size(); i++) {
    const Trapezoid& t = srb[i];
    TZNode& tzn = vt.refB(i);
    tzn.type = t.type;
    tzn.topLeftMargin = (srb.getTopLeftMargin(i) + 1); // add 1 to move
    tzn.bottomRightMargin = (srb.getBottomRightMargin(i) + 1);
    tzn.bx = &vx.refB(t.bx);
    tzn.by = &vy.refB(t.by);
    tzn.ex = &vx.refB(t.ex);
    tzn.ey = &vy.refB(t.ey);
    tzn.bxey = &vy.refB(t.getBXEY());
    tzn.exby = &vy.refB(t.getEXBY());
  }
  if (isRectShape()) {
    const StripRegion& sre = region.sre;
    for (int i = 0; i < (int)sre.size(); i++) {
      const Trapezoid& t = sre[i];
      TZNode& tzn = vt.refE(i);
      tzn.type = t.type;
      tzn.topLeftMargin = (sre.getTopLeftMargin(i) + 1);// add 1 to move
      tzn.bottomRightMargin = (sre.getBottomRightMargin(i) + 1);
      tzn.bx = &vx.refE(t.bx);
      tzn.by = &vy.refE(t.by);
      tzn.ex = &vx.refE(t.ex);
      tzn.ey = &vy.refE(t.ey);
      tzn.bxey = &vy.refE(t.getBXEY());
      tzn.exby = &vy.refE(t.getEXBY());
    }
  }
}
TEMPLATE
void
CLASS::
initVXY()
{// set XYNode::null() when resize();
  for (int x = srb.bx0; x <= srb.ex0; x++) {
    XNode& xn = vx.refB(x);
    xn.xCol = scoreManager.getXCol(x);
  }
  if (isRectShape()) {
    for (int x = sre.bx0; x <= sre.ex0; x++) {
      XNode& xn = vx.refE(x);
      xn.xCol = scoreManager.getXCol(x);
    }
  }
  for (int y = srb.by0; y <= srb.ey0; y++) {
    YNode& yn = vy.refB(y);
    yn.yCol = scoreManager.getYCol(y);
  }
  if (isRectShape()) {
    for (int y = max((srb.ey0 + 1), sre.by0); y <= sre.ey0; y++) {
      YNode& yn = vy.refE(y);
      yn.yCol = scoreManager.getYCol(y);
    }
  }
  for (int x = srb.bx0; x <= (srb.ex0 + 1); x++) {// +1 <- for prev_pairX
    XNode& xn = vx.refB(x);
    if (isTriIn()) {
      xn.pn_beg = &vpairX.ref(x, x - 1); // access (x, x-1)
    } else if (isRectIn()) {
      xn.pn_beg = &vpairX.ref(x, sre.bx0 - 1);
    } else {// outside
      xn.pn_beg = &vpairX.ref(x - 1, sre.ex0); // access(-1,x)
    }
  }
  for (int y = srb.by0; y <= (srb.ey0 + 1); y++) {// +1 <- for prev_pairY
    YNode& yn = vy.refB(y);
    if (isTriIn()) {
      yn.pn_beg = &vpairY.ref(y, y - 1); // access (y, y-1)
    } else if (isRectIn()) {
      yn.pn_beg = &vpairY.ref(y, max(y, sre.by0) - 1);
    }  else {// outside
      yn.pn_beg = &vpairY.ref(y - 1, sre.ey0); // access(-1,y)
    }
  }
  for (int y = srb.by0; y <= sre.ey0; y++) {
    if (srb.ey0 < y && y < sre.by0) continue;

    YNode& yn = (y <= srb.ey0 ? vy.refB(y) : vy.refE(y));
    yn.bxn = region.getBXN(y);
    yn.exn = region.getEXN(y);
    if (yn.bxn < 0) continue;

    if (insideType) {// access ex0+1
      Point pt = {(-1), y};
      int shift = (-1);
      if (yn.exn <= region.bxn_last) {
	pt.x = region.convertXN2XK(yn.exn);
	shift = 1;
      } else if (region.bxn_last < yn.bxn) {
	pt.x = region.convertXN2XK(yn.bxn);
	shift = 0;
      } else {
	pt.x = region.convertXN2XK(region.bxn_last);
	shift = 1;
      }
      yn.pSkipCountB = (&vskip.skipCountB_ref(region, pt) + shift);
      yn.pSkipShiftB = (&vskip.skipShiftB_ref(region, pt) + shift);
      yn.pSkipCountE = (&vskip.skipCountE_ref(region, pt) + shift);
      yn.pSkipShiftE = (&vskip.skipShiftE_ref(region, pt) + shift);
    } else {//outsideType
      Point pt_b = {(-1), y};
      Point pt_e = {(-1), y};
      int shift_b = (-1);
      int shift_e = (-1);

      pt_b.x = region.convertXN2XK(yn.bxn);
      shift_b = (-1);
      yn.pSkipCountB = (&vskip.skipCountB_ref(region, pt_b) + shift_b);
      yn.pSkipShiftB = (&vskip.skipShiftB_ref(region, pt_b) + shift_b);

      pt_e.x = region.convertXN2XK(yn.exn);
      shift_e = 1;
      yn.pSkipCountE = (&vskip.skipCountE_ref(region, pt_e) + shift_e);
      yn.pSkipShiftE = (&vskip.skipShiftE_ref(region, pt_e) + shift_e);
      yn.pSkipCountE0 = yn.pSkipCountE;
    }
  }

  for (int y = srb.by0; y <= sre.ey0; y++) {
    if (srb.ey0 < y && y < sre.by0) continue;

    YNode& yn = (y <= srb.ey0 ? vy.refB(y) : vy.refE(y));
    if (y < sre.by0) {
      const int& yex = srb.getEX(y);
      Assert(yex <= srb.ex0);
      yn.ex = &vx.refB(yex);
      yn.ext = (&vt.getHeadB() + srb.getEXTIndex(y));
    } else if (y <= srb.ey0) {
      const int& yex = sre.getEX(y);
      if (yex <= srb.ex0) {
	yn.ex = &vx.refB(yex);
	yn.ext = (&vt.getHeadB() + srb.getEXTIndex(y));
      } else {
	yn.ex = &vx.refE(yex);
	yn.ext = (&vt.getHeadE() + sre.getEXTIndex(y));
      }
    }
  }
}
TEMPLATE
void
CLASS::
initVPair()
{
  for (int bx = srb.bx0; bx <= srb.ex0; bx++) {
    for (int ex = max((bx+1), (sre.bx0-1)); ex <= (sre.ex0-1); ex++) {
      PairNode& pn = vpairX.ref(bx, ex);
      pn.pairCol = scoreManager.getPairColX(bx, ex);
    }
  }
  for (int by = srb.by0; by <= srb.ey0; by++) {
    for (int ey = max((by+1), (sre.by0-1)); ey <= (sre.ey0-1); ey++) {
      PairNode& pn = vpairY.ref(by, ey);
      pn.pairCol = scoreManager.getPairColY(by, ey);
    }
  }
}
TEMPLATE
void
CLASS::
initVSC()
{// no-op
}
TEMPLATE
void
CLASS::
initVSkip()
{
  StripRegionSkip::RowIterator<true> itrb(srkb);
  StripRegionSkip::RowIterator<true> itre(srke);
  StripRegionSkip::RowIterator<true> itrm(srkm);

  while (itrb.getNext()) {
    vskip.skipCountB0_ref(region, itrb.x) = itrb.ly;
  }
  if (!region.isTriShape()) {
    while (itre.getNext()) {
      vskip.skipCountE0_ref(region, itre.x) = itre.ly;
    }
  }
  while (itrm.getNext()) {
    vskip.skipCountM0_ref(region, itrm.x) = itrm.ly;
  }
  
  for (int by = region.srb.by0; by <= region.srb.ey0; by++) {
    const int& bxk_beg = region.srkb.getFirstSkipRow(by);
    const int& bxk_last = region.srkb.getLastSkipRow(by);
    if (bxk_beg < 0) continue;

    Rect bbox = {
      {{bxk_beg, region.srb.by0}},
      {{bxk_last, region.srb.ey0}}
    };
    itrb.reset(bbox);
    while (itrb.getNext()) {
      const Point pt = {itrb.x, by};
      const int& y_beg = itrb.y_beg;
      const int& y_last = itrb.y_last;
      const int& ly = itrb.ly;
      if (ly == 0) continue;

      const int& lyL = skipSet.getNSkipPointY(pt.x, y_beg, pt.y - 1);
      const int& lyR = skipSet.getNSkipPointY(pt.x, pt.y + 1, y_last);
      const int& lyB = (insideType ? lyL : lyR);
      const int& lyE = (insideType ? lyR : lyL);

      vskip.skipCountB_ref(region, pt) = (ly - lyB);
      vskip.skipShiftB_ref(region, pt) = lyB;
      vskip.skipCountE_ref(region, pt) = (ly - lyE);
      vskip.skipShiftE_ref(region, pt) = lyE;
    }
  }
  if (!region.isTriShape()) {
    for (int ey = region.sre.by0; ey <= region.sre.ey0; ey++) {
      const int& exk_beg = region.srke.getFirstSkipRow(ey);
      const int& exk_last = region.srke.getLastSkipRow(ey);
      if (exk_beg < 0) continue;

      Rect bbox = {
	{{exk_beg, region.sre.by0}},
	{{exk_last, region.sre.ey0}}
      };
      itre.reset(bbox);
      while (itre.getNext()) {
	const Point pt = {itre.x, ey};
	const int& y_beg = itre.y_beg;
	const int& y_last = itre.y_last;
	const int& ly = itre.ly;
	if (ly == 0) continue;

	const int& lyL = skipSet.getNSkipPointY(pt.x, y_beg, pt.y - 1);
	const int& lyR = skipSet.getNSkipPointY(pt.x, pt.y + 1, y_last);
	const int& lyB = (insideType ? lyL : lyR);
	const int& lyE = (insideType ? lyR : lyL);

	vskip.skipCountB_ref(region, pt) = (ly - lyB);
	vskip.skipShiftB_ref(region, pt) = lyB;
	vskip.skipCountE_ref(region, pt) = (ly - lyE);
	vskip.skipShiftE_ref(region, pt) = lyE;
      }
    }
  }
}
TEMPLATE
void
CLASS::
initVBE()
{
  initVBE_SC();
  initVBE_ESC();

  if (isTriIn()) {
    initVBE_TriIn();
  } else if (isRectIn()) {
    initVBE_RectIn(getRegionTriB(), getRegionTriE());
    if (hasBFType2()) {
      initVBE_RectIn_M(getRegionRectB(), getRegionRectE());
    }
  } else if (isTriOut()) {
    initVBE_TriOut();
  } else if (isRectOut()) {
    initVBE_RectOut(getRegionTriB(), getRegionTriE());
    if (hasBFType2()) {
      if (isRegionType(CARET_CUT_L)) {
	initVBE_RectOut_L(getRegionRectO(), getRegionRectI());
      } else if (isRegionType(CARET_CUT_R)) {
	initVBE_RectOut_R(getRegionRectO(), getRegionRectI());
      } else {
	Die("bad regionType");
      }
    }
  } else {
    Die("bad regionType");
  }
}
TEMPLATE
void
CLASS::
initVBE_SC()
{
  StripRegion::Iterator<(!insideType)> itb(srb);
  Point e0;
  while (itb.getNext()) {
    const int& diag = (itb.pt.y - itb.pt.x);
    if (isTriIn()) {
      e0 = itb.pt;
    } else if (isRectIn()) {
      e0.setVal(sre.bx0, max(itb.pt.y, sre.by0));
    } else {// outside
      e0 = sre.e0;
    }
    vbe.ref(region, itb.pt).sc = &vsc.ref(region, diag, e0);
  }
}
TEMPLATE
void
CLASS::
initVBE_ESC()
{
  StripRegion::Iterator<true> itb(srb);
  while (itb.getNext()) {
    Point pt = itb.pt;
    BENode& bn = vbe.ref(region, pt);
    
    if (insideType) {
      if (pt.x < srb.ex0 && pt.y < srb.ey0) {
	bn.mColB = scoreManager.getMatchCol(pt);
      }
    } else {// outside
      Point pt1 = {pt.x - 1, pt.y - 1};
      if (srb.includes(pt1)) {
	bn.mColB = scoreManager.getMatchCol(pt1);
      }
    }
  }
  StripRegion::Iterator<true> ite(sre);
  while (ite.getNext()) {
    Point pt = ite.pt;
    BENode& en = vbe.ref(region, pt);
    if (insideType) {
      Point pt1 = {pt.x - 1, pt.y - 1};
      if (sre.includes(pt1)) {
	en.mColE = scoreManager.getMatchCol(pt1);
      }
    } else { // outside
      if (pt.x < sre.ex0 && pt.y < sre.ey0) {
	en.mColE = scoreManager.getMatchCol(pt);
      }
    }
  }
}
TEMPLATE
void
CLASS::
initVBE_TriIn()
{
  StripRegionSkip::Iterator<true> it(region.srke);
  Rect rect;
  while (it.getNext()) {
    BENode& ben = vbe.ref(region, it.pt);
    rect.setVal(it.pt, it.pt);
#ifdef SLR_STATE
    ben.st[0] = &vste.ref(region, rect, Model::SR_st);//en store, vary
    ben.st[1] = &vstb.ref(region, rect, Model::SL_st);//bn store
    ben.st[2] = &vstb.ref(region, rect, Model::SL_st);//start position for rect1
#else
    ben.st[0] = &vste.ref(region, rect);//en store, vary
    ben.st[1] = &vstb.ref(region, rect);//bn store
    ben.st[2] = &vstb.ref(region, rect);//start position for rect1
#endif
  }
}
TEMPLATE
void
CLASS::
initVBE_RectIn(const Region& regionTriB, const Region& regionTriE)
{
  Rect rect;
  Rect bbox;

  StripRegionSkip::Iterator<true> itb(region.srkb);
  StripRegionSkip::Iterator<true> ite(region.srke);

#ifdef SLR_STATE
  dpTask.loadVSTB(regionTriB, Model::SL_st);
#else
  dpTask.loadVSTB(regionTriB);
#endif

  bbox.e = region.sre.e0;
  while (itb.getNext()) {
    BENode& bn = vbe.ref(region, itb.pt);
#ifdef SLR_STATE
    bn.st[0] = &vstb.ref(regionTriB, rect.setVal(itb.pt, itb.pt), Model::SL_st);
#else
    bn.st[0] = &vstb.ref(regionTriB, rect.setVal(itb.pt, itb.pt));
#endif

    bbox.b = itb.pt;
    ite.reset(bbox);
    if (ite.getNext()) {
#ifdef SLR_STATE
      bn.st[1] = &vstb.ref(region, rect.setVal(itb.pt, ite.pt), Model::SL_st);//store
#else
      bn.st[1] = &vstb.ref(region, rect.setVal(itb.pt, ite.pt));//store
#endif
    }
  }

#ifdef SLR_STATE
  dpTask.loadVSTE(regionTriE, Model::SR_st);
#else
  dpTask.loadVSTE(regionTriE);
#endif

  StripRegionSkip::Iterator<false> itb1(region.srkb);
  ite.reset(region.sre.rect0);
  bbox.b = region.srb.b0;
  while (ite.getNext()) {
    BENode& en = vbe.ref(region, ite.pt);

    bbox.e = ite.pt;
    itb1.reset(bbox);
    if (itb1.getNext()) {
#ifdef SLR_STATE
      en.st[0] = &vste.ref(region, rect.setVal(itb1.pt, ite.pt), Model::SR_st);//store
#else
      en.st[0] = &vste.ref(region, rect.setVal(itb1.pt, ite.pt));//store
#endif
    }
#ifdef SLR_STATE
    en.st[1] = &vste.ref(regionTriE, rect.setVal(ite.pt, ite.pt), Model::SR_st);
#else
    en.st[1] = &vste.ref(regionTriE, rect.setVal(ite.pt, ite.pt));
#endif
  }
}
TEMPLATE
void
CLASS::
initVBE_RectIn_M(const Region& regionRectB, const Region& regionRectE)
{
  const Rect& rectkm = regionRectB.srke.getRectK();
  if (!rectkm.isNormal()) return;

#ifdef SLR_STATE
  dpTask.loadVSTB(regionRectB, Model::SL_st);
#else
  dpTask.loadVSTB(regionRectB);
#endif

  Rect rect;
  StripRegionSkip::Iterator<true> itb(region.srkb);
  while (itb.getNext()) {
    BENode& bn = vbe.ref(region, itb.pt);
#ifdef SLR_STATE
    bn.st[2] = &vstb.ref(regionRectB, rect.setVal(itb.pt, rectkm.b), Model::SL_st);
#else
    bn.st[2] = &vstb.ref(regionRectB, rect.setVal(itb.pt, rectkm.b));
#endif
  }

#ifdef SLR_STATE
  dpTask.loadVSTE(regionRectE, Model::SR_st);
#else
  dpTask.loadVSTE(regionRectE);
#endif
  StripRegionSkip::Iterator<true> ite(region.srke);
  while (ite.getNext()) {
    BENode& en = vbe.ref(region, ite.pt);
#ifdef SLR_STATE
    en.st[2] = &vste.ref(regionRectE, rect.setVal(rectkm.b, ite.pt), Model::SR_st);
#else
    en.st[2] = &vste.ref(regionRectE, rect.setVal(rectkm.b, ite.pt));
#endif
  }
}
TEMPLATE
void
CLASS::
initVBE_TriOut()
{
  Assert(dpTask.parent);
  Rect rect;
  Rect bbox;

  StripRegionSkip::Iterator<true> itb(region.srkb);
  StripRegionSkip::Iterator<true> ite(region.srke);
  StripRegionSkip::Iterator<false> ite1(region.srke);

#ifdef SLR_STATE
  dpTask.parent->loadVSTE(region, Model::SL_st);
#else
  dpTask.parent->loadVSTE(region);
#endif

  bbox.e = region.sre.e0;
  while (itb.getNext()) {
    const Point& pt = itb.pt;
    BENode& bn = vbe.ref(region, pt);
    bbox.b = pt;
    ite1.reset(bbox);
    if (ite1.getNext()) {
#ifdef SLR_STATE
      bn.st[0] = &vstb.ref(region, rect.setVal(pt, ite1.pt), Model::BF_st);//store
#else
      bn.st[0] = &vstb.ref(region, rect.setVal(pt, ite1.pt));//store
#endif
    }
#ifdef SLR_STATE
    bn.st[3] = &dpTask.parent->vste.ref(region,rect.setVal(pt, pt), Model::SL_st);
#else
    bn.st[3] = &dpTask.parent->vste.ref(region,rect.setVal(pt, pt));
#endif
  }

#ifdef SLR_STATE
  dpTask.parent->loadVSTB(region, Model::SR_st);
#else
  dpTask.parent->loadVSTB(region);
#endif
    
  ite.reset();
  bbox.b = region.srb.b0;
  while (ite.getNext()) {
    const Point& pt = ite.pt;
    BENode& en = vbe.ref(region, pt);
    bbox.e = pt;
    itb.reset(bbox);
    if (itb.getNext()) {
#ifdef SLR_STATE
      en.st[1] = &vste.ref(region, rect.setVal(itb.pt, pt), Model::BF_st);//store
#else
      en.st[1] = &vste.ref(region, rect.setVal(itb.pt, pt));//store
#endif
    }
#ifdef SLR_STATE
    en.st[2] = &dpTask.parent->vstb.ref(region, rect.setVal(pt, pt), Model::SR_st);
#else
    en.st[2] = &dpTask.parent->vstb.ref(region, rect.setVal(pt, pt));
#endif
  }
}
TEMPLATE
void
CLASS::
initVBE_RectOut(const Region& regionTriB, const Region& regionTriE)
{// TriOut and RectOut are similar <- merge
  Rect rect;
  Rect bbox;
  StripRegionSkip::Iterator<true> itb(region.srkb);
  StripRegionSkip::Iterator<true> ite(region.srke);
  StripRegionSkip::Iterator<false> ite1(region.srke);

#ifdef SLR_STATE
  dpTask.loadVSTE(regionTriB, Model::SL_st);
#else
  dpTask.loadVSTE(regionTriB);
#endif

  bbox.e = region.sre.e0;
  while (itb.getNext()) {
    BENode& bn = vbe.ref(region, itb.pt);
    bbox.b = itb.pt;
    ite1.reset(bbox);
    if (ite1.getNext()) {
#ifdef SLR_STATE
      bn.st[0] = &vstb.ref(region, rect.setVal(itb.pt, ite1.pt), Model::BF_st);//store
#else
      bn.st[0] = &vstb.ref(region, rect.setVal(itb.pt, ite1.pt));//store
#endif
    }
#ifdef SLR_STATE
    bn.st[1] = &vste.ref(regionTriB, rect.setVal(itb.pt, itb.pt), Model::SL_st);
#else
    bn.st[1] = &vste.ref(regionTriB, rect.setVal(itb.pt, itb.pt));
#endif
  }

#ifdef SLR_STATE
  dpTask.loadVSTB(regionTriE, Model::SR_st);
#else    
  dpTask.loadVSTB(regionTriE);
#endif

  bbox.b = region.srb.b0;
  ite.reset();
  while (ite.getNext()) {
    BENode& en = vbe.ref(region, ite.pt);
    bbox.e = ite.pt;
    itb.reset(bbox);
    if (itb.getNext()) {
#ifdef SLR_STATE
      en.st[1] = &vste.ref(region, rect.setVal(itb.pt, ite.pt), Model::BF_st);//store
#else
      en.st[1] = &vste.ref(region, rect.setVal(itb.pt, ite.pt));//store
#endif
    }
#ifdef SLR_STATE
    en.st[0] = &vstb.ref(regionTriE, rect.setVal(ite.pt, ite.pt), Model::SR_st);
#else    
    en.st[0] = &vstb.ref(regionTriE, rect.setVal(ite.pt, ite.pt));
#endif
  }
}
TEMPLATE
void
CLASS::
initVBE_RectOut_L(const Region& regionRectO, const Region& regionRectI)
{
  const Rect& rectkm = regionRectO.srke.getRectK();
  if (!rectkm.isNormal()) return;

#ifdef SLR_STATE
  dpTask.loadVSTB(regionRectO, Model::BF_st);
#else
  dpTask.loadVSTB(regionRectO);
#endif

  Rect rect;
  StripRegionSkip::Iterator<true> itb(region.srkb);
  while (itb.getNext()) {
    BENode& bn = vbe.ref(region, itb.pt);
#ifdef SLR_STATE
    bn.st[2] = &vstb.ref(regionRectO, rect.setVal(itb.pt, rectkm.b), Model::BF_st);
#else
    bn.st[2] = &vstb.ref(regionRectO, rect.setVal(itb.pt, rectkm.b));
#endif
  }

#ifdef SLR_STATE
  dpTask.loadVSTB(regionRectI, Model::SR_st);
#else
  dpTask.loadVSTB(regionRectI);
#endif

  StripRegionSkip::Iterator<true> ite(region.srke);
  while (ite.getNext()) {
    BENode& en = vbe.ref(region, ite.pt);
#ifdef SLR_STATE
    en.st[2] = &vstb.ref(regionRectI, rect.setVal(ite.pt, rectkm.b), Model::SR_st);
#else
    en.st[2] = &vstb.ref(regionRectI, rect.setVal(ite.pt, rectkm.b));
#endif
  }
}
TEMPLATE
void
CLASS::
initVBE_RectOut_R(const Region& regionRectO, const Region& regionRectI)
{
  Rect rectkm = regionRectO.srkb.getRectK();
  if (!rectkm.isNormal()) return;

#ifdef SLR_STATE
  dpTask.loadVSTE(regionRectI, Model::SL_st);
#else
  dpTask.loadVSTE(regionRectI);
#endif

  Rect rect;
  StripRegionSkip::Iterator<true> itb(region.srkb);
  while (itb.getNext()) {
    BENode& bn = vbe.ref(region, itb.pt);
#ifdef SLR_STATE
    bn.st[2] = &vste.ref(regionRectI, rect.setVal(rectkm.b, itb.pt), Model::SL_st);
#else
    bn.st[2] = &vste.ref(regionRectI, rect.setVal(rectkm.b, itb.pt));
#endif
  }

#ifdef SLR_STATE
  dpTask.loadVSTE(regionRectO, Model::BF_st);
#else
  dpTask.loadVSTE(regionRectO);
#endif

  StripRegionSkip::Iterator<true> ite(region.srke);
  while (ite.getNext()) {
    BENode& en = vbe.ref(region, ite.pt);
#ifdef SLR_STATE
    en.st[2] = &vste.ref(regionRectO, rect.setVal(rectkm.b, ite.pt), Model::BF_st);
#else
    en.st[2] = &vste.ref(regionRectO, rect.setVal(rectkm.b, ite.pt));
#endif
  }
}
TEMPLATE
void
CLASS::
clean()
{
  finalScore = NegInfDPScore;
  finalState = Model::X_st;

  dpAreaMargin = 0;

  state0    = Model::X_st;
  state_lim = Model::X_st;

  scb_store = NULL;
  sce_store = NULL;

  tau    = NULL;
  tau_BF = NULL;

  bt_lim = NULL;
  bt     = NULL;
  et0    = NULL;
  et_lim = NULL;

  bn  = NULL;
  en0 = NULL;

  bxn_beg  = (-1); 
  bxn_last = (-1);
  exn_beg  = (-1); 
  exn_last = (-1);
  mxn_beg  = (-1); 
  mxn_last = (-1);

  bxr0 = (-1);
  exr0 = (-1);

  skipSize = 0;
  pSkipCountB0 = NULL;
  pSkipCountE0 = NULL;
  pSkipCountM0 = NULL;
}
TEMPLATE
void
CLASS::
initState0()
{
  if (isOutsideType()) {
    state0 = ((isTriShape() || isRegionNum(REGION_BE))
	      ? dpTask.states[0] : Model::X_st);
  } else if (isTriIn()) {
    state0 = Model::X_st; 
  }
  if (isRectIn()) {
    if (isRegionType(CARET)
	|| isRegionType(CARET_CUT_L)
	|| isRegionType(CARET_CUT_R)) {
      state0 = dpTask.states[1];
    } else if (isRegionType(NOSE)) {
      switch (regionNum) {
      case REGION_BE:
	state0 = Model::X_st;
	break;
      case REGION_ME:
	state0 = dpTask.states[2];
	break;
      case REGION_BM:
	state0 = dpTask.states[1];
	break;
      default:
	Die("bad regionNum=%s",
	    DPRegion::regionNum2str(regionNum));
	break;
      }
    } else {
      Die("bad regionType = %s",
	  DPRegion::type2str(regionType));
    }
  }
}
TEMPLATE
void
CLASS::
initStateLim()
{// only consider limited situation
  state_lim = Model::X_st;
  if (isInsideType()) {
    if (isRegionTypeNum(TRIANGLE, REGION_EE)
	|| isRegionNum(REGION_BE)) {
      state_lim = dpTask.states[0];
    }
  }
}
TEMPLATE
void
CLASS::
initSCStore()
{
  if (isSmallCYK()) {
    // scb_store - storeB - store values at the last bx = bx_lim
    if (storeB()) {
      scb_store = (isInsideType()
		   ? &vsfbwB.head(region)
		   : &vsfbwB.tail(region));
    }
    // sce_store - storeE - store values at each ex = ex_lim
    if (storeE()) {
      sce_store = (isInsideType()
		   ? &vsfbwE.head(region)
		   : &vsfbwE.tail(region));
    }
  } else if (isInside()) {
    sce_store = &vfw.head(region);
  } else if (isOutside()) {
    sce_store = &vbw.tail(region);
  }
}
TEMPLATE
void
CLASS::
initTau()
{
  if (isCYK()) {
    tau = &vtb.head(region);
    tau_BF = &vtbbf.head(region);
  }
}
TEMPLATE
void
CLASS::
initTZNodes()
{
  bt_lim = (insideType ? &vt.getHeadB() : &vt.getTailB());
  bt = (insideType ? &vt.getTailB() : &vt.getHeadB());
  et0 = (insideType ? &vt.getHeadE() : &vt.getTailE());
  et_lim = (insideType ? &vt.getTailE() : &vt.getHeadE());
}
TEMPLATE
void
CLASS::
initBENodes()
{
  bn = (insideType ? &vbe.getTailB() : &vbe.getHeadB());
  en0 = (insideType ? &vbe.getHeadE() : &vbe.getTailE());
}
TEMPLATE
void
CLASS::
initSkip()
{
  bxn_beg = region.bxn_beg;
  bxn_last = region.bxn_last;
  exn_beg = region.exn_beg;
  exn_last = region.exn_last;
  mxn_beg = region.mxn_beg;
  mxn_last = region.mxn_last;

  if (insideType) {
    if (bxn_beg >= 0) {
      bxr0 = (region.srb.ex0 - region.bxk_last + 1);
    }
    if (exn_beg >= 0) {
      if (isTriShape()) {
	exr0 = 1;
      } else {
	exr0 = (region.exk_beg - region.sre.bx0 + 1);
      }
    }
  } else {
    if (bxn_beg >= 0) {
      bxr0 = (region.bxk_beg - region.srb.bx0 + 1);
    }
    if (exn_beg >= 0) {
      exr0 = (region.sre.ex0 - region.exk_last + 1);
    }
  }

  skipSize = dpRegion.getSkipSize();
  if (bxn_beg >= 0) {
    pSkipCountB0 = (insideType
		    ? (&vskip.getTailB0() + 1)
		    : (&vskip.getHeadB0() - 1));
  }
  if (exn_beg >= 0) {
    pSkipCountE0 = (insideType
		    ? &vskip.getTailE0()
		    : (&vskip.getTailE0() + 1));
  }
  if (hasBFType2()) {
    if (mxn_beg >= 0) {
      pSkipCountM0 = &vskip.getHeadM0();
    }
  }
}
TEMPLATE
int
CLASS::
getSkipRowCountM0() const
{
  if (isInsideType()) {
    Assert(isRegionTypeNum(NOSE, REGION_BE));
    return dpRegion.getSubRegion(REGION_BM).srke.getNSkipRow();
  } else {// outside type
    if (isRegionTypeNum(CARET_CUT_L, REGION_BM)) {
      return dpRegion.getSubRegion(REGION_ME).srke.getNSkipRow();
    } else if (isRegionTypeNum(CARET_CUT_R, REGION_ME)) {
      return dpRegion.getSubRegion(REGION_BM).srkb.getNSkipRow();
    } else {
      Die("no bfType2");
    }
  }
  return 0;
}

TEMPLATE
bool
CLASS::
onlyToLoadVST() const
{
  if (toLoadVSTB() || toLoadVSTE()) return true;

  return false;
}
TEMPLATE
bool
CLASS::
toLoadVSTB() const
{
  if (!isSmallCYK()) return false;

  switch (regionType) {
  case CARET_CUT_L: return isRegionNum(REGION_EE);
  case CARET_CUT_R: return isRegionNum(REGION_BB);
  case NOSE:
    switch (regionNum) {
    case REGION_EE: return true;
    case REGION_BB: return true;
    case REGION_BE: return true;
    case REGION_MM: return true;
    case REGION_BM: return dpTask.parent->dpRegion.isType(CARET_CUT_R);//outside
    case REGION_ME: return dpTask.parent->dpRegion.isType(CARET_CUT_L);//outside
    default: return false;
    }
  default: return false;
  }
}
TEMPLATE
bool
CLASS::
toLoadVSTE() const
{
  if (!isSmallCYK()) return false;

  switch (regionType) {
  case CARET_CUT_L: return isRegionNum(REGION_EE);
  case CARET_CUT_R: return isRegionNum(REGION_BB);
  case NOSE:
    switch (regionNum) {
    case REGION_EE: return true;
    case REGION_BB: return true;
    case REGION_BE: return true;
    case REGION_MM: return true;
    case REGION_BM: return dpTask.parent->dpRegion.isType(CARET_CUT_R);
    case REGION_ME: return dpTask.parent->dpRegion.isType(CARET_CUT_L);
    default: return false;
    }
  default: return false;
  }
}
//TEMPLATE
//void
//CLASS::
//loadVST()
//{
//  if (region.getVSTSize() == 0) return;
//
//  Assert(dpTask.parent != NULL);
//  const DPTask& dpTask1 = *dpTask.parent;
//  const DPRegion& dpRegion1 = dpTask1.dpRegion;
//  //const Rect rect0 = {{region.srb.b0}, {region.sre.e0}};
//  const Region& region1 = dpRegion1.findRegion(region);
//  const VSTB& vstb1 = dpTask1.vstb;
//  const VSTE& vste1 = dpTask1.vste;
//
//  vstb.resize(region);
//  vste.resize(region);
//#ifdef SLR_STATE
//  vstb.copy(region, vstb1, region1, Model::SL_st);
//  vstb.copy(region, vstb1, region1, Model::SR_st);
//  vstb.copy(region, vstb1, region1, Model::BF_st);// needed for NOSE REGION_BE
//  vste.copy(region, vste1, region1, Model::SL_st);
//  vste.copy(region, vste1, region1, Model::SR_st);
//  vste.copy(region, vste1, region1, Model::BF_st);// this is unnecessary
//#else
//  vstb.copy(region, vstb1, region1);
//  vste.copy(region, vste1, region1);
//#endif
//}
TEMPLATE
void
CLASS::
loadVST()
{
#ifdef SLR_STATE
  if (isInsideType()) {
    dpTask.loadVSTB(region, Model::SL_st);
    dpTask.loadVSTE(region, Model::SR_st);
  } else {
    dpTask.loadVSTB(region, Model::BF_st);
    dpTask.loadVSTE(region, Model::BF_st);
  }
#else
  dpTask.loadVSTB(region);
  dpTask.loadVSTE(region);
#endif
}
TEMPLATE
const DPDeckPSCFG::DPRegion::Region&
CLASS::
getRegionTriE() const
{
  const struct {int regionNum, triE;} table[] = {
    {REGION_BE, REGION_EE},
    {REGION_BM, REGION_MM},
    {REGION_ME, REGION_EE}
  };
  const int n = (sizeof(table) / sizeof(table[0]));
  for (int i = 0; i < n; i++) {
    if (isRegionNum(table[i].regionNum)) {
      return dpRegion.getSubRegion(table[i].triE);
    }
  }
  Die("no triE");
  return region;
}
TEMPLATE
const DPDeckPSCFG::DPRegion::Region&
CLASS::
getRegionTriB() const
{
  const struct {int regionNum, triB;} table[] = {
    {REGION_BE, REGION_BB},
    {REGION_BM, REGION_BB},
    {REGION_ME, REGION_MM}
  };
  const int n = (sizeof(table) / sizeof(table[0]));
  for (int i = 0; i < n; i++) {
    if (isRegionNum(table[i].regionNum)) {
      return dpRegion.getSubRegion(table[i].triB);
    }
  }
  Die("no triB");
  return region;
}
TEMPLATE
const DPDeckPSCFG::DPRegion::Region&
CLASS::
getRegionRectB() const
{
  if (isRegionNum(REGION_BE)) {
    return dpRegion.getSubRegion(REGION_BM);
  }
  Die("no rectB");
  return region;
}
TEMPLATE
const DPDeckPSCFG::DPRegion::Region&
CLASS::
getRegionRectE() const
{
  if (isRegionNum(REGION_BE)) {
    return dpRegion.getSubRegion(REGION_ME);
  }
  Die("no rectB");
  return region;
}
TEMPLATE
const DPDeckPSCFG::DPRegion::Region&
CLASS::
getRegionRectO() const
{
  return dpRegion.getSubRegion(REGION_BE);
}
TEMPLATE
const DPDeckPSCFG::DPRegion::Region&
CLASS::
getRegionRectI() const
{
  if (isRegionNum(REGION_BM)) {
    return dpRegion.getSubRegion(REGION_ME);
  } else if (isRegionNum(REGION_ME)) {
    return dpRegion.getSubRegion(REGION_BM);
  }
  Die("no rectI");
  return region;
}
TEMPLATE
const DPDeckPSCFG::DPRegion::Region&
CLASS::
getRegion1(const int& bfType) const
{// left region or outside region
  if (isTriIn()) {
    Assert(bfType == BF_TYPE_0);
    return region;
  } else if (isRectIn()) {
    switch (bfType) {
    case BF_TYPE_0: return getRegionTriB();
    case BF_TYPE_1: return region;
    case BF_TYPE_2: return getRegionRectB();
    }
  } else if (isTriOut()) {
    switch (bfType) {
    case BF_TYPE_0: return region;
    case BF_TYPE_1: return region;
    }
  } else if (isRectOut()) {
    switch (bfType) {
    case BF_TYPE_0: return region;
    case BF_TYPE_1: return region;
    case BF_TYPE_2: return getRegionRectO();
    }
  }
  Die("bad bfType=%d", bfType);
  return region;
}
TEMPLATE
const DPDeckPSCFG::DPRegion::Region&
CLASS::
getRegion2(const int& bfType) const
{// right region or inside region
  if (isTriIn()) {
    Assert(bfType == BF_TYPE_0);
    return region;
  } else if (isRectIn()) {
    switch (bfType) {
    case BF_TYPE_0: return region;
    case BF_TYPE_1: return getRegionTriE();
    case BF_TYPE_2: return getRegionRectE();
    }
  } else if (isTriOut()) {
    switch (bfType) {
    case BF_TYPE_0: return region;
    case BF_TYPE_1: return region;
    }
  } else if (isRectOut()) {
    switch (bfType) {
    case BF_TYPE_0: return getRegionTriE();
    case BF_TYPE_1: return getRegionTriB();
    case BF_TYPE_2: return getRegionRectI();
    }
  }
  Die("bad bfType=%d", bfType);
  return region;
}

TEMPLATE
const DPDeckPSCFG::DPScoreT& 
CLASS::
#ifdef SLR_STATE
vstb_ref(int bx, int by, int ex, int ey, State s) const 
#else
vstb_ref(int bx, int by, int ex, int ey) const 
#endif
{
  const Rect rect = {{{bx, by}}, {{ex, ey}}};
#ifdef SLR_STATE
  return vstb.ref(region, rect, s);
#else
  return vstb.ref(region, rect);
#endif
}
TEMPLATE
const DPDeckPSCFG::DPScoreT& 
CLASS::
#ifdef SLR_STATE
vste_ref(int bx, int by, int ex, int ey, State s) const 
#else
vste_ref(int bx, int by, int ex, int ey) const 
#endif
{
  const Rect rect = {{{bx, by}}, {{ex, ey}}};
#ifdef SLR_STATE
  return vste.ref(region, rect, s);
#else
  return vste.ref(region, rect);
#endif
}
TEMPLATE
const DPDeckPSCFG::DPScoreT&
CLASS::
vst_rect1_ref(int bfType, int bx, int by, int ex, int ey) const
{// left region or outside region
  const Rect rect1 = {{{bx, by}}, {{ex, ey}}};
  Assert(dpRegion.includes(rect1));
  Assert(dpRegion.isSkipRect(rect1));
  const DPRegion::Region& reg1 = getRegion1(bfType);
  Assert(reg1.includes(rect1));
#ifdef SLR_STATE
#define VSTB_REF(reg, rect, s) vstb.ref(reg, rect, s)
#define VSTE_REF(reg, rect, s) vste.ref(reg, rect, s)
#else
#define VSTB_REF(reg, rect, s) vstb.ref(reg, rect)
#define VSTE_REF(reg, rect, s) vste.ref(reg, rect)
#endif
  if (isTriIn()) {
    Assert(bfType == BF_TYPE_0);
    return VSTB_REF(reg1, rect1, Model::SL_st);
  } else if (isRectIn()) {
    switch (bfType) {
    case BF_TYPE_0: return VSTB_REF(reg1, rect1, Model::SL_st);
    case BF_TYPE_1: return VSTB_REF(reg1, rect1, Model::SL_st);
    case BF_TYPE_2: return VSTB_REF(reg1, rect1, Model::SL_st);
    }
  } else if (isTriOut()) {
    switch (bfType) {
    case BF_TYPE_0: return VSTB_REF(reg1, rect1, Model::BF_st);
    case BF_TYPE_1: return VSTE_REF(reg1, rect1, Model::BF_st);
    }
  } else if (isRectOut()) {
    switch (bfType) {
    case BF_TYPE_0: return VSTB_REF(reg1, rect1, Model::BF_st);
    case BF_TYPE_1: return VSTE_REF(reg1, rect1, Model::BF_st);
    case BF_TYPE_2:
      if (dpRegion.isType(CARET_CUT_L)) {
	return VSTB_REF(reg1, rect1, Model::BF_st);
      } else if (dpRegion.isType(CARET_CUT_R)) {
	return VSTE_REF(reg1, rect1, Model::BF_st);
      } else {
	Die("bad dpRegion.type=%d", dpRegion.getType());
      }
    }
  }
  Die("bad bfType=%d", bfType);
  return NegInfDPScore;
#undef VSTB_REF
#undef VSTE_REF
}
TEMPLATE
const DPDeckPSCFG::DPScoreT&
CLASS::
vst_rect2_ref(int bfType, int bx, int by, int ex, int ey) const
{// right region or inside region
  Rect rect2 = {{{bx, by}}, {{ex, ey}}};
  Assert(dpRegion.includes(rect2));
  Assert(dpRegion.isSkipRect(rect2));
  const DPRegion::Region& reg2 = getRegion2(bfType);
  Assert(reg2.includes(rect2));
#ifdef SLR_STATE
#define VSTB_REF(reg, rect, s) vstb.ref(reg, rect, s)
#define VSTE_REF(reg, rect, s) vste.ref(reg, rect, s)
#else
#define VSTB_REF(reg, rect, s) vstb.ref(reg, rect)
#define VSTE_REF(reg, rect, s) vste.ref(reg, rect)
#endif
  if (isTriIn()) {
    Assert(bfType == BF_TYPE_0);
    return VSTE_REF(reg2, rect2, Model::SR_st);
  } else if (isRectIn()) {
    switch (bfType) {
    case BF_TYPE_0: return VSTE_REF(reg2, rect2, Model::SR_st);
    case BF_TYPE_1: return VSTE_REF(reg2, rect2, Model::SR_st);
    case BF_TYPE_2: return VSTE_REF(reg2, rect2, Model::SR_st);
    }
  } else if (isTriOut()) {
    switch (bfType) {
    case BF_TYPE_0: return dpTask.parent->VSTB_REF(reg2, rect2, Model::SR_st);
    case BF_TYPE_1: return dpTask.parent->VSTE_REF(reg2, rect2, Model::SL_st);
    }
  } else if (isRectOut()) {
    switch (bfType) {
    case BF_TYPE_0: return VSTB_REF(reg2, rect2, Model::SR_st);
    case BF_TYPE_1: return VSTE_REF(reg2, rect2, Model::SL_st);
    case BF_TYPE_2:
      if (dpRegion.isType(CARET_CUT_L)) {
	return VSTB_REF(reg2, rect2, Model::SR_st);
      } else if (dpRegion.isType(CARET_CUT_R)) {
	return VSTE_REF(reg2, rect2, Model::SL_st);
      } else {
	Die("bad dpRegion.type=%d", dpRegion.getType());
      }
    }
  }
  Die("bad bfType=%d", bfType);
  return NegInfDPScore;
#undef VSTB_REF
#undef VSTE_REF
}

TEMPLATE
const DPDeckPSCFG::BENode& 
CLASS::
vbe_ref(int x, int y) const 
{
  const Point pt = {x, y};
  return vbe.ref(region, pt);
}
TEMPLATE
const DPDeckPSCFG::ScoreManager::SCNode& 
CLASS::
vsc_ref(int diag, int ex, int ey) const 
{
  const Point e = {ex, ey};
  return vsc.ref(region, diag, e);
}
TEMPLATE
const DPDeckPSCFG::ScoreManager::SCNode& 
CLASS::
vsfbwB_refI(int by, int ex, int ey) const 
{
  return vsfbwB.refI(region, by, ex, ey);
}
TEMPLATE
const DPDeckPSCFG::ScoreManager::SCNode& 
CLASS::
vsfbwB_refO(int by, int ex, int ey) const 
{
  return vsfbwB.refO(region, by, ex, ey);
}
TEMPLATE
const DPDeckPSCFG::ScoreManager::SCNode& 
CLASS::
vsfbwE_refI(int bx, int by, int ey) const 
{
  return vsfbwE.refI(region, bx, by, ey);
}
TEMPLATE
const DPDeckPSCFG::ScoreManager::SCNode& 
CLASS::
vsfbwE_refO(int bx, int by, int ey) const 
{
  return vsfbwE.refO(region, bx, by, ey);
}
TEMPLATE  
const int& 
CLASS::
vskip_skipCountB0_ref(int x) const 
{
  return vskip.skipCountB0_ref(region, x);
}
TEMPLATE  
const int& 
CLASS::
vskip_skipCountE0_ref(int x) const 
{
  return vskip.skipCountE0_ref(region, x);
}
TEMPLATE  
const int& 
CLASS::
vskip_skipCountM0_ref(int x) const 
{
  return vskip.skipCountM0_ref(region, x);
}
TEMPLATE  
const int& 
CLASS::
vskip_skipCount0_ref(int x) const {
  return vskip.skipCount0_ref(region, x);
}
TEMPLATE  
const int& 
CLASS::
vskip_skipCountB_ref(int x, int y) const {
  const Point pt = {x, y};
  return vskip.skipCountB_ref(region, pt);
}
TEMPLATE  
const int& 
CLASS::
vskip_skipShiftB_ref(int x, int y) const 
{
  const Point pt = {x, y};
  return vskip.skipShiftB_ref(region, pt);
}
TEMPLATE  
const int& 
CLASS::
vskip_skipCountE_ref(int x, int y) const 
{
  const Point pt = {x, y};
  return vskip.skipCountE_ref(region, pt);
}
TEMPLATE  
const int& 
CLASS::
vskip_skipShiftE_ref(int x, int y) const 
{
  const Point pt = {x, y};
  return vskip.skipShiftE_ref(region, pt);
}
TEMPLATE  
const DPDeckPSCFG::TBNode& 
CLASS::
vtb_ref(int bx, int by, int ex, int ey) const
{
  const Rect rect = {{{bx, by}},{{ex, ey}}};
  return vtb.ref(region, rect);
}
TEMPLATE  
const DPDeckPSCFG::TBBFNode& 
CLASS::
vtbbf_ref(int bx, int by, int ex, int ey) const
{
  const Rect rect = {{{bx, by}},{{ex, ey}}};
  return vtbbf.ref(region, rect);
}
TEMPLATE  
const DPDeckPSCFG::ScoreManager::SCNode& 
CLASS::
vfw_ref(int bx, int by, int ex, int ey) const
{
  const Rect rect = {{{bx, by}},{{ex, ey}}};
  return vfw.ref(region, rect);
}
TEMPLATE  
const DPDeckPSCFG::ScoreManager::SCNode& 
CLASS::
vbw_ref(int bx, int by, int ex, int ey) const
{
  const Rect rect = {{{bx, by}},{{ex, ey}}};
  return vbw.ref(region, rect);
}

#undef TEMPLATE
#undef CLASS
#endif
