/*
 * DPDeckPSCFGTBTree.hpp
 *
 */
#ifndef DP_DECK_PSCFG_TB_TREE_HPP
#define DP_DECK_PSCFG_TB_TREE_HPP

#include "DPPath.hpp"
#include "DPDeckPSCFG.hpp"

class CT;
class 
DPDeckPSCFG::
TBTree 
{
public:
  class Node;
  typedef Node* Link;
  typedef Node node_type;

  TBTree();
  TBTree(const TBTree& tbTree);
  TBTree(const Vector<Link>& treeNodes, const Vector<State>& states);
  ~TBTree();
  TBTree& operator=(const TBTree& tbTree);
  string toString() const;
  void print() const;

  void clear();
  void reset(const Vector<Link>& treeNodes,
	     const Vector<State>& states);
  bool empty() const {return (head == NULL);}
  Link getRoot() {return head;}
  Link getRoot() const {return head;}
  DPPath getDPPath() const;
  DPPath getDPPath(const DPPath& dpPath0) const;
  bool isLeftEnd(const Node*const& link) const;
  bool isRightEnd(const Node*const& link) const;
  State getStateLeft() const;
  State getStateRight() const;
  void replaceNode(Link& link, const Vector<Link>& treeNodes);
  void testConsistency() const;
private:
  Link head;
  Link left;
  Link right;
  State state_h;
  State state_l;
  State state_r;

  friend class Node;
  enum {
    M_st, I_st, D_st
  };
  class AlnElem {
  public:
    Point pt;
    int state;
    AlnElem* pair;
    AlnElem(const Point& pt, int state)
      : pt(pt), state(state), pair(NULL) {}
  };
  typedef deque<AlnElem*> AlnFlagments;
};
class 
DPDeckPSCFG::TBTree::
Node 
{
public:
  Link parent;
  Link leftChild;
  Link rightChild;
  Rect rect;
  State state;

  Node();
  Node(const Link& parent,
       const Link& leftChild,
       const Link& rightChild,
       const Rect& rect,
       const State& state);
  Node(const Node& node);
  ~Node();
  Node& operator=(const Node& node);
  string toString() const;
  void print() const;

  bool hasParent() const;
  bool hasChild(int which) const;
  Link getParent() const;
  Link getChild(int which) const;

  int getNChild() const;
  void replaceChild(const Link& curr, const Link& next);
  bool isLeftChild(const Link& link) const;
  bool isRightChild(const Link& link) const;
  bool isChild(const Link& link) const;
  void testConsistencyR();

private:
  friend class TBTree;
  void deleteR();
  AlnFlagments getAlnFlagmentsR() const;
};

#endif
