/*
 * DPDeckPSCFGVectors.cpp
 *
 */
#ifndef DP_DECK_PSCFG_VECTORS_CPP
#define DP_DECK_PSCFG_VECTORS_CPP
#include "DPDeckPSCFGVectors.hpp"

DPDeckPSCFG::VTZ::
VTZ()
  : Super(), headB(-1), tailB(-1), headE(-1), tailE(-1)
{
}
DPDeckPSCFG::VTZ::
~VTZ()
{
}
string
DPDeckPSCFG::VTZ::
toString() const
{
  ostringstream oss;
  for (int i = 0; i < (int) size(); i++) {
    const TZNode& tn = get(i);
    oss << setw(4) << i << ":";
    oss << "type=" << Trapezoid::type2str(tn.type) << ",";
    oss << "topLeftMargin=" << tn.topLeftMargin << ",";
    oss << "bottomRightMargin=" << tn.bottomRightMargin << ",";
  }
  return oss.str();
}
void
DPDeckPSCFG::VTZ::
print() const
{
  cout << toString() << flush;
}
bool
DPDeckPSCFG::VTZ::
empty() const
{
  return (Super::empty() || ((Super::size() - marginSize) > 0));
}
int
DPDeckPSCFG::VTZ::
size() const
{
  return (empty() ? 0 : (Super::size() - marginSize));
}
int
DPDeckPSCFG::VTZ::
computeSize(const DPRegion::Region& region)
{
  const int& lenB = region.srb.size();
  const int& lenE = region.sre.size();
  if (region.isTriShape()) {
    return (lenB + marginSize);
  } else {
    return (lenB + lenE + marginSize);
  }
}
int
DPDeckPSCFG::VTZ::
computeSizeInBytes(const DPRegion::Region& region)
{
  return computeSize(region) * sizeof(ValueT);
}
void
DPDeckPSCFG::VTZ::
resize(const DPRegion::Region& region)
{
  const int& lenB = region.srb.size();
  const int& lenE = region.sre.size();
  Super::resize(computeSize(region));
  if (region.isTriShape()) {
    headB = margin_l;
    tailB = margin_l + (lenB - 1);
    headE = headB;
    tailE = tailB;
  } else {
    headB = margin_l;
    tailB = headB + (lenB - 1);
    headE = margin_l + lenB + margin_m;
    tailE = headE + (lenE - 1);
  }
  Super::fill(ValueT::null());
}
string
DPDeckPSCFG::VX::
toString() const
{
  ostringstream oss;
  return oss.str();
}
int
DPDeckPSCFG::VX::
computeSize(const DPRegion::Region& region)
{
  if (region.isTriShape()) {
    return (region.srb.rect0.lengthX() + marginSize);
  } else {
    return (region.srb.rect0.lengthX()
	    + region.sre.rect0.lengthX()
	    + marginSize);
  }
}
int
DPDeckPSCFG::VX::
computeSizeInBytes(const DPRegion::Region& region)
{
  return computeSize(region) * sizeof(ValueT);
}
void
DPDeckPSCFG::VX::
resize(const DPRegion::Region& region)
{
  Super::resize(computeSize(region));
  if (region.isTriShape()) {
    headB = headE = margin_l;
  } else {
    headB = margin_l;
    headE = margin_l + region.srb.rect0.lengthX() + margin_m;
  }
  bx_beg = region.srb.bx0;
  bx_last = region.srb.ex0;
  ex_beg = region.sre.bx0;
  ex_last = region.sre.ex0;
  Super::fill(ValueT::null());
}
int
DPDeckPSCFG::VX::
getIndexB(int x) const
{
  Assert(bx_beg - 1 <= x && x <= bx_last + 1);
  return (headB + (x - bx_beg));
}
int
DPDeckPSCFG::VX::
getIndexE(int x) const
{
  Assert(ex_beg - 1 <= x && x <= ex_last + 1);
  return (headE + (x - ex_beg));
}
int
DPDeckPSCFG::VX::
getPos(ValueT*const& p) const
{
  const int& idx = (p - &(Super::get(0)));
  if (idx < (headE - 2)) {
    return bx_beg + (idx - headB);
  } else {
    return ex_beg + (idx - headE);
  }
}
string
DPDeckPSCFG::VY::
toString() const
{
  ostringstream oss;
  return oss.str();
}
int
DPDeckPSCFG::VY::
computeSize(const DPRegion::Region& region)
{
  if (region.isTriShape()) {
    return (region.srb.rect0.lengthY() + marginSize);
  } else {// rect shape
    return (region.sre.ey0 - region.srb.by0 + 1 + marginSize);
  }
}
int
DPDeckPSCFG::VY::
computeSizeInBytes(const DPRegion::Region& region)
{
  return computeSize(region) * sizeof(ValueT);
}
void
DPDeckPSCFG::VY::
resize(const DPRegion::Region& region)
{
  Super::resize(computeSize(region));
  if (region.isTriShape()) {
    headB = headE = margin_l;
  } else if (region.isRectShape()) {
    headB = margin_l;
    headE = margin_l + (region.sre.by0 - region.srb.by0);
  }
  by_beg = region.srb.by0;
  by_last = region.srb.ey0;
  ey_beg = region.sre.by0;
  ey_last = region.sre.ey0;
  Super::fill(ValueT::null());
}
int
DPDeckPSCFG::VY::
getIndexB(int y) const
{
  Assert(by_beg - 1 <= y && y <= by_last + 1);
  return (headB + (y - by_beg));
}
int
DPDeckPSCFG::VY::
getIndexE(int y) const
{
  Assert(ey_beg - 1 <= y && y <= ey_last + 1);
  return (headE + (y - ey_beg));
}
int
DPDeckPSCFG::VY::
getPos(ValueT*const& p) const
{
  const int& idx = (p - &(Super::get(0)));
  if (idx < (headE - 2)) {
    return by_beg + (idx - headB);
  } else {
    return ey_beg + (idx - headE);
  }
}
void
DPDeckPSCFG::VPairX::
getLengths(const DPRegion::Region& region,
	   int& rowLength, int& colLength)
{
  rowLength = (region.srb.rect0.lengthX() + rowMarginB + rowMarginE);
  colLength = (region.sre.rect0.lengthX() + colMarginB + colMarginE);
}
int
DPDeckPSCFG::VPairX::
computeSize(const DPRegion::Region& region)
{
  int rowLength;
  int colLength;
  getLengths(region, rowLength, colLength);
  return (rowLength * colLength);
}
int
DPDeckPSCFG::VPairX::
computeSizeInBytes(const DPRegion::Region& region)
{
  return (computeSize(region) * sizeof(ValueT));
}
void
DPDeckPSCFG::VPairX::
resize(const DPRegion::Region& region)
{
  int rowLength;
  int colLength;
  getLengths(region, rowLength, colLength);
  Super::setLengths(rowLength, colLength);
  Super::fill(ValueT::null());
  bx_beg = region.srb.bx0;
  bx_last = region.srb.ex0;
  ex_beg = region.sre.bx0;
  ex_last = region.sre.ex0;
}
const DPDeckPSCFG::VPairX::ValueT&
DPDeckPSCFG::VPairX::
ref(int bx, int ex) const
{
  const int& row = rowMarginB + (bx - bx_beg);
  const int& col = colMarginB + (ex - ex_beg);
  return Super::ref(row, col);
}
DPDeckPSCFG::VPairX::ValueT&
DPDeckPSCFG::VPairX::
ref(int bx, int ex)
{
  const int& row = rowMarginB + (bx - bx_beg);
  const int& col = colMarginB + (ex - ex_beg);
  return Super::ref(row, col);
}
void
DPDeckPSCFG::VPairY::
getLengths(const DPRegion::Region& region,
	   int& rowLength, int& colLength)
{
  rowLength = (region.srb.rect0.lengthY() + rowMarginB + rowMarginE);
  colLength = (region.sre.rect0.lengthY() + colMarginB + colMarginE);
}
int
DPDeckPSCFG::VPairY::
computeSize(const DPRegion::Region& region)
{
  int rowLength, colLength;
  getLengths(region, rowLength, colLength);
  return (rowLength * colLength);
}
int
DPDeckPSCFG::VPairY::
computeSizeInBytes(const DPRegion::Region& region)
{
  return (computeSize(region) * sizeof(ValueT));
}
void
DPDeckPSCFG::VPairY::
resize(const DPRegion::Region& region)
{
  int rowLength, colLength;
  getLengths(region, rowLength, colLength);
  Super::setLengths(rowLength, colLength);
  Super::fill(ValueT::null());
  by_beg = region.srb.by0;
  by_last = region.srb.ey0;
  ey_beg = region.sre.by0;
  ey_last = region.sre.ey0;
}
const DPDeckPSCFG::VPairY::ValueT&
DPDeckPSCFG::VPairY::
ref(int by, int ey) const
{
  const int& row = rowMarginB + (by - by_beg);
  const int& col = colMarginB + (ey - ey_beg);
  return Super::ref(row, col);
}
DPDeckPSCFG::VPairY::ValueT&
DPDeckPSCFG::VPairY::
ref(int by, int ey)
{
  const int& row = rowMarginB + (by - by_beg);
  const int& col = colMarginB + (ey - ey_beg);
  return Super::ref(row, col);
}
DPDeckPSCFG::VBE::
VBE()
  : Super(), bSize(0), eSize(0),
    headB(-1), tailB(-1), headE(-1), tailE(-1)
{
}
DPDeckPSCFG::VBE::
~VBE()
{
}
string
DPDeckPSCFG::VBE::
toString() const
{
  ostringstream oss;
  return oss.str();
}
int
DPDeckPSCFG::VBE::
computeSize(const DPRegion::Region& region)
{
  return region.getVBESize();
}
int
DPDeckPSCFG::VBE::
computeSizeInBytes(const DPRegion::Region& region)
{
  return computeSize(region) * sizeof(BENode);
}
void
DPDeckPSCFG::VBE::
resize(const DPRegion::Region& region)
{
  if (region.isTriShape()) {
    bSize = region.srb.getAreaMargin();
    eSize = 0;
    headB = region.srb.getAreaMarginBefore(region.srb.b0);
    tailB = region.srb.getAreaMarginBefore(region.srb.e0);
    headE = headB;
    tailE = tailB;
  } else {
    bSize = region.srb.getAreaMargin();
    eSize = region.sre.getAreaMargin();  
    headB = region.srb.getAreaMarginBefore(region.srb.b0);
    tailB = region.srb.getAreaMarginBefore(region.srb.e0);
    headE = bSize + region.sre.getAreaMarginBefore(region.sre.b0);
    tailE = bSize + region.sre.getAreaMarginBefore(region.sre.e0);
  }
#ifdef DEBUG
  int n = (bSize + eSize);
  Assert(headB < n && tailB < n && headE < n && tailE < n);
#endif
  Super::resize(bSize + eSize);
  Super::fill(BENode::null());
}
int
DPDeckPSCFG::VBE::
getIndexB(const StripRegion& srb, const Point& pt) const
{
  return srb.getAreaMarginBefore(pt);
}
int
DPDeckPSCFG::VBE::
getIndexE(const StripRegion& sre, const Point& pt) const
{
  return (bSize + sre.getAreaMarginBefore(pt));
}
DPDeckPSCFG::VSC::
VSC()
{
}
DPDeckPSCFG::VSC::
~VSC()
{
}
string
DPDeckPSCFG::VSC::
toString() const
{
  ostringstream oss;
  for (int i = 0; i < (int) size(); i++) {
    if (!(get(i) == ValueT::null())) {
      oss << i << ":" << get(i).toString() << "\n";
    }
  }
  return oss.str();
}
int
DPDeckPSCFG::VSC::
computeSize(const DPRegion::Region& region)
{
  const int& diagWidthB = region.srb.getDiagWidth();
  const int& areaMarginE = region.sre.getAreaMargin();
  // + left right Margin ? 
  return (diagWidthB * areaMarginE);
}
int
DPDeckPSCFG::VSC::
computeSizeInBytes(const DPRegion::Region& region)
{
  return computeSize(region) * sizeof(ValueT);
}
void
DPDeckPSCFG::VSC::
resize(const DPRegion::Region& region)
{
  Super::resize(computeSize(region));
  fill(ValueT::null());
}
int
DPDeckPSCFG::VSC::
getIndex(const DPRegion::Region& region,
	 const int& diagB, const Point& ptE) const
{// Note that diag = (y - x)
  Assert(region.srb.getMinDiag() <= diagB
	 && diagB <= region.srb.getMaxDiag());
  Assert(region.sre.includes(ptE));
  const int& diagOffsetB = (-region.srb.getMinDiag());
  const int& areaMarginE = region.sre.getAreaMargin();
  const int& indexE = region.sre.getAreaMarginBefore(ptE);
  return (areaMarginE * (diagOffsetB + diagB) + indexE);
}
DPDeckPSCFG::VSkip::
VSkip()
{
}
DPDeckPSCFG::VSkip::
~VSkip()
{
}
string
DPDeckPSCFG::VSkip::
toString() const
{
  ostringstream oss;
  return oss.str();
}
int
DPDeckPSCFG::VSkip::
computeSizeInBytes(const DPRegion::Region& region)
{
  //const bool& tri = region.isTriShape();
  int n = 0;
  // vSkipCount0
  n += (region.srkb.getNSkipRow()
	+ (region.isTriShape() ? 0 : region.srke.getNSkipRow())
	+ (region.srkm.empty() ? 0 : region.srkm.getNSkipRow()));

  //vSkipCountB
  n += region.getVSkipArea();
  //vSkipShiftB
  n += region.getVSkipArea();
  //vSkipCountE
  n += region.getVSkipArea();
  //vSkipShiftE
  n += region.getVSkipArea();

  return n * sizeof(int);
}
void
DPDeckPSCFG::VSkip::
resize(const DPRegion::Region& region)
{
  int b0Size = region.getB0Size();
  int e0Size = region.getE0Size();
  int m0Size = region.getM0Size();

  int n = 0;
  if (b0Size > 0) {
    headB0 = n;
    tailB0 = (headB0 + b0Size - 1);
    n += b0Size;
  } else {
    headB0 = -1;
    tailB0 = -1;
  }
  if (e0Size > 0) {
    headE0 = n;
    tailE0 = (headE0 + e0Size - 1);
    n += e0Size;
  } else {
    headE0 = headB0;
    tailE0 = tailB0;
  }
  if (m0Size > 0) {
    headM0 = n;
    tailM0 = (headM0 + m0Size - 1);
    n += m0Size;
  } else {
    headM0 = -1;
    tailM0 = -1;
  }
  vSkipCount0.resize(n);
  vSkipCount0.fill(0);

  vSkipCountB.resize(region.getVSkipArea());
  vSkipShiftB.resize(region.getVSkipArea());
  vSkipCountB.fill(0);
  vSkipShiftB.fill(0);// this zero is important

  vSkipCountE.resize(region.getVSkipArea());
  vSkipShiftE.resize(region.getVSkipArea());
  vSkipCountE.fill(0);
  vSkipShiftE.fill(0);// this zero is important
}
int&
DPDeckPSCFG::VSkip::
skipCountB0_ref(const DPRegion::Region& region, const int& bx) {
  Assert(region.skipSet.isSkipRow(bx));
  Assert(region.bxk_beg <= bx && bx <= region.bxk_last);
  return skipCount0_ref(region, bx);
}
int&
DPDeckPSCFG::VSkip::
skipCountE0_ref(const DPRegion::Region& region, const int& ex)
{
  Assert(region.skipSet.isSkipRow(ex));
  Assert(region.exk_beg <= ex && ex <= region.exk_last);
  return skipCount0_ref(region, ex);
}
int&
DPDeckPSCFG::VSkip::
skipCountM0_ref(const DPRegion::Region& region, const int& mx)
{
  Assert(region.skipSet.isSkipRow(mx));
  Assert(region.mxk_beg <= mx && mx <= region.mxk_last);
  return skipCount0_ref(region, mx);
}
int&
DPDeckPSCFG::VSkip::
skipCount0_ref(const DPRegion::Region& region, const int& xk)
{
  Assert(region.skipSet.isSkipRow(xk));
  const int xn = region.convertXK2XN(xk);
  Assert(xn >= 0);
  return vSkipCount0[xn];
}
int& 
DPDeckPSCFG::VSkip::
skipCountB_ref(const DPRegion::Region& region, const Point& pt)
{
  return vSkipCountB[region.getVSkipIndex(pt)];
}
int& 
DPDeckPSCFG::VSkip::
skipShiftB_ref(const DPRegion::Region& region, const Point& pt)
{
  return vSkipShiftB[region.getVSkipIndex(pt)];
}
int& 
DPDeckPSCFG::VSkip::
skipCountE_ref(const DPRegion::Region& region, const Point& pt)
{
  return vSkipCountE[region.getVSkipIndex(pt)];
}
int& 
DPDeckPSCFG::VSkip::
skipShiftE_ref(const DPRegion::Region& region, const Point& pt)
{
  return vSkipShiftE[region.getVSkipIndex(pt)];
}

int
DPDeckPSCFG::VSTB::
computeSize(const DPRegion::Region& region)
{
  return region.getVSTSize();
}
int
DPDeckPSCFG::VSTB::
computeSizeInBytes(const DPRegion::Region& region)
{
  return (computeSize(region) * sizeof(ValueT));
}
void
DPDeckPSCFG::VSTB::
resize(const DPRegion::Region& region, State s)
{
  data[indexOf(region)].resize(region, s);
}
void
DPDeckPSCFG::VSTB::Datum::
resize(const DPRegion::Region& region, State s)
{
  periodX = region.skipSet.getSkipRowPeriod();
  periodY = region.skipSet.getHorizPeriod();
#ifdef SLR_STATE
  const int& idx = Model::getSTStateIndex(s);
#else
  const int& idx = 0;
#endif
  const int& vstSize = region.getVSTSize();
  if (vstSize == 0) {
    for (int i = 0; i < NST_STATE; i++) {
      v[i].clear();
      v_ptrs[i] = NULL;
    }
    vBYRange.clear();
    vEXRange.clear();
    vEYRange.clear();
    return;
  }
  v[idx].resize(vstSize);
  v_ptrs[idx] = &v[idx][0];
  v[idx].fill(NegInfDPScore);
  vBYRange.resize(region.srkb.getNSkipRow());
  vEXRange.resize(region.srkb.getArea());
  vEYRange.resize(region.getVSTBVEYRangeSize());

  BYRange* pBYRange = &vBYRange.back();
  EXRange* pEXRange = &vEXRange.back();
  EYRange* pEYRange = &vEYRange.front();
  //ValueT*  pVal     = &v.front();
  int offset = 0;

  bxRange.pBYRange = pBYRange;
#ifdef DEBUG
  bxRange.bx_beg = region.srkb.getFirstSkipRow();
#endif
  bxRange.bx_last = region.srkb.getLastSkipRow();
  StripRegionSkip::RowIterator<false> ritb(region.srkb);
  StripRegionSkip::RowIterator<true> rite(region.srke);
  Rect bboxe = region.sre.rect0;
  while (ritb.getNext()) {
    BYRange& byRange = (*pBYRange--);
    if (ritb.ly == 0) {
      byRange.pEXRange = NULL;
#ifdef DEBUG
      byRange.by_beg = (-1);
#endif
      byRange.by_last = (-2);
      continue;
    }
    byRange.pEXRange = pEXRange;
#ifdef DEBUG
    byRange.by_beg  = ritb.y_beg;
#endif
    byRange.by_last = ritb.y_last;
    for (int by = ritb.y_last; by >= ritb.y_beg; by -= periodY) {
      EXRange& exRange = (*pEXRange--);
      const int& ex_beg  = max(ritb.x, region.srke.getFirstSkipRow());
      const int& ex_last = region.srke.getLastSkipRow();
      if (ex_last < 0) {
	exRange.pEYRange = NULL;
	exRange.ex_beg = (-1);
#ifdef DEBUG
	exRange.ex_last = (-2);
#endif
	continue;
      }
      exRange.pEYRange = pEYRange;
      exRange.ex_beg  = ex_beg;
#ifdef DEBUG
      exRange.ex_last = ex_last;
#endif

      bboxe.bx = ritb.x;
      bboxe.by = by;
      rite.reset(bboxe);
      while (rite.getNext()) {
	EYRange& eyRange = (*pEYRange++);
	if (rite.ly == 0) {
	  //eyRange.pVal = NULL;
	  eyRange.offset = (-1);
	  eyRange.ey_beg = (-1);
#ifdef DEBUG
	  eyRange.ey_last = (-2);
#endif
	  continue;
	}
	//eyRange.pVal = pVal;
	eyRange.offset = offset;
	eyRange.ey_beg = rite.y_beg;
#ifdef DEBUG
	eyRange.ey_last = rite.y_last;
#endif
	//pVal += rite.ly;
	offset += rite.ly;
      }
    }
  }
#ifdef DEBUG
  Assert((pBYRange + 1) == &vBYRange.front());
  Assert((pEXRange + 1) == &vEXRange.front());
  Assert((pEYRange - 1) == &vEYRange.back());
  Assert(offset == (int)v[idx].size());
  DPRegion::Region::Iterator<true> dpit(region);
  while (dpit.getNext()) {
    if (!region.skipSet.isSkipRect(dpit.rect)) continue;
#ifdef SLR_STATE
    const ValueT& val = ref(region, dpit.rect, s);
#else
    const ValueT& val = ref(region, dpit.rect);
#endif
    Assert((&val - &v[idx][0]) == region.getVSTBIndex(dpit.rect));
  }
#endif
}
const DPDeckPSCFG::VSTB::ValueT&
DPDeckPSCFG::VSTB::Datum::
#ifdef SLR_STATE
ref(const DPRegion::Region& region, const Rect& rect, const State& s) const
#else
ref(const DPRegion::Region& region, const Rect& rect) const
#endif
{
  Assert(region.includes(rect) && region.skipSet.isSkipRect(rect));
  BYRange& byRange = bxRange.getBYRange(rect.bx, periodX);
  EXRange& exRange = byRange.getEXRange(rect.by, periodY);
  EYRange& eyRange = exRange.getEYRange(rect.ex, periodX);
  //ValueT& val = eyRange.ref(rect.ey, periodY);
  const int& offset = eyRange.ref(rect.ey, periodY);
#ifdef SLR_STATE
  const int& idx = Model::getSTStateIndex(s);
#else
  const int& idx = 0;
#endif
  Assert(0 <= offset && offset < (int)v[idx].size());
  return v_ptrs[idx][offset];
}
DPDeckPSCFG::VSTB::ValueT&
DPDeckPSCFG::VSTB::Datum::
#ifdef SLR_STATE
ref(const DPRegion::Region& region, const Rect& rect, const State& s)
#else
ref(const DPRegion::Region& region, const Rect& rect)
#endif
{
  Assert(region.includes(rect) && region.skipSet.isSkipRect(rect));
  BYRange& byRange = bxRange.getBYRange(rect.bx, periodX);
  EXRange& exRange = byRange.getEXRange(rect.by, periodY);
  EYRange& eyRange = exRange.getEYRange(rect.ex, periodX);
  //ValueT& val = eyRange.ref(rect.ey, periodY);
  const int& offset = eyRange.ref(rect.ey, periodY);
#ifdef SLR_STATE
  const int& idx = Model::getSTStateIndex(s);
#else
  const int& idx = 0;
#endif
  Assert(0 <= offset && offset < (int)v[idx].size());
  return v_ptrs[idx][offset];
}
#ifdef SLR_STATE
void
DPDeckPSCFG::VSTB::
copy(const DPRegion::Region& region,
     const VSTE& vste, const State& s)
{
  Assert(computeSize(region) == vste.computeSize(region));
  if (computeSize(region) == 0) return;

  StripRegionSkip::Iterator<false> itb(region.srkb);
  StripRegionSkip::Iterator<true> ite(region.srke);
  Rect bbox = region.sre.rect0;
  DPScoreT* st = &head(region, s);
  Rect rect;
  while (itb.getNext()) {
    bbox.b = itb.pt;
    rect.b = itb.pt;
    ite.reset(bbox);
    while (ite.getNext()) {
      rect.e = ite.pt;
      Assert(st == &ref(region, rect, s));
      (*st++) = vste.ref(region, rect, s);
    }
  }
  Assert((st - 1) == &tail(region, s));
}
void
DPDeckPSCFG::VSTB::
copy(const DPRegion::Region& region,
     const VSTB& vstb1, const DPRegion::Region& region1, const State& s)
{
  Assert(computeSize(region) <= vstb1.computeSize(region));
  if (computeSize(region) == 0) return;

  StripRegionSkip::Iterator<false> itb(region.srkb);
  StripRegionSkip::RowIterator<true> ite(region.srke);
  Rect bbox = region.sre.rect0;
  DPScoreT* st = &head(region, s);
  const DPScoreT* st1 = NULL;
  Rect rect;
  while (itb.getNext()) {
    bbox.b = itb.pt;
    rect.b = itb.pt;
    ite.reset(bbox);
    while (ite.getNext()) {
      if (ite.ly == 0) continue;

      rect.e.setVal(ite.x, ite.y_beg);    
      Assert(st == &ref(region, rect, s));
      st1 = &vstb1.ref(region1, rect, s);
      std::copy(st1, st1 + ite.ly, st);
      st += ite.ly;
    }
  }
  Assert((st - 1) == &tail(region, s));
}
void
DPDeckPSCFG::VSTB::
copy(const DPRegion::Region& region,
     const VSTE& vste1, const DPRegion::Region& region1, const State& s)
{
  Assert(computeSize(region) <= vste1.computeSize(region));
  if (computeSize(region) == 0) return;

  StripRegionSkip::Iterator<false> itb(region.srkb);
  StripRegionSkip::Iterator<true> ite(region.srke);
  Rect bbox = region.sre.rect0;
  DPScoreT* st = &head(region, s);
  Rect rect;
  while (itb.getNext()) {
    bbox.b = itb.pt;
    rect.b = itb.pt;
    ite.reset(bbox);
    while (ite.getNext()) {
      rect.e = ite.pt;
      Assert(st == &ref(region, rect, s));
      (*st++) = vste1.ref(region1, rect, s);
    }
  }
  Assert((st - 1) == &tail(region, s));
}
#else

void
DPDeckPSCFG::VSTB::
copy(const DPRegion::Region& region, const VSTE& vste)
{
  Assert(computeSize(region) == vste.computeSize(region));
  if (computeSize(region) == 0) return;

  StripRegionSkip::Iterator<false> itb(region.srkb);
  StripRegionSkip::Iterator<true> ite(region.srke);
  Rect bbox = region.sre.rect0;
  DPScoreT* st = &head(region);
  Rect rect;
  while (itb.getNext()) {
    bbox.b = itb.pt;
    rect.b = itb.pt;
    ite.reset(bbox);
    while (ite.getNext()) {
      rect.e = ite.pt;
      Assert(st == &ref(region, rect));
      (*st++) = vste.ref(region, rect);
    }
  }
  Assert((st - 1) == &tail(region));
}
void
DPDeckPSCFG::VSTB::
copy(const DPRegion::Region& region,
     const VSTB& vstb1, const DPRegion::Region& region1)
{
  Assert(computeSize(region) <= vstb1.computeSize(region));
  if (computeSize(region) == 0) return;

  StripRegionSkip::Iterator<false> itb(region.srkb);
  StripRegionSkip::RowIterator<true> ite(region.srke);
  Rect bbox = region.sre.rect0;

#if 1

  DPScoreT* st = &head(region);
  const DPScoreT* st1 = NULL;
  Rect rect;
  while (itb.getNext()) {
    bbox.b = itb.pt;
    rect.b = itb.pt;
    ite.reset(bbox);
    while (ite.getNext()) {
      if (ite.ly == 0) continue;

      rect.e.setVal(ite.x, ite.y_beg);    
#ifdef DEBUG
      Assert(st == &ref(region, rect));
#endif

      st1 = &vstb1.ref(region1, rect);
      std::copy(st1, st1 + ite.ly, st);
      st += ite.ly;
    }
  }
  Assert((st - 1) == &tail(region));

#elif
  const SkipSet& skipSet = region.getSkipSet();
  Rect bbox1 = region1.sre.rect0;
  StripRegionSkip::Iterator<false> itb1(region1.srkb);
  StripRegionSkip::RowIterator<true> ite1(region1.srke);
  const DPScoreT* st1 = &vstb1.head(region1);
  while (itb.getNext()) {
    while (itb1.getNext()) {
      if (itb.pt == itb1.pt) {
	bbox.b = itb.pt;
	bbox1.b = itb1.pt;
	ite.reset(bbox);
	ite1.reset(bbox);
	while (ite.getNext()) {
	  ite1.getNext();
	  Assert(ite.x == ite1.x);
	  const int& offset = 
	    (skipSet.getNSkipPointY(ite.x, ite1.y_beg, ite.y_beg) - 1);
#ifdef DEBUG
	  Rect rect = {{itb.pt}, {ite.x, ite.y_beg}};
	  Assert(st == &ref(region, rect));
	  Assert(st1 + offset == &vstb1.ref(region1, rect));
#endif
	  std::copy(st1 + offset, st1 + offset + ite.ly, st);
	  st += ite.ly;
	  st1 += ite1.ly;
	}
	while (ite1.getNext()) {
	  st1 += ite1.ly;
	}
	break;
      } else {
	st1 += region1.srke.getAreaBottomRight(itb1.pt);
      }
    }
  }
  Assert(st - 1 == &tail(region));
  // write more check codes for this method
#endif
}
void
DPDeckPSCFG::VSTB::
copy(const DPRegion::Region& region,
     const VSTE& vste1, const DPRegion::Region& region1)
{
  Assert(computeSize(region) <= vste1.computeSize(region));
  if (computeSize(region) == 0) return;

  StripRegionSkip::Iterator<false> itb(region.srkb);
  StripRegionSkip::Iterator<true> ite(region.srke);
  Rect bbox = region.sre.rect0;
  DPScoreT* st = &head(region);
  Rect rect;
  while (itb.getNext()) {
    bbox.b = itb.pt;
    rect.b = itb.pt;
    ite.reset(bbox);
    while (ite.getNext()) {
      rect.e = ite.pt;
      Assert(st == &ref(region, rect));
      (*st++) = vste1.ref(region1, rect);
    }
  }
  Assert((st - 1) == &tail(region));
}
#endif
int
DPDeckPSCFG::VSTE::
computeSize(const DPRegion::Region& region)
{
  return region.getVSTSize();
}
int
DPDeckPSCFG::VSTE::
computeSizeInBytes(const DPRegion::Region& region)
{
  return (region.getVSTSize() * sizeof(DPScoreT));
}
void
DPDeckPSCFG::VSTE::
resize(const DPRegion::Region& region, State s)
{
  data[indexOf(region)].resize(region, s);
}
void
DPDeckPSCFG::VSTE::Datum::
resize(const DPRegion::Region& region, State s)
{
  periodX = region.skipSet.getSkipRowPeriod();
  periodY = region.skipSet.getHorizPeriod();
#ifdef SLR_STATE
  const int& idx = Model::getSTStateIndex(s);
#else
  const int& idx = 0;
#endif
  const int& vstSize = region.getVSTSize();
  if (vstSize == 0) {
    for (int i = 0; i < NST_STATE; i++) {
      v[i].clear();
      v_ptrs[i] = NULL;
    }
    vEYRange.clear();
    vBXRange.clear();
    vBYRange.clear();
    exRange.pEYRange = NULL;
    exRange.ex_beg = (-1);
#ifdef DEBUG
    exRange.ex_last = (-2);
#endif
    return;
  }
  v[idx].resize(vstSize);
  v_ptrs[idx] = &v[idx][0];
  v[idx].fill(NegInfDPScore);// this is needed until backward bifurcation is implemented.
  vEYRange.resize(region.srke.getNSkipRow());
  vBXRange.resize(region.srke.getArea());
  vBYRange.resize(region.getVSTEVBYRangeSize());

  EYRange* pEYRange = &vEYRange.front();
  BXRange* pBXRange = &vBXRange.front();
  BYRange* pBYRange = &vBYRange.front();
  //ValueT*  pVal     = &v.front(); 
  int offset = 0;

  exRange.pEYRange = pEYRange;
  exRange.ex_beg = region.srke.getFirstSkipRow();
#ifdef DEBUG
  exRange.ex_last = region.srke.getLastSkipRow();
#endif
  StripRegionSkip::RowIterator<true> rite(region.srke);
  StripRegionSkip::RowIterator<true> ritb(region.srkb);
  Rect bboxb = region.srb.rect0;
  while (rite.getNext()) {
    EYRange& eyRange = (*pEYRange++);
    if (rite.ly == 0) {
      eyRange.pBXRange = NULL;
      eyRange.ey_beg = (-1);
#ifdef DEBUG
      eyRange.ey_last = (-2);
#endif
      continue;
    }
    eyRange.pBXRange = pBXRange;
    eyRange.ey_beg  = rite.y_beg;
#ifdef DEBUG
    eyRange.ey_last = rite.y_last;
#endif
    for (int ey = rite.y_beg; ey <= rite.y_last; ey += periodY) {
      BXRange& bxRange = (*pBXRange++);
      const int& bx_beg  = region.srkb.getFirstSkipRow();
#ifdef DEBUG
      const int& bx_last = min(rite.x, region.srkb.getLastSkipRow());
#endif
      if (bx_beg < 0) {
	bxRange.pBYRange = NULL;
	bxRange.bx_beg = (-1);
#ifdef DEBUG
	bxRange.bx_last = (-2);
#endif
	continue;
      }
      bxRange.pBYRange = pBYRange;
      bxRange.bx_beg  = bx_beg;
#ifdef DEBUG
      bxRange.bx_last = bx_last;
#endif

      bboxb.ex = rite.x;
      bboxb.ey = ey;
      ritb.reset(bboxb);
      while (ritb.getNext()) {
	BYRange& byRange = (*pBYRange++);
	if (ritb.ly == 0) {
	  //byRange.pVal = NULL;
	  byRange.offset = (-1);
	  byRange.by_beg = (-1);
#ifdef DEBUG
	  byRange.by_last = (-2);
#endif
	  continue;
	}
	//byRange.pVal = pVal;
	byRange.offset = offset;
	byRange.by_beg = ritb.y_beg;
#ifdef DEBUG
	byRange.by_last = ritb.y_last;
#endif
	//pVal += ritb.ly;
	offset += ritb.ly;
      }
    }
  }
#ifdef DEBUG
  Assert((pEYRange - 1) == &vEYRange.back());
  Assert((pBXRange - 1) == &vBXRange.back());
  Assert((pBYRange - 1) == &vBYRange.back());
  Assert(offset == (int)v[idx].size());
  DPRegion::Region::Iterator<true> dpit(region);
  while (dpit.getNext()) {
    if (!region.skipSet.isSkipRect(dpit.rect)) continue;
#if SLR_STATE
    const ValueT& val = ref(region, dpit.rect, s);
#else
    const ValueT& val = ref(region, dpit.rect);
#endif
    Assert((&val - &v[idx][0]) == region.getVSTEIndex(dpit.rect));
  }
#endif
}
const DPDeckPSCFG::VSTE::ValueT&
DPDeckPSCFG::VSTE::Datum::
#ifdef SLR_STATE
ref(const DPRegion::Region& region, const Rect& rect, const State& s) const
#else
ref(const DPRegion::Region& region, const Rect& rect) const
#endif
{
  Assert(region.includes(rect) && region.skipSet.isSkipRect(rect));
  EYRange& eyRange = exRange.getEYRange(rect.ex, periodX);
  BXRange& bxRange = eyRange.getBXRange(rect.ey, periodY);
  BYRange& byRange = bxRange.getBYRange(rect.bx, periodX);
  //ValueT& val = byRange.ref(rect.by, periodY);
  const int& offset = byRange.ref(rect.by, periodY);
#ifdef SLR_STATE
  const int& idx = Model::getSTStateIndex(s);
#else
  const int& idx = 0;
#endif
  Assert(0 <= offset && offset < (int)v[idx].size());
  return v_ptrs[idx][offset];
}
DPDeckPSCFG::VSTE::ValueT&
DPDeckPSCFG::VSTE::Datum::
#ifdef SLR_STATE
ref(const DPRegion::Region& region, const Rect& rect, const State& s)
#else
ref(const DPRegion::Region& region, const Rect& rect)
#endif
{
  Assert(region.includes(rect) && region.skipSet.isSkipRect(rect));
  EYRange& eyRange = exRange.getEYRange(rect.ex, periodX);
  BXRange& bxRange = eyRange.getBXRange(rect.ey, periodY);
  BYRange& byRange = bxRange.getBYRange(rect.bx, periodX);
  //ValueT& val = byRange.ref(rect.by, periodY);
  const int& offset = byRange.ref(rect.by, periodY);
#ifdef SLR_STATE
  const int& idx = Model::getSTStateIndex(s);
#else
  const int& idx = 0;
#endif
  Assert(0 <= offset && offset < (int)v[idx].size());
  return v_ptrs[idx][offset];
}
#ifdef SLR_STATE
void
DPDeckPSCFG::VSTE::
copy(const DPRegion::Region& region,
     const VSTB& vstb, const State& s)
{
  Assert(computeSize(region) == vstb.computeSize(region));
  if (computeSize(region) == 0) return;

  StripRegionSkip::Iterator<true> ite(region.srke);
  StripRegionSkip::Iterator<true> itb(region.srkb);
  Rect bbox = region.srb.rect0;
  DPScoreT* st = &head(region, s);
  Rect rect;
  while (ite.getNext()) {
    bbox.e = ite.pt;
    rect.e = ite.pt;
    itb.reset(bbox);
    while (itb.getNext()) {
      rect.b = itb.pt;
      Assert(st == &ref(region, rect, s));
      (*st++) = vstb.ref(region, rect, s);
    }
  }
  Assert((st - 1) == &tail(region, s));
}
void
DPDeckPSCFG::VSTE::
copy(const DPRegion::Region& region,
     const VSTE& vste1, const DPRegion::Region& region1, const State& s)
{
  Assert(computeSize(region) <= vste1.computeSize(region));
  if (computeSize(region) == 0) return;

  StripRegionSkip::Iterator<true> ite(region.srke);
  StripRegionSkip::RowIterator<true> itb(region.srkb);
  Rect bbox = region.srb.rect0;

  DPScoreT* st = &head(region, s);
  const DPScoreT* st1 = NULL;
  Rect rect;
  while (ite.getNext()) {
    bbox.e = ite.pt;
    rect.e = ite.pt;
    itb.reset(bbox);
    while (itb.getNext()) {
      if (itb.ly == 0) continue;

      rect.b.setVal(itb.x, itb.y_beg);
      Assert(st == &ref(region, rect, s));
      st1 = &vste1.ref(region1, rect, s);
      std::copy(st1, st1 + itb.ly, st);
      st += itb.ly;
    }
  }
  Assert((st - 1) == &tail(region, s));
}
void
DPDeckPSCFG::VSTE::
copy(const DPRegion::Region& region,
     const VSTB& vstb1, const DPRegion::Region& region1, const State& s)
{
  Assert(computeSize(region) <= vstb1.computeSize(region));
  if (computeSize(region) == 0) return;

  StripRegionSkip::Iterator<true> ite(region.srke);
  StripRegionSkip::Iterator<true> itb(region.srkb);
  Rect bbox = region.srb.rect0;
  DPScoreT* st = &head(region, s);
  Rect rect;
  while (ite.getNext()) {
    bbox.e = ite.pt;
    rect.e = ite.pt;
    itb.reset(bbox);
    while (itb.getNext()) {
      rect.b = itb.pt;
      Assert(st == &ref(region, rect, s));
      (*st++) = vstb1.ref(region1, rect, s);
    }
  }
  Assert((st - 1) == &tail(region, s));
}
#else

void
DPDeckPSCFG::VSTE::
copy(const DPRegion::Region& region, const VSTB& vstb)
{
  Assert(computeSize(region) == vstb.computeSize(region));
  if (computeSize(region) == 0) return;

  StripRegionSkip::Iterator<true> ite(region.srke);
  StripRegionSkip::Iterator<true> itb(region.srkb);
  Rect bbox = region.srb.rect0;
  DPScoreT* st = &head(region);
  Rect rect;
  while (ite.getNext()) {
    bbox.e = ite.pt;
    rect.e = ite.pt;
    itb.reset(bbox);
    while (itb.getNext()) {
      rect.b = itb.pt;
      Assert(st == &ref(region, rect));
      (*st++) = vstb.ref(region, rect);
    }
  }
  Assert((st - 1) == &tail(region));
}
void
DPDeckPSCFG::VSTE::
copy(const DPRegion::Region& region,
     const VSTE& vste1, const DPRegion::Region& region1)
{
  Assert(computeSize(region) <= vste1.computeSize(region));
  if (computeSize(region) == 0) return;

  StripRegionSkip::Iterator<true> ite(region.srke);
  StripRegionSkip::RowIterator<true> itb(region.srkb);
  Rect bbox = region.srb.rect0;

#if 1
  DPScoreT* st = &head(region);
  const DPScoreT* st1 = NULL;
  Rect rect;
  while (ite.getNext()) {
    bbox.e = ite.pt;
    rect.e = ite.pt;
    itb.reset(bbox);
    while (itb.getNext()) {
      if (itb.ly == 0) continue;

      rect.b.setVal(itb.x, itb.y_beg);
#ifdef DEBUG
      Assert(st == &ref(region, rect));
#endif
      st1 = &vste1.ref(region1, rect);
      std::copy(st1, st1 + itb.ly, st);
      st += itb.ly;
    }
  }
  Assert(st - 1 == &tail(region));

#else

  const SkipSet& skipSet = region.getSkipSet();
  StripRegionSkip::Iterator<true> ite1(region1.srke);
  StripRegionSkip::RowIterator<true> itb1(region1.srkb);

  Rect bbox1 = region1.srb.rect0;
  const DPScoreT* st1 = &vste1.head(region1);
  while (ite.getNext()) {
    while (ite1.getNext()) {
      if (ite.pt == ite1.pt) {
	bbox.e = ite.pt;
	bbox1.e = ite1.pt;
	itb.reset(bbox);
	itb1.reset(bbox);
	while (itb.getNext()) {
	  while (ite1.getNext()) {
	    if (itb.x != itb1.x) break;
	  }
	  Assert(itb.x == itb1.x);
	  const int& offset = 
	    (skipSet.getNSkipPointY(itb.x, itb1.y_beg, itb.y_beg) - 1);
#ifdef DEBUG
	  Rect rect = {{itb.x, itb.y_beg}, {ite.pt}};
	  Assert(st == &ref(region, rect));
	  Assert(st1 + offset == &vste1.ref(region1, rect));
#endif
	  std::copy(st1 + offset, st1 + offset + itb.ly, st);
	  st += itb.ly;
	  st1 += itb1.ly;
	}
	break;
      } else {
	st1 += region1.srkb.getAreaTopLeft(ite1.pt);
      }
    }
  }
  Assert(st - 1 == &tail(region));
#endif
}
void
DPDeckPSCFG::VSTE::
copy(const DPRegion::Region& region,
     const VSTB& vstb1, const DPRegion::Region& region1)
{
  Assert(computeSize(region) <= vstb1.computeSize(region));
  if (computeSize(region) == 0) return;

  StripRegionSkip::Iterator<true> ite(region.srke);
  StripRegionSkip::Iterator<true> itb(region.srkb);
  Rect bbox = region.srb.rect0;
  DPScoreT* st = &head(region);
  Rect rect;
  while (ite.getNext()) {
    bbox.e = ite.pt;
    rect.e = ite.pt;
    itb.reset(bbox);
    while (itb.getNext()) {
      rect.b = itb.pt;
      Assert(st == &ref(region, rect));
      (*st++) = vstb1.ref(region1, rect);
    }
  }
  Assert((st - 1) == &tail(region));
}
#endif
string
DPDeckPSCFG::VTB::
toString() const
{
  ostringstream oss;
  return oss.str();
}
void
DPDeckPSCFG::VTB::
print() const
{
  cout << toString() << flush;
}
int
DPDeckPSCFG::VTB::
computeSizeInBytes(const DPRegion::Region& region)
{
  return region.getVTBSize() * sizeof(TBNode);
}
void
DPDeckPSCFG::VTB::
resize(const DPRegion::Region& region)
{
  data[indexOf(region)].resize(region);
}
void
DPDeckPSCFG::VTB::Datum::
resize(const DPRegion::Region& region)
{
  vBYRange.resize(region.srb.rect0.lengthX());
  vEXRange.resize(region.srb.getArea());
  vEYRange.resize(region.getVEYRangeSize());
  v.resize(region.getVTBSize());
  v.fill(ValueT::null());

  BYRange* pBYRange = &vBYRange.back();
  EXRange* pEXRange = &vEXRange.back();
  EYRange* pEYRange = &vEYRange[0];
  ValueT*  pVal     = &v[0]; 

  const int& bx_beg  = region.srb.bx0;
  const int& bx_last = region.srb.ex0;
  bxRange.pBYRange = pBYRange;
#ifdef DEBUG
  bxRange.bx_beg = bx_beg;
#endif
  bxRange.bx_last = bx_last;

  for (int bx = bx_last; bx >= bx_beg; bx--) {
    BYRange& byRange = (*pBYRange--);
    const int& by_beg  = region.srb.getBY(bx);
    const int& by_last = region.srb.getEY(bx);
    byRange.pEXRange = pEXRange;
#ifdef DEBUG
    byRange.by_beg  = by_beg;
#endif
    byRange.by_last = by_last;

    for (int by = by_last; by >= by_beg; by--) {
      EXRange& exRange = (*pEXRange--);
      const int& ex_beg  = max(bx, region.sre.bx0);
      const int& ex_last = region.sre.ex0;
      exRange.pEYRange = pEYRange;
      exRange.ex_beg  = ex_beg;
#ifdef DEBUG
      exRange.ex_last = ex_last;
#endif

      for (int ex = ex_beg; ex <= ex_last; ex++) {
	EYRange& eyRange = (*pEYRange++);
	const int& ey_beg = max(by, region.sre.getBY(ex));
	const int& ey_last = region.sre.getEY(ex);
	const int& ley = (ey_last - ey_beg + 1);
	eyRange.pVal = pVal;
	eyRange.ey_beg = ey_beg;
#ifdef DEBUG
	eyRange.ey_last = ey_last;
#endif
	pVal += ley;
      }
    }
  }
  Assert((pBYRange + 1) == &vBYRange.front());
  Assert((pEXRange + 1) == &vEXRange.front());
  Assert((pEYRange - 1) == &vEYRange.back());
  Assert((pVal - 1) == &v.back());

#ifdef DEBUG
  DPRegion::Region::Iterator<true> dpit(region);
  while (dpit.getNext()) {
    const ValueT& val = ref(region, dpit.rect);
    Assert((&val - &v[0]) == region.getVTBIndex(dpit.rect));
  }
#endif
}
const DPDeckPSCFG::VTB::ValueT&
DPDeckPSCFG::VTB::Datum::
ref(const DPRegion::Region& region, const Rect& rect) const
{
  Assert(region.includes(rect));
  BYRange& byRange = bxRange.getBYRange(rect.bx);
  EXRange& exRange = byRange.getEXRange(rect.by);
  EYRange& eyRange = exRange.getEYRange(rect.ex);
  ValueT& val = eyRange.ref(rect.ey);
  return val;
}
DPDeckPSCFG::VTB::ValueT&
DPDeckPSCFG::VTB::Datum::
ref(const DPRegion::Region& region, const Rect& rect)
{
  Assert(region.includes(rect));
  BYRange& byRange = bxRange.getBYRange(rect.bx);
  EXRange& exRange = byRange.getEXRange(rect.by);
  EYRange& eyRange = exRange.getEYRange(rect.ex);
  ValueT& val = eyRange.ref(rect.ey);
  return val;
}
int
DPDeckPSCFG::VTBBF::
computeSizeInBytes(const DPRegion::Region& region)
{
  return region.getVSTSize() * sizeof(TBBFNode);
}
void
DPDeckPSCFG::VTBBF::
resize(const DPRegion::Region& region)
{
  data[indexOf(region)].resize(region.getVSTSize());
  data[indexOf(region)].fill(ValueT::null());
}

int
DPDeckPSCFG::VSFBWB::
computeSizeInBytes(const DPRegion::Region& region)
{
  return (sizeof(ValueT) *
	  max(region.getVSFBWBSizeI(),
	      (region.isTriShape() ? 0 : region.getVSFBWBSizeO())));
}
void
DPDeckPSCFG::VSFBWB::
resizeI(const DPRegion::Region& region)
{
  data[indexOf(region)].resize(region.getVSFBWBSizeI());
  data[indexOf(region)].fill(ValueT::null());
}
void
DPDeckPSCFG::VSFBWB::
resizeO(const DPRegion::Region& region)
{
  const int& i = indexOf(region);
  data[i].resize(region.getVSFBWBSizeO());
  data[i].fill(ValueT::null());
}
int
DPDeckPSCFG::VSFBWE::
computeSizeInBytes(const DPRegion::Region& region)
{
  return (sizeof(ValueT) *
	  max(region.getVSFBWESizeI(),
	      (region.isTriShape() ? 0 : region.getVSFBWESizeO())));
}
void
DPDeckPSCFG::VSFBWE::
resizeI(const DPRegion::Region& region)
{
  const int& i = indexOf(region);
  data[i].resize(region.getVSFBWESizeI());
  data[i].fill(ValueT::null());
}
void
DPDeckPSCFG::VSFBWE::
resizeO(const DPRegion::Region& region)
{
  const int& i = indexOf(region);
  data[i].resize(region.getVSFBWESizeO());
  data[i].fill(ValueT::null());
}
int
DPDeckPSCFG::VFBW::
computeSizeInBytes(const DPRegion::Region& region)
{
  return region.getVTBSize() * sizeof(ValueT);
}
void
DPDeckPSCFG::VFBW::
resize(const DPRegion::Region& region)
{
  data[indexOf(region)].resize(region);
}
void
DPDeckPSCFG::VFBW::Datum::
resize(const DPRegion::Region& region)
{
  vBYRange.resize(region.srb.rect0.lengthX());
  vEXRange.resize(region.srb.getArea());
  vEYRange.resize(region.getVEYRangeSize());
  v.resize(region.getVTBSize());
  v.fill(ValueT::null());

  BYRange* pBYRange = &vBYRange.back();
  EXRange* pEXRange = &vEXRange.back();
  EYRange* pEYRange = &vEYRange[0];
  ValueT*  pVal     = &v[0]; 

  const int& bx_beg  = region.srb.bx0;
  const int& bx_last = region.srb.ex0;
  bxRange.pBYRange = pBYRange;
  bxRange.bx_beg = bx_beg;
  bxRange.bx_last = bx_last;

  for (int bx = bx_last; bx >= bx_beg; bx--) {
    BYRange& byRange = (*pBYRange--);
    const int& by_beg  = region.srb.getBY(bx);
    const int& by_last = region.srb.getEY(bx);
    byRange.pEXRange = pEXRange;
    byRange.by_beg  = by_beg;
    byRange.by_last = by_last;

	
    for (int by = by_last; by >= by_beg; by--) {
      EXRange& exRange = (*pEXRange--);
      const int& ex_beg  = max(bx, region.sre.bx0);
      const int& ex_last = region.sre.ex0;
      exRange.pEYRange = pEYRange;
      exRange.ex_beg  = ex_beg;
      exRange.ex_last = ex_last;

      for (int ex = ex_beg; ex <= ex_last; ex++) {
	EYRange& eyRange = (*pEYRange++);
	const int& ey_beg = max(by, region.sre.getBY(ex));
	const int& ey_last = region.sre.getEY(ex);
	const int& ley = (ey_last - ey_beg + 1);
	eyRange.pVal = pVal;
	eyRange.ey_beg = ey_beg;
	eyRange.ey_last = ey_last;
	pVal += ley;
      }
    }
  }
  Assert((pBYRange + 1) == &vBYRange.front());
  Assert((pEXRange + 1) == &vEXRange.front());
  Assert((pEYRange - 1) == &vEYRange.back());
  Assert((pVal - 1) == &v.back());

#ifdef DEBUG
  DPRegion::Region::Iterator<true> dpit(region);
  while (dpit.getNext()) {
    const ValueT& val = ref(region, dpit.rect);
    Assert((&val - &v[0]) == region.getVTBIndex(dpit.rect));
  }
#endif
}
const DPDeckPSCFG::VFBW::ValueT&
DPDeckPSCFG::VFBW::Datum::
ref(const DPRegion::Region& region, const Rect& rect) const
{
  Assert(region.includes(rect));
  BYRange& byRange = bxRange.getBYRange(rect.bx);
  EXRange& exRange = byRange.getEXRange(rect.by);
  EYRange& eyRange = exRange.getEYRange(rect.ex);
  ValueT& val = eyRange.ref(rect.ey);
  return val;
}
DPDeckPSCFG::VFBW::ValueT&
DPDeckPSCFG::VFBW::Datum::
ref(const DPRegion::Region& region, const Rect& rect)
{
  Assert(region.includes(rect));
  BYRange& byRange = bxRange.getBYRange(rect.bx);
  EXRange& exRange = byRange.getEXRange(rect.by);
  EYRange& eyRange = exRange.getEYRange(rect.ex);
  ValueT& val = eyRange.ref(rect.ey);
  return val;
}

#endif
