/*
 * PSCFG_Base.hpp
 *
 */
#ifndef PSCFG_BASE_HPP
#define PSCFG_BASE_HPP

#include "Rect.hpp"
class PSCFG_Base {
public:
  enum State {
    ST_st = 0,
    MP_st = 1, 
    MR_st = 2,
    IR_st = 3,
    DR_st = 4,
    ML_st = 5,
    IL_st = 6,
    DL_st = 7,

    BF_st = 8,

    BEG_st = 9,
    END_st = 10,
    X_st   = 11,

    NSTATE  = 9,
    NSTATE0 = 12,
    NEMIT_STATE = 7
  };

#define ST ST_st
#define MP MP_st
#define MR MR_st
#define IR IR_st
#define DR DR_st
#define ML ML_st
#define IL IL_st
#define DL DL_st
#define BF BF_st
#define BG BEG_st
#define EN END_st
#define XX X_st

  enum {
    NLOOP = 1
  };
  static int getNLoopType() {
    return NLOOP;
  }
  static const int& getLoopTypeIndex(const State& s) {
    static const CArray<int, NSTATE0> table = {{
      (-1),
      (-1), 
      0, 
      0,
      0,
      0,
      0,
      0,

      (-1),

      (-1),
      (-1),
      (-1)
    }};
    return table[s];
  }
  static bool isEmissionState(const State& s) {
    static const Rect rect0 = {{{0, 0}}, {{0, 0}}};
    return (getNEmit(s) == rect0);
  }
  static bool isXXState(const State& s) {
    return (s == X_st);
  }
  static State beginState() {return ST_st;}
  static State lastState() {return BF_st;}
  static bool isNormalState(const int& s) {
    return (beginState() <= s && s <= lastState());
  }
  static bool isValidState(const int& s) {
    return (beginState() <= s && s <= X_st);
  }
  inline static int getNEmitState() {
    return NEMIT_STATE;
  };
  inline static State getEmitState(const int& i) {
    Assert(0 <= i && i < NEMIT_STATE);
    return State(MP_st + i); // emission states should be consective
  }
  inline static bool isEmitState(const State& s) {
    return (MP_st <= s && s < (MP_st + NEMIT_STATE));
  }
  static bool bifurcates(const State& s) {
    return (s == BF_st);
  }
  inline static const PSCFG_Base::State&
  toPSCFGBaseState(const State& s) {return s;}
  inline static const State& 
  getStateForBase(const PSCFG_Base::State& s, const int& i)
  {
    return s;
  }
  inline static int
  getNStateForBase(const PSCFG_Base::State& s) {return 1;}

  static const char* state2str(const State& s) {
    static char names[NSTATE0+1][10] = {
      "ST",
      "MP",
      "MR",
      "IR",
      "DR",
      "ML",
      "IL",
      "DL",
      "BF",
      "BEG",
      "END",
      "XX",
      "??"
    };
    return names[isValidState(s) ? s : NSTATE0];
  }
  inline static const Rect& getNEmit(const State& s) {
    static const Rect nEmit[NSTATE0] = {
      {{{0, 0}},{{0, 0}}},//ST
      {{{1, 1}},{{1, 1}}},//MP
      {{{0, 0}},{{1, 1}}},//MR
      {{{0, 0}},{{0, 1}}},//IR
      {{{0, 0}},{{1, 0}}},//DR
      {{{1, 1}},{{0, 0}}},//ML
      {{{0, 1}},{{0, 0}}},//IL
      {{{1, 0}},{{0, 0}}},//DL
      {{{0, 0}},{{0, 0}}},//BF
      {{{0, 0}},{{0, 0}}},//BG
      {{{0, 0}},{{0, 0}}},//EN
      {{{0, 0}},{{0, 0}}} //XX
    };
    return nEmit[s];
  }
  inline static Rect addNEmit(const Rect& rect, const State& s) {
    const Rect& d = getNEmit(s);
    Rect rect1 = {
      {{rect.bx + d.bx, rect.by + d.by}},
      {{rect.ex - d.ex, rect.ey - d.ey}}
    };
    return rect1;
  }
  inline static Rect subtractNEmit(const Rect& rect, const State& s) {
    const Rect& d = getNEmit(s);
    Rect rect1 = {
      {{rect.bx - d.bx, rect.by - d.by}},
      {{rect.ex + d.ex, rect.ey + d.ey}}
    };
    return rect1;
  }
  static bool emitB(const State& s) {
    static const Point pt0 = {0, 0};
    return (!(getNEmit(s).b == pt0));
  }
  static bool emitE(const State& s) {
    static const Point pt0 = {0, 0};
    return (!(getNEmit(s).e == pt0));
  }
  static bool emitBX(const State& s) {
    return (!(getNEmit(s).bx == 0));
  }
  static bool emitEX(const State& s) {
    return (!(getNEmit(s).ex == 0));
  }
  static bool isChild(const State& s, const State& s1) {
    static const bool is_child[NSTATE0][NSTATE0] = {
      //ST MP MR IR DR ML IL DL BF BG EN XX
      {  0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0},//<-ST
      {  0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0},//<-MP
      {  0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0},//<-MR
      {  0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0},//<-IR
      {  0, 1, 1, 0, 1, 1, 1, 1, 1, 0, 0, 0},//<-DR
      {  0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0},//<-ML
      {  0, 1, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0},//<-IL
      {  0, 1, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0},//<-DL
      {  1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},//<-BF
      {  1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},//<-BG
      {  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},//<-EN
      {  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},//<-XX
    };
    return is_child[s][s1];
  }
  static bool isParent(const State& s, const State& s1) {
    return isChild(s1, s);
  }
  inline static const int& getNChild(const State& s) {
    static const int nChild[NSTATE0] = {
      8,//ST->MP,MR,IR,DR,ML,IL,DL,BF
      8,//MP->MP,MR,IR,DR,ML,IL,DL,BF
      8,//MR->MP,MR,IR,DR,ML,IL,DL,BF
      7,//IR->MP,MR,IR,ML,IL,DL,BF
      7,//DR->MP,MR,DR,ML,IL,DL,BF
      5,//ML->MP,ML,IL,DL,BF
      4,//IL->MP,ML,IL,BF
      4,//DL->MP,ML,DL,BF
      1,//BF->ST
      1,//BG->ST
      0,//EN->
      0 //XX->                         
    };
    return nChild[s];
  }
  enum {
    NESC0 = NSTATE,
    NTSC0 = (8+//ST->MP,MR,IR,DR,ML,IL,DL,BF  
	     8+//MP->MP,MR,IR,DR,ML,IL,DL,BF  
	     8+//MR->MP,MR,IR,DR,ML,IL,DL,BF  
	     7+//IR->MP,MR,IR,ML,IL,DL,BF     
	     7+//DR->MP,MR,DR,ML,IL,DL,BF     
	     5+//ML->MP,ML,IL,DL,BF           
	     4+//IL->MP,ML,IL,BF              
	     4+//DL->MP,ML,DL,BF
            
	     1 //BF->ST
                      
	     //1+//BG->ST                       
	     //0+//EN->                         
	     //0 //XX->                         
	     )
    // = 52
  };
  static int getNESC0() {
    return NESC0;
  }
  static int getESC0Index(const State& s) {
    return s;
  }
   static int getNTSC0() {
    return NTSC0;
  }
  static  int getTSC0Index(const State& s, const State& s1) {
    // should have accumulated number of NChildren
    int n = 0;
    for (int i = 0; i < s; i++) {
      n += getNChild(State(i));
    }
    n += getChildIndex(s, s1);
    return n;
  }
  inline static const State& getChild(const State& s, const int& i) {
    static const State children[NSTATE0][9] = {
      {MP,MR,IR,DR,ML,IL,DL,BF,XX},//<-ST
      {MP,MR,IR,DR,ML,IL,DL,BF,XX},//<-MP
      {MP,MR,IR,DR,ML,IL,DL,BF,XX},//<-MR
      {MP,MR,IR,ML,IL,DL,BF,XX,XX},//<-IR
      {MP,MR,DR,ML,IL,DL,BF,XX,XX},//<-DR
      {MP,ML,IL,DL,BF,XX,XX,XX,XX},//<-ML
      {MP,ML,IL,BF,XX,XX,XX,XX,XX},//<-IL
      {MP,ML,DL,BF,XX,XX,XX,XX,XX},//<-DL
      {ST,XX,XX,XX,XX,XX,XX,XX,XX},//<-BF
      {ST,XX,XX,XX,XX,XX,XX,XX,XX},//<-BG
      {XX,XX,XX,XX,XX,XX,XX,XX,XX},//<-EN
      {XX,XX,XX,XX,XX,XX,XX,XX,XX},//<-XX
    };
    return children[s][i];
  }
  inline static const signed char&
  getChildIndex(const State& s, const State& s1){
    static const struct ChildIndex {
      typedef PSCFG_Base Model;
      typedef PSCFG_Base::State State;
      enum {
	LENGTH = Model::NSTATE0,
	//SIZE = (LENGTH * LENGTH),
	NULL_INDEX = (-1)
      };
      signed char data[LENGTH][LENGTH];// index is upto 127

      ChildIndex() {
	for (int i = 0; i < LENGTH; i++) {
	  for (int j = 0; j < LENGTH; j++) {
	    data[i][j] = NULL_INDEX;
	  }
	}
	for (int i = 0; i < LENGTH; i++) {
	  const State& s = (State)i;
	  int k = 0;
	  for (int j = 0; j < Model::getNChild(s); j++) {
	    const State& s1 = Model::getChild(s, j);
	    data[s][s1] = k++;
	  }
	}
      }
      inline const signed char& 
      getVal(const State& s, const State& s1) const {
	return data[s][s1];
      }
    } childIndex;
    return childIndex.getVal(s, s1);
  }
  inline static const int& getNParent(const State& s) {
    static const int nParent[NSTATE0] = {
      1,//ST<-BF
      8,//MP<-ST,MP,MR,IR,DR,ML,IL,DL
      5,//MR<-ST,MP,MR,IR,DR
      4,//IR<-ST,MP,MR,IR
      4,//DR<-ST,MP,MR,DR
      8,//ML<-ST,MP,MR,IR,DR,ML,IL,DL
      7,//IL<-ST,MP,MR,IR,DR,ML,IL
      7,//DL<-ST,MP,MR,IR,DR,ML,DL
      8,//BF<-ST,MP,MR,IR,DR,ML,IL,DL
      0,//BG<-
      1,//EN<-MR
      0 //XX<-
    };
    return nParent[s];
  }
  inline static const State& getParent(const State& s, const int& i) {
    static const State parents[NSTATE0][9] = {
      {BF,XX,XX,XX,XX,XX,XX,XX,XX},//->ST
      {ST,MP,MR,IR,DR,ML,IL,DL,XX},//->MP
      {ST,MP,MR,IR,DR,XX,XX,XX,XX},//->MR
      {ST,MP,MR,IR,XX,XX,XX,XX,XX},//->IR
      {ST,MP,MR,DR,XX,XX,XX,XX,XX},//->DR
      {ST,MP,MR,IR,DR,ML,IL,DL,XX},//->ML
      {ST,MP,MR,IR,DR,ML,IL,XX,XX},//->IL
      {ST,MP,MR,IR,DR,ML,DL,XX,XX},//->DL
      {ST,MP,MR,IR,DR,ML,IL,DL,XX},//->BF
      {XX,XX,XX,XX,XX,XX,XX,XX,XX},//->BG
      {MR,XX,XX,XX,XX,XX,XX,XX,XX},//->EN
      {XX,XX,XX,XX,XX,XX,XX,XX,XX} //->XX
    };
    return parents[s][i];
  }
  class Iterator;
  class TBNode;
  class ScoreManager;

#undef ST 
#undef MP 
#undef MR 
#undef IR 
#undef DR 
#undef ML 
#undef IL 
#undef DL 
#undef BF 
#undef BG 
#undef EN 
#undef XX 
};

class PSCFG_Base::Iterator {
public:
  State s;
  Iterator() : s(X_st), next_s(beginState()) {}
  void reset() {s = X_st; next_s = beginState();}
  bool getNext() {
    if (lastState() < next_s) return false;
  
    s = State(next_s++);
    return true;
  }
private:
  int next_s;
};

struct PSCFG_Base::TBNode {//3byte data
  /* memory configulation
     struct {
     unsigned int IL : 2;
     unsigned int MP : 3;
     unsigned int ST : 3;

     unsigned int DL : 2;
     unsigned int IR : 3;
     unsigned int MR : 3;

     unsigned int altered_flag : 1;
     unsigned int end_flag : 1;
     unsigned int ML : 3;
     unsigned int DR : 3;
     };
  */
  unsigned char data[3];
#ifdef DEBUG
  State isSet[NSTATE];
#endif

  static const TBNode& null() {
    static const TBNode nullNode = {
      {0, 0, 0}
#ifdef DEBUG
      , {X_st, X_st, X_st, X_st, X_st, X_st, X_st, X_st, X_st}
#endif
    };
    return nullNode;
  }
  bool isNull() const {return (*this == null());}
  void setNull() {*this = null();}
  bool operator==(const TBNode& tbn) const {
    if (!(data[0] == tbn.data[0])) return false;
#ifdef DEBUG
    if (!equal(&isSet[0], &isSet[0]+NSTATE, &tbn.isSet[0]))
      return false;
#endif
    return true;
  }
  bool equals(const TBNode& tau) const;
  //static const unsigned char FLAG_ALTERED = (1 << 6);
  //static const unsigned char FLAG_INVALID = (1 << 7);
  //void setAlteredBit() {data[2] |= FLAG_ALTERED;}
  //void unsetAlteredBit() {data[2] &= ~FLAG_ALTERED;}
  //bool isAltered() const {return data[2] & FLAG_ALTERED;}
  struct Conv {
    unsigned int index; //data[index=0..2]
    unsigned int shift;
    unsigned int mask_negate;
    unsigned int mask;
  };
  inline static const Conv& getConv(const State& s) {
    static const CArray<Conv, NSTATE> conv = {{
      //idx sft,              msk_n         msk
      {   0,  0, ~(~(~0 << 3) << 0), ~(~0 << 3)}, //ST
      {   0,  3, ~(~(~0 << 3) << 3), ~(~0 << 3)}, //MP
      {   1,  0, ~(~(~0 << 3) << 0), ~(~0 << 3)}, //MR
      {   1,  3, ~(~(~0 << 3) << 3), ~(~0 << 3)}, //IR
      {   2,  0, ~(~(~0 << 3) << 0), ~(~0 << 3)}, //DR
      {   2,  3, ~(~(~0 << 3) << 3), ~(~0 << 3)}, //ML
      {   0,  6, ~(~(~0 << 2) << 6), ~(~0 << 2)}, //IL
      {   1,  6, ~(~(~0 << 2) << 6), ~(~0 << 2)}, //DL
      {   2,  6, ~(~(~0 << 2) << 6), ~(~0 << 2)}, //BF
    }};
    return conv[s];
  }
  const State& getVal(const State& s) const {
    Assert(isNormalState(s));
    const Conv& c = getConv(s);
    const unsigned int& d = data[c.index];
    const unsigned int& i = ((d >> c.shift) & c.mask);
    const State& s1 = getChild(s, i);
#ifdef DEBUG
    Assert(isSet[s] != X_st);
    Assert(isSet[s] == s1);
#endif
    return s1;
  }
  inline void setVal(const State& s, const int& childIndex) {
    const Conv& c = getConv(s);
    unsigned char& d = data[c.index];
    d &= c.mask_negate;
    d |= (childIndex << c.shift);

#ifdef DEBUG
    const State s1 = getChild(s, childIndex);
    Assert(isNormalState(s) && s != BF_st);
    Assert(isNormalState(s1));
    Assert(isChild(s, s1));
    isSet[s] = s1;
    Assert(getVal(s) == s1);
#endif
  }
  inline void setVal0(const State& s, const State s1) {
    //assuming data is nulified.
    static const unsigned char X = 0xff;
    static const unsigned char state2code[NSTATE][NSTATE] = {
      //ST,    MP,    MR,    IR,    DR,    ML,    IL,    DL,    BF
      {  X,(0<<0),(1<<0),(2<<0),(3<<0),(4<<0),(5<<0),(6<<0),(7<<0)},//ST
      {  X,(0<<3),(1<<3),(2<<3),(3<<3),(4<<3),(5<<3),(6<<3),(7<<3)},//MP
      {  X,(0<<0),(1<<0),(2<<0),(3<<0),(4<<0),(5<<0),(6<<0),(7<<0)},//MR
      {  X,(0<<3),(1<<3),(2<<3),     X,(3<<3),(4<<3),(5<<3),(6<<3)},//IR
      {  X,(0<<0),(1<<0),     X,(2<<0),(3<<0),(4<<0),(5<<0),(6<<0)},//DR
      {  X,(0<<3),     X,     X,     X,(1<<3),(2<<3),(3<<3),(4<<3)},//ML
      {  X,(0<<6),     X,     X,     X,(1<<6),(2<<6),     X,(3<<6)},//IL
      {  X,(0<<6),     X,     X,     X,(1<<6),     X,(2<<6),(3<<6)},//DL
      {  X,     X,     X,     X,     X,     X,     X,     X,     X},//BF
    };
    const Conv& c = getConv(s);
    data[c.index] |= state2code[s][s1];
    unsigned char& d = data[c.index];
    d &= c.mask_negate;
    d |= state2code[s][s1];

#ifdef DEBUG
    Assert(isNormalState(s) && s != BF_st);
    Assert(isNormalState(s1));
    Assert(isChild(s, s1));
    isSet[s] = s1;
    Assert(getVal(s) == s1);
#endif
  }
  //bool isValid() const {
  //  if (data[2] & FLAG_INVALID) return false;
  //
  //  Iterator it;
  //  while (it.getNext()) {
  //    if (getVal(it.s) == X_st) return false;
  //  }
  //  return true;
  //}
//#ifdef DEBUG
//  bool valIsSet(const State& s) const {
//    Assert(isNormalState(s));
//    return (isSet[s] != X_st);
//  }
//  bool isAltered() const {
//    Iterator it;
//    while (it.getNext()) {
//      if (it.s == BF_st) continue;
//
//      if (valIsSet(it.s)) return true;
//    }
//    return false;
//  }
//#endif
  string toString() const {
    ostringstream oss;
    oss << "{ ";
    //if (!isValid()) {
    //  oss << "Invalid: true, ";
    //}
    //if (isAltered()) {
    //  oss << "isAltered: true, ";
    //}
    oss << "data: { ";
    Iterator it;
    while (it.getNext()) {
      if (it.s == BF_st) continue;

      const State& s1 = getVal(it.s);
      oss << PSCFG_Base::state2str(it.s) << ": " 
	  << PSCFG_Base::state2str(s1);
      oss << (it.s == DR_st ? " }" : ", ");
    }
    return oss.str();
  }
  void print() const {
    cout << toString() << flush;
  }
};

#endif
