/*
 * PSCFG_Energy.hpp
 *
 */
#ifndef PSCFG_ENERGY_HPP
#define PSCFG_ENERGY_HPP

//#include "scfg/SCFG_Base.hpp"
//#include "scfg/SCFG_Energy.hpp"
#include "PSCFG_Base.hpp"

////#define USE_PSCFG_BASE
//#define USE_PSCFG_STEMLOC_SUPERPAIR
////#define PSCFG_BASE_HAIRPIN
////#define PSCFG_SHORT_BULGE
//#define SLR_STATE
////#define PSCFG_LOOP_TYPE_DEPENDENT_SCORE
//#define PSCFG_SHORT_BULGE
//#define USE_PSCFG_BASE_NO_ML
#define USE_PSCFG_BASE
#define USE_PSCFG_NUSSINOV

class 
PSCFG_Energy 
{
public:
#if defined(USE_PSCFG_BASE)
  enum State {
    ST_st,         

    MP_st,         
    MR_st,          
    IR_st,          
    DR_st,          
    ML_st,          
    IL_st,          
    DL_st,     
#ifdef PSCFG_BASE_HAIRPIN
    MR_Hairpin_st,          
    IR_Hairpin_st,          
    DR_Hairpin_st,   
#endif
#ifdef PSCFG_SHORT_BULGE
    MR_Bulge1_st,
    ML_Bulge1_st,
#endif   

    BF_st,          

    BEG_st,         
    END_st,         
    X_st,           

    NSTATE         = BEG_st,
    NSTATE0        = (NSTATE + 3),

    BEGIN_STATE    = ST_st,// BEGIN_STATE == 0
    LAST_STATE     = BF_st,// last normal state states values should be continuous
    EMIT_STATE_BEG = MP_st, // emission states should be continuous
    NEMIT_STATE    = (BF_st - ST_st - 1)
  };

#define ST ST_st
#define MP MP_st
#define MR MR_st
#define IR IR_st
#define DR DR_st
#define ML ML_st
#define IL IL_st
#define DL DL_st

#ifdef PSCFG_BASE_HAIRPIN
#define MRH MR_Hairpin_st   
#define IRH IR_Hairpin_st       
#define DRH DR_Hairpin_st
#endif

#ifdef PSCFG_SHORT_BULGE
#define MRB1 MR_Bulge1_st
#define MLB1 ML_Bulge1_st
#endif 
  
#define BF BF_st
#define BG BEG_st
#define EN END_st
#define XX X_st

  enum LoopType {
    LOOP_TYPE_LOOP     = 0,
    NLOOP              = 1,
    INVALID_LOOP_TYPE  = (-1)
  };
  enum {
    MAX_NCHILD          = (
#if defined(USE_PSCFG_NUSSINOV)
			     8
#else
                             8
#ifdef PSCFG_BASE_HAIRPIN
			   + 3
#endif
#ifdef PSCFG_SHORT_BULGE
			   + 2
#endif
#endif
			   ),
    MAX_NPARENT         = (
#if defined(USE_PSCFG_NUSSINOV)
			     7
#else
			     8
#ifdef PSCFG_SHORT_BULGE
			   + 2
#endif
#endif
			   ),
    MAX_NSTATE_FOR_BASE = (1
#ifdef PSCFG_BASE_HAIRPIN
			   +1
#endif
#ifdef PSCFG_SHORT_BULGE
			   +1
#endif
			   ),
    TBNODE_NBYTE        = (
#if defined(USE_PSCFG_NUSSINOV)
			   1
#else
			   4
#ifdef PSCFG_BASE_HAIRPIN
			   +0
#endif
#ifdef PSCFG_SHORT_BULGE
			   +0
#endif
#endif
			  ),
    NESC0               = NSTATE,
    NTSC0               = (
#if defined(USE_PSCFG_NUSSINOV)
			   (16 - 2)
#else
			   (54 - 19)
#ifdef PSCFG_BASE_HAIRPIN
			   + (10 - 5)
#endif
#ifdef PSCFG_SHORT_BULGE
			   + (4 - 3)
#endif
#endif
			   )
  };

#elif defined(USE_PSCFG_BASE_NO_ML)
  enum State {
    ST_st          = 0,

    MP_st          = 1, 
    MR_st          = 2,
    IR_st          = 3,
    DR_st          = 4,

    BF_st          = 5,

    BEG_st         = 6,
    END_st         = 7,
    X_st           = 8,

    NSTATE         = 6,
    NSTATE0        = 9,

    BEGIN_STATE    = ST_st,// BEGIN_STATE == 0
    LAST_STATE     = BF_st,// last normal state states values should be continuous
    EMIT_STATE_BEG = MP_st, // emission states should be continuous
    NEMIT_STATE    = (DR_st - MP_st + 1)
  };

#define ST ST_st
#define MP MP_st
#define MR MR_st
#define IR IR_st
#define DR DR_st
#define BF BF_st
#define BG BEG_st
#define EN END_st
#define XX X_st

  enum LoopType {
    LOOP_TYPE_LOOP     = 0,
    NLOOP              = 1,
    INVALID_LOOP_TYPE  = (-1)
  };
  enum {
    MAX_NCHILD          = 5,
    MAX_NPARENT         = 5,
    MAX_NSTATE_FOR_BASE = 1,
    TBNODE_NBYTE        = 2,
    NESC0               = NSTATE,
    NTSC0               = (26 - 6)
  };

#elif defined(USE_PSCFG_STEMLOC_STEMPAIR)
  enum State {
    ST_st           =  0,

    MP_StemMatLR_st =  1,
    ML_StemMatL_st  =  2,
    MR_StemMatR_st  =  3,
    
    DL_StemDelL_st  =  4,
    DR_StemDelR_st  =  5,
    IL_StemInsL_st  =  6,
    DL_StemInsR_st  =  7,
    
    ML_LoopMat_st   =  8,
    DL_LoopDel_st   =  9,
    IL_LoopIns_st   = 10,

    BF_st           = 11,

    BEG_st          = 12,
    END_st          = 13,
    X_st            = 14,

    NSTATE          = 12,
    NSTATE0         = 15,

    BEGIN_STATE     = ST_st,// BEGIN_STATE == 0
    LAST_STATE      = BF_st,// last normal state states values should be continuous
    EMIT_STATE_BEG  = MP_StemMatLR_st, // emission states should be continuous
    NEMIT_STATE     = 10
  };

#define ST  ST_st
#define MPS MP_StemMatLR_st
#define MLS ML_StemMatL_st 
#define MRS MR_StemMatR_st 
    
#define DLS DL_StemDelL_st 
#define DRS DR_StemDelR_st 
#define ILS IL_StemInsL_st 
#define DRS DL_StemInsR_st 
    
#define MLL ML_LoopMat_st  
#define DLL DL_LoopDel_st  
#define ILL IL_LoopIns_st  
#define BF  BF_st
#define BG  BEG_st
#define EN  END_st
#define XX  X_st

  enum LoopType {
    LOOP_TYPE_STEM     = 0,
    LOOP_TYPE_LOOP     = 1,
    NLOOP              = 2,
    INVALID_LOOP_TYPE  = (-1)
  };
  enum {
    MAX_NCHILD          = 8,
    MAX_NPARENT         = 8,
    MAX_NSTATE_FOR_BASE = 2,
    TBNODE_NBYTE        = 4,
    NESC0               = NSTATE,
    NTSC0               = (70 - 52)
  };

#elif defined(USE_PSCFG_STEMLOC_SUPERPAIR)

  enum State {
#ifdef SLR_STATE
    SL_st             ,
    SR_st             ,
#else
    ST_st             ,
#endif
    MP_Stem_st        ,
#ifdef PSCFG_SHORT_BULGE
    MR_Bulge1_st      ,
    ML_Bulge1_st      ,
#ifdef PSCFG_SHORT_BULGE_INDEL
    IR_Bulge1_st      ,
    DR_Bulge1_st      ,
    IL_Bulge1_st      ,
    DL_Bulge1_st      ,
#endif
#endif   
    ML_LoopMatch_st   ,
    IL_LoopIns_st     ,
    DL_LoopDel_st     ,
    ML_LBulgeMatch_st ,
    IL_LBulgeIns_st   ,
    DL_LBulgeDel_st   ,
    MR_RBulgeMatch_st ,
    IR_RBulgeIns_st   ,
    DR_RBulgeDel_st   ,
    ML_LRBulgeMatch_st,
    IL_LRBulgeIns_st  ,
    DL_LRBulgeDel_st  ,


    BF_st             ,

    BEG_st            ,
    END_st            ,
    X_st              ,

    NSTATE             = (BEG_st),
    NSTATE0            = (X_st + 1),

    BEGIN_STATE        = (
#ifdef SLR_STATE
			  SL_st
#else
			  ST_st
#endif
			  ), // BEGIN_STATE == 0
    LAST_STATE         = BF_st,// last normal state states values should be continuous
    EMIT_STATE_BEG     = MP_Stem_st, // emission states should be continuous
    NEMIT_STATE        = (BF_st - MP_Stem_st)
  };

#ifdef SLR_STATE
#define SL  SL_st
#define SR  SR_st
#else
#define ST  ST_st  
#endif                        
#define MP  MP_Stem_st     
  
#ifdef PSCFG_SHORT_BULGE
#define MRB1 MR_Bulge1_st
#define MLB1 ML_Bulge1_st
#ifdef PSCFG_SHORT_BULGE_INDEL
#define IRB1 IR_Bulge1_st
#define DRB1 DR_Bulge1_st
#define ILB1 IL_Bulge1_st
#define DLB1 DL_Bulge1_st
#endif
#endif 
 
#define MLH ML_LoopMatch_st   
#define ILH IL_LoopIns_st     
#define DLH DL_LoopDel_st     
#define MLB ML_LBulgeMatch_st 
#define ILB IL_LBulgeIns_st   
#define DLB DL_LBulgeDel_st   
#define MRB MR_RBulgeMatch_st 
#define IRB IR_RBulgeIns_st   
#define DRB DR_RBulgeDel_st   
#define MLI ML_LRBulgeMatch_st
#define ILI IL_LRBulgeIns_st  
#define DLI DL_LRBulgeDel_st                 
#define BF  BF_st                              
#define BG  BEG_st            
#define EN  END_st            
#define XX  X_st              

  enum LoopType {
    LOOP_TYPE_HAIRPIN  = 0,
    LOOP_TYPE_BULGE    = 1,
    LOOP_TYPE_INTERIOR = 2,
    NLOOP              = 3,
    INVALID_LOOP_TYPE  = (-1)
  };
  enum {
    MAX_NCHILD          = (14
#ifdef SLR_STATE
			   +1
#endif
#ifdef PSCFG_SHORT_BULGE
			   + 1
#ifdef PSCFG_SHORT_BULGE_INDEL
			   + 4
#endif
#endif
			   ),
    MAX_NPARENT         = (8
                           +1
#ifdef PSCFG_SHORT_BULGE
			   +2
#ifdef PSCFG_SHORT_BULGE_INDEL
			   + 4
#endif
#endif
			   ),
    MAX_NSTATE_FOR_BASE = (3
#ifdef PSCFG_SHORT_BULGE
			   + 1
#endif
#ifdef PSCFG_SHORT_BULGE_INDEL
			   + 0
#endif
			   ),

    TBNODE_NBYTE        = (6
#ifdef SLR_STATE
			   +0
#endif
#ifdef PSCFG_SHORT_BULGE
			   + 0
#ifdef PSCFG_SHORT_BULGE_INDEL
			   + 0
#endif
#endif
			   ),
    NESC0               = NSTATE,
#if 0
    NTSC0               = (85 - 22)
#else
    NTSC0               = ((85 - 54)
#ifdef SLR_STATE
			   + (4 - 1)
#endif
#ifdef PSCFG_LOOP_TYPE_DEPENDENT_SCORE
			   + (34 - 21)
#endif
#ifdef PSCFG_SHORT_BULGE
			   + (4 - 3)
#ifdef PSCFG_SHORT_BULGE_INDEL
			   + (17 - 15)
#endif
#endif
			   )
#endif
  };

#else

  enum State {
    ST_st = 0,

    MP_st = 1, 

    MR_Exterior_st = 2,
    IR_Exterior_st = 3,
    DR_Exterior_st = 4,

    MR_Hairpin_st = 5,
    IR_Hairpin_st = 6,
    DR_Hairpin_st = 7,

    MR_Bulge_st = 8,
    IR_Bulge_st = 9,
    DR_Bulge_st = 10,
    ML_Bulge_st = 11,
    IL_Bulge_st = 12,
    DL_Bulge_st = 13,

    MR_Interior_st = 14,
    IR_Interior_st = 15,
    DR_Interior_st = 16,
    ML_Interior_st = 17,
    IL_Interior_st = 18,
    DL_Interior_st = 19,

    MR_Multi1_st = 20,
    IR_Multi1_st = 21,
    DR_Multi1_st = 22,
    ML_Multi1_st = 23,
    IL_Multi1_st = 24,
    DL_Multi1_st = 25,

    MR_Multi_st = 26,
    IR_Multi_st = 27,
    DR_Multi_st = 28,
    ML_Multi_st = 29,
    IL_Multi_st = 30,
    DL_Multi_st = 31,

    BF_st = 32,

    BEG_st = 33,
    END_st = 34,
    X_st = 35,

    NSTATE = 33,
    NSTATE0 = 36,
    NEMIT_STATE = 31
  };

#define ST ST_st
#define MP MP_st

#define MRE MR_Exterior_st
#define IRE IR_Exterior_st
#define DRE DR_Exterior_st

#define MRH MR_Hairpin_st
#define IRH IR_Hairpin_st
#define DRH DR_Hairpin_st

#define MRB MR_Bulge_st
#define IRB IR_Bulge_st
#define DRB DR_Bulge_st
#define MLB ML_Bulge_st
#define ILB IL_Bulge_st
#define DLB DL_Bulge_st

#define MRI MR_Interior_st
#define IRI IR_Interior_st
#define DRI DR_Interior_st
#define MLI ML_Interior_st
#define ILI IL_Interior_st
#define DLI DL_Interior_st

#define MR1 MR_Multi1_st
#define IR1 IR_Multi1_st
#define DR1 DR_Multi1_st
#define ML1 ML_Multi1_st
#define IL1 IL_Multi1_st
#define DL1 DL_Multi1_st

#define MRM MR_Multi_st
#define IRM IR_Multi_st
#define DRM DR_Multi_st
#define MLM ML_Multi_st
#define ILM IL_Multi_st
#define DLM DL_Multi_st

#define BF BF_st

#define BG BEG_st
#define EN END_st
#define XX X_st

  typedef PSCFG_Base EmitModel;
  typedef PSCFG_Base::State EmitState;
  enum LoopType {
    LOOP_TYPE_HAIRPIN  = 0,
    LOOP_TYPE_BULGE    = 1,
    LOOP_TYPE_INTERIOR = 2,
    LOOP_TYPE_MULTI    = 3,
    NLOOP              = 4,
    INVALID_LOOP_TYPE  = (-1)
  };
  enum {
    MAX_NCHILD = 20,
    MAX_NPARENT = 17,
    MAX_NSTATE_FOR_BASE = 6
    TBNODE_NBYTE        = 10,
    NESC0               = NSTATE,
    NTSC0               = 153
  };
#endif

  struct Datum {
    typedef CArray<State, (NSTATE0 + 1)> Children;
    State state;
    const char* name;
    int loopTypeIndex;
    PSCFG_Base::State baseState;
    Children children;
  };
  typedef PSCFG_Energy Model;
  typedef CArray<Datum,NSTATE0> Data;
  PSCFG_Energy() {
    typedef PSCFG_Base B;
    Data data = {{
#define Pair(state) state, #state
 
#if defined(USE_PSCFG_BASE)
#if defined(USE_PSCFG_NUSSINOV)
      {Pair(ST),-1,B::ST_st ,{{MP,MR,IR,DR,ML,IL,DL,BF,EN,XX}}},
      {Pair(MP),-1,B::MP_st ,{{ST,XX}}},
      {Pair(MR), 0,B::MR_st ,{{ST,XX}}},
      {Pair(IR), 0,B::IR_st ,{{ST,XX}}},
      {Pair(DR), 0,B::DR_st ,{{ST,XX}}},
      {Pair(ML), 0,B::ML_st ,{{ST,XX}}},
      {Pair(IL), 0,B::IL_st ,{{ST,XX}}},
      {Pair(DL), 0,B::DL_st ,{{ST,XX}}},
      {Pair(BF),-1,B::BF_st ,{{XX}}},
      {Pair(BG),-1,B::BEG_st,{{ST,XX}}},
      {Pair(EN),-1,B::END_st,{{XX}}},
      {Pair(XX),-1,B::X_st  ,{{XX}}}
#else
      {Pair(ST),-1,B::ST_st ,{{MP,MR,IR,DR,ML,IL,DL,BF,XX}}},
      {Pair(MP),-1,B::MP_st ,{{MP,MR,IR,DR,ML,IL,DL,
#ifdef PSCFG_BASE_HAIRPIN
			       MRH,IRH,DRH,
#endif
#ifdef PSCFG_SHORT_BULGE
			       MRB1,MLB1,
#endif
			       BF,XX}}},
      {Pair(MR), 0,B::MR_st ,{{MP,MR,IR,DR,ML,IL,DL,BF,
#ifndef PSCFG_BASE_HAIRPIN
			       EN,
#endif
			       XX}}},
      {Pair(IR), 0,B::IR_st ,{{MP,MR,IR,ML,IL,DL,BF,
#ifndef PSCFG_BASE_HAIRPIN
			       EN,
#endif
			       XX}}},
      {Pair(DR), 0,B::DR_st ,{{MP,MR,DR,ML,IL,DL,BF,
#ifndef PSCFG_BASE_HAIRPIN
			       EN,
#endif
			       XX}}},
      {Pair(ML), 0,B::ML_st ,{{MP,ML,IL,DL,BF,XX}}},
      {Pair(IL), 0,B::IL_st ,{{MP,ML,IL,BF,XX}}},
      {Pair(DL), 0,B::DL_st ,{{MP,ML,DL,BF,XX}}},
#ifdef PSCFG_BASE_HAIRPIN
      {Pair(MRH), 0,B::MR_st ,{{MRH,IRH,DRH,EN,XX}}},
      {Pair(IRH), 0,B::IR_st ,{{MRH,IRH,EN,XX}}},
      {Pair(DRH), 0,B::DR_st ,{{MRH,DRH,EN,XX}}},
#endif
#ifdef PSCFG_SHORT_BULGE
      {Pair(MRB1), 0,B::MR_st ,{{MP,XX}}},
      {Pair(MLB1), 0,B::ML_st ,{{MP,XX}}},
#endif      
      {Pair(BF),-1,B::BF_st ,{{XX}}},
      {Pair(BG),-1,B::BEG_st,{{ST,
#ifdef PSCFG_BASE_HAIRPIN
			       MRH,IRH,DRH,
#endif
			       XX}}},
      {Pair(EN),-1,B::END_st,{{XX}}},
      {Pair(XX),-1,B::X_st  ,{{XX}}}
#endif
#elif defined(USE_PSCFG_BASE_NO_ML)
      {Pair(ST),-1,B::ST_st ,{{MP,MR,IR,DR,BF,XX}}},
      {Pair(MP),-1,B::MP_st ,{{MP,MR,IR,DR,BF,XX}}},
      {Pair(MR), 0,B::MR_st ,{{MP,MR,IR,DR,BF,EN,XX}}},
      {Pair(IR), 0,B::IR_st ,{{MP,MR,IR,BF,EN,XX}}},
      {Pair(DR), 0,B::DR_st ,{{MP,MR,DR,BF,EN,XX}}},
      {Pair(BF),-1,B::BF_st ,{{XX}}},
      {Pair(BG),-1,B::BEG_st,{{ST,XX}}},
      {Pair(EN),-1,B::END_st,{{XX}}},
      {Pair(XX),-1,B::X_st  ,{{XX}}}
#elif defined(USE_PSCFG_STEMLOC_STEMPAIR)
      {Pair(ST ),-1,B::ST_st ,{{MPS,MLS,MRS,DLS,DRS,ILS,IRS,BF,XX}}},
      {Pair(MPS), 0,B::MP_st ,{{MPS,MLS,MRS,DLS,DRS,ILS,IRS,MLL,DLL,ILL,BF,XX}}},
      {Pair(MLS), 0,B::ML_st ,{{MPS,MLS,MRS,DLS,DRS,ILS,IRS,XX}}},
      {Pair(MRS), 0,B::MR_st ,{{MPS,MLS,MRS,DLS,DRS,ILS,IRS,XX}}},   
      {Pair(DLS), 0,B::DL_st ,{{MPS,MLS,MRS,DLS,DRS,ILS,IRS,XX}}},
      {Pair(DRS), 0,B::DR_st ,{{MPS,MLS,MRS,DLS,DRS,ILS,IRS,XX}}},
      {Pair(ILS), 0,B::IL_st ,{{MPS,MLS,MRS,DLS,DRS,ILS,IRS,XX}}},
      {Pair(IRS), 0,B::DR_st ,{{MPS,MLS,MRS,DLS,DRS,ILS,IRS,XX}}},  
      {Pair(MLL), 1,B::ML_st ,{{MLL,DLL,ILL, EN}}},
      {Pair(DLL), 1,B::DL_st ,{{MLL,DLL,ILL, EN}}},
      {Pair(ILL), 1,B::IL_st ,{{MLL,DLL,ILL, EN}}},
      {Pair(BF ),-1,B::BF_st ,{{ XX}}},
      {Pair(BG ),-1,B::BEG_st,{{MPS,MLS,MRS,DLS,DRS,ILS,DRS,MLL,DLL,ILL,BF}}},
      {Pair(EN ),-1,B::END_st,{{ XX}}},
      {Pair(XX ),-1,B::X_st  ,{{ XX}}}
#elif defined(USE_PSCFG_STEMLOC_SUPERPAIR)
#ifdef SLR_STATE
      {Pair(SL ),-1,B::ST_st ,{{ MP,MLB,ILB,DLB, XX}}},
      {Pair(SR ),-1,B::ST_st ,{{ MP,MLB,ILB,DLB,MRB,IRB,DRB,MLI,ILI,DLI, BF, XX}}},
#else
      {Pair(ST ),-1,B::ST_st ,{{ MP,MLB,ILB,DLB,MRB,IRB,DRB,MLI,ILI,DLI, BF, XX}}},
#endif
      {Pair(MP ),-1,B::MP_st ,{{ MP,
#ifdef PSCFG_SHORT_BULGE
				 MRB1,MLB1,
#ifdef PSCFG_SHORT_BULGE_INDEL
				 IRB1,DRB1,ILB1,DLB1,
#endif
#endif
				 MLH,ILH,DLH,MLB,ILB,DLB,MRB,IRB,DRB,MLI,ILI,DLI, BF, XX}}},
#ifdef PSCFG_SHORT_BULGE
      {Pair(MRB1), 1,B::MR_st ,{{MP,
#ifdef PSCFG_SHORT_BULGE_INDEL
				 MLB1,ILB1,DLB1,
#endif
				 XX}}},
      {Pair(MLB1), 1,B::ML_st ,{{MP,XX}}},
#ifdef PSCFG_SHORT_BULGE_INDEL
      {Pair(IRB1), 1,B::IR_st ,{{MP,MLB1,ILB1,DLB1,XX}}},
      {Pair(DRB1), 1,B::DR_st ,{{MP,MLB1,ILB1,DLB1,XX}}},
      {Pair(ILB1), 1,B::IL_st ,{{MP,XX}}},
      {Pair(DLB1), 1,B::DL_st ,{{MP,XX}}},
#endif
#endif      
      {Pair(MLH), 0,B::ML_st ,{{MLH,ILH,DLH, EN, XX}}},
      {Pair(ILH), 0,B::IL_st ,{{MLH,ILH,DLH, EN, XX}}},
      {Pair(DLH), 0,B::DL_st ,{{MLH,ILH,DLH, EN, XX}}},
      {Pair(MLB), 1,B::ML_st ,{{ MP,MLB,ILB,DLB, XX}}},
      {Pair(ILB), 1,B::IL_st ,{{ MP,MLB,ILB,DLB, XX}}},
      {Pair(DLB), 1,B::DL_st ,{{ MP,MLB,ILB,DLB, XX}}},
      {Pair(MRB), 1,B::MR_st ,{{ MP,MRB,IRB,DRB, BF, XX}}},
      {Pair(IRB), 1,B::IR_st ,{{ MP,MRB,IRB,DRB, BF, XX}}},
      {Pair(DRB), 1,B::DR_st ,{{ MP,MRB,IRB,DRB, BF, XX}}},
      {Pair(MLI), 2,B::ML_st ,{{ MRB,IRB,DRB,MLI,ILI,DLI, BF, XX}}},
      {Pair(ILI), 2,B::IL_st ,{{ MRB,IRB,DRB,MLI,ILI,DLI, BF, XX}}},
      {Pair(DLI), 2,B::DL_st ,{{ MRB,IRB,DRB,MLI,ILI,DLI, BF, XX}}},
      {Pair(BF ),-1,B::BF_st ,{{ XX}}},
      //{Pair(BG ),-1,B::BEG_st,{{ ST, MP,MLH,ILH,DLH,MLB,ILB,DLB,MRB,IRB,DRB,MLI,ILI,DLI, XX}}},
#ifdef SLR_STATE
      // because of ST->BF transition,
      // BEG->ST and BEG->BF cannot co-exist unambigously
      // tentatively disallow BEG->BF transition due to complication
      {Pair(BG ),-1,B::BEG_st,{{SL, SR, MP,MLH,ILH,DLH,MLB,ILB,DLB,MRB,IRB,DRB,MLI,ILI,DLI, XX}}},
#else
      {Pair(BG ),-1,B::BEG_st,{{ ST,MLH,ILH,DLH, XX}}},
#endif
      {Pair(EN ),-1,B::END_st,{{ XX}}},
      {Pair(XX ),-1,B::X_st  ,{{ XX}}},
#else
      {Pair(ST_st         ),-1,B::ST_st ,
       {{ MP,MR1,IR1,DR1,ML1,IL1,DL1,MRM,IRM,DRM,MLM,ILM,DLM, BF, XX}}},
      {Pair(MP_st  	  ),-1,B::MP_st ,
       {{ MP,MRH,IRH,DRH,MRB,IRB,DRB,MLB,ILB,DLB,MRI,IRI,DRI,MRM,IRM,DRM,MLM,ILM,DLM, BF, XX}}},
      {Pair(MR_Exterior_st), 3,B::MR_st ,{{MRE,IRE,DRE, EN, XX}}}, 
      {Pair(IR_Exterior_st), 3,B::IR_st ,{{MRE,IRE, XX}}}, 
      {Pair(DR_Exterior_st), 3,B::DR_st ,{{MRE,DRE, XX}}}, 
      {Pair(MR_Hairpin_st ), 0,B::MR_st ,{{MRH,IRH,DRH, EN, XX}}},
      {Pair(IR_Hairpin_st ), 0,B::IR_st ,{{MRH,IRH, XX}}},
      {Pair(DR_Hairpin_st ), 0,B::DR_st ,{{MRH,DRH, XX}}},
      {Pair(MR_Bulge_st   ), 1,B::MR_st ,{{ MP,MRB,IRB,DRB, XX}}},
      {Pair(IR_Bulge_st   ), 1,B::IR_st ,{{ MP,MRB,IRB, XX}}},
      {Pair(DR_Bulge_st   ), 1,B::DR_st ,{{ MP,MRB,DRB, XX}}},
      {Pair(ML_Bulge_st   ), 1,B::ML_st ,{{ MP,MLB,ILB,DLB, XX}}},
      {Pair(IL_Bulge_st   ), 1,B::IL_st ,{{ MP,MLB,ILB, XX}}},
      {Pair(DL_Bulge_st   ), 1,B::DL_st ,{{ MP,MLB,DLB, XX}}},
      {Pair(MR_Interior_st), 2,B::MR_st ,{{MRI,IRI,DRI,MLI,ILI,DLI, XX}}}, 
      {Pair(IR_Interior_st), 2,B::IR_st ,{{MRI,IRI,MLI,ILI,DLI, XX}}}, 
      {Pair(DR_Interior_st), 2,B::DR_st ,{{MRI,DRI,MLI,ILI,DLI, XX}}}, 
      {Pair(ML_Interior_st), 2,B::ML_st ,{{ MP,MLI,ILI,DLI, XX}}}, 
      {Pair(IL_Interior_st), 2,B::IL_st ,{{ MP,MLI,ILI, XX}}}, 
      {Pair(DL_Interior_st), 2,B::DL_st ,{{ MP,MLI,DLI, XX}}}, 
      {Pair(MR_Multi1_st  ), 3,B::MR_st ,{{ MP,MR1,IR1,DR1,ML1,IL1,DL1, XX}}},
      {Pair(IR_Multi1_st  ), 3,B::IR_st ,{{ MP,MR1,IR1,ML1,IL1,DL1, XX}}},
      {Pair(DR_Multi1_st  ), 3,B::DR_st ,{{ MP,MR1,DR1,ML1,IL1,DL1, XX}}},
      {Pair(ML_Multi1_st  ), 3,B::ML_st ,{{ MP,ML1,IL1,DL1, XX}}},
      {Pair(IL_Multi1_st  ), 3,B::IL_st ,{{ MP,ML1,IL1, XX}}},
      {Pair(DL_Multi1_st  ), 3,B::DL_st ,{{ MP,ML1,DL1, XX}}},
      {Pair(MR_Multi_st   ), 3,B::MR_st ,{{MRM,IRM,DRM,MLM,ILM,DLM, BF, XX}}},
      {Pair(IR_Multi_st   ), 3,B::IR_st ,{{MRM,IRM,MLM,ILM,DLM, BF, XX}}},
      {Pair(DR_Multi_st   ), 3,B::DR_st ,{{MRM,DRM,MLM,ILM,DLM, BF, XX}}},
      {Pair(ML_Multi_st   ), 3,B::ML_st ,{{MLM,ILM,DLM, BF, XX}}},
      {Pair(IL_Multi_st   ), 3,B::IL_st ,{{MLM,ILM, BF, XX}}},
      {Pair(DL_Multi_st   ), 3,B::DL_st ,{{MLM,DLM, BF, XX}}},
      {Pair(BF_st	  ),-1,B::BF_st ,{{XX}}},
      {Pair(BEG_st 	  ),-1,B::BEG_st,{{ST,MRE,IRE,DRE, XX}}},
      {Pair(END_st 	  ),-1,B::END_st,{{XX}}},
      {Pair(X_st          ),-1,B::X_st  ,{{XX}}}, 

#endif
#undef Pair
    }};

    CArray<State,2> tsc_constraints[] = {
#if defined(USE_PSCFG_BASE)
#if defined(USE_PSCFG_NUSSINOV)
      {{ST,IR}},{{ST,DR}},{{XX,XX}},
      {{ST,IL}},{{ST,DL}},{{XX,XX}}
#else
#ifdef PSCFG_BASE_HAIRPIN
      {{MP,DRH}},{{MP,IRH}},{{XX,XX}},
      {{MRH,DRH}},{{MRH,IRH}},{{DRH,MRH}},{{IRH,MRH}},{{XX,XX}},
      {{DRH,DRH}},{{IRH,IRH}},{{XX,XX}},
#endif
#ifdef PSCFG_SHORT_BULGE
      {{MP,MLB1}},{{MP,MRB1}},{{MLB1,MP}},{{MRB1,MP}},{{XX,XX}},
#ifdef PSCFG_SHORT_BULGE_INDEL
      {{MP,DLB1}},{{MP,DRB1}},{{MP,ILB1}},{{MP,IRB1}},
      {{DLB1,MP}},{{DRB1,MP}},{{ILB1,MP}},{{IRB1,MP}},{{XX,XX}},
      {{MRB1,MLB1}},{{MRB1,ILB1}},{{MRB1,DLB1}},
      {{IRB1,MLB1}},{{IRB1,ILB1}},{{IRB1,DLB1}},
      {{DRB1,MLB1}},{{DRB1,ILB1}},{{DRB1,DLB1}},{{XX,XX}},
#endif
#endif
      {{ST,IR}},{{ST,DR}},{{XX,XX}},
      {{ST,IL}},{{ST,DL}},{{XX,XX}},
      {{MP,IR}},{{MP,DR}},{{XX,XX}},
      {{MP,IL}},{{MP,DL}},{{XX,XX}},
      {{MR,IR}},{{MR,DR}},{{XX,XX}},
      {{MR,IL}},{{MR,DL}},{{XX,XX}},
      {{ML,IL}},{{ML,DL}},{{XX,XX}},
      {{IR,IL}},{{IR,DL}},{{DR,IL}},{{DR,DL}},{{XX,XX}},
      {{IR,MP}},{{DR,MP}},{{XX,XX}},
      {{IL,MP}},{{DL,MP}},{{XX,XX}},
      {{IR,MR}},{{DR,MR}},{{XX,XX}},
      {{IL,ML}},{{DL,ML}},{{XX,XX}},
      {{IR,BF}},{{DR,BF}},{{XX,XX}},
      {{IL,BF}},{{DL,BF}},{{XX,XX}},
      {{IL,IL}},{{DL,DL}},{{XX,XX}},
      {{IR,IR}},{{DR,DR}},{{XX,XX}},
      {{IR,ML}},{{DR,ML}},{{XX,XX}}
#endif
#elif defined(USE_PSCFG_BASE_NO_ML)
      {{ST, IR}},{{ST, DR}},{{XX, XX}},
      {{MP, IR}},{{MP, DR}},{{XX, XX}},
      {{MR, IR}},{{MR, DR}},{{XX, XX}},
      {{IR, MP}},{{DR, MP}},{{XX, XX}},
      {{IR, MR}},{{DR, MR}},{{XX, XX}},
      {{IR, BF}},{{DR, BF}},{{XX, XX}}
#elif defined(USE_PSCFG_STEMLOC_STEMPAIR)
      {{ST,MPS}},{{ST,MLS}},{{ST,MRS}},{{ST,DLS}},{{ST,DRS}},{{ST,ILS}},{{ST,DRS}},{{XX,XX}},
      {{MPS,MLS}},{{MPS,MRS}},{{MLS,MPS}},{{MRS,MPS}},{{XX,XX}},
      {{MPS,DLS}},{{MPS,ILS}},{{MPS,DRS}},{{MPS,IRS}},{{XX,XX}},
      {{DLS,MPS}},{{ILS,MPS}},{{DRS,MPS}},{{IRS,MPS}},{{XX,XX}},
      {{MPS,DLL}},{{MPS,ILL}},{{XX,XX}},
      {{MLS,DLS}},{{MLS,ILS}},{{MRS,DRS}},{{MRS,IRS}},{{DLS,MLS}},{{ILS,MLS}},{{DRS,MRS}},{{IRS,MRS}},{{XX,XX}},
      {{MLS,DRS}},{{MLS,IRS}},{{MRS,DLS}},{{MRS,ILS}},{{DRS,MLS}},{{IRS,MLS}},{{DLS,MRS}},{{ILS,MRS}},{{XX,XX}},
      {{DLS,DLS}},{{ILS,ILS}},{{DRS,DRS}},{{IRS,IRS}},{{XX,XX}},
      {{DLS,ILS}},{{ILS,DLS}},{{DRS,IRS}},{{IRS,DRS}},{{XX,XX}},
      {{DLS,DRS}},{{DLS,IRS}},{{ILS,DRS}},{{ILS,IRS}},{{DRS,DLS}},{{IRS,DLS}},{{DRS,ILS}},{{IRS,ILS}},{{XX,XX}},
      {{MLS,MLS}},{{MRS,MRS}},{{XX,XX}},
      {{DLS,DLS}},{{ILS,ILS}},{{DRS,DRS}},{{IRS,IRS}},{{XX,XX}},
      {{MLL,DLL}},{{MLL,ILL}},{{DLL,MLL}},{{ILL,MLL}},{{XX,XX}},
      {{DLL,DLL}},{{ILL,ILL}},{{XX,XX}}
#elif defined(USE_PSCFG_STEMLOC_SUPERPAIR)
#if 0
      {{ ST,ILB}},{{ ST,DLB}},{{ XX, XX}},
      {{ ST,IRB}},{{ ST,DRB}},{{ XX, XX}},
      {{ MP,ILH}},{{ MP,DLH}},{{ XX, XX}},
      {{ MP,ILB}},{{ MP,DLB}},{{ XX, XX}},
      {{ILB, MP}},{{DLB, MP}},{{ XX, XX}},
      {{ MP,IRB}},{{ MP,DRB}},{{ XX, XX}},
      {{IRB, MP}},{{DRB, MP}},{{ XX, XX}},
      {{ MP,ILI}},{{ MP,DLI}},{{ XX, XX}},
      {{ILI, MP}},{{DLI, MP}},{{ XX, XX}},
      {{MLH,ILH}},{{MLH,DLH}},{{ XX, XX}},
      {{ILH,MLH}},{{DLH,MLH}},{{ XX, XX}},
      {{MLB,ILB}},{{MLB,DLB}},{{ XX, XX}},
      {{ILB,MLB}},{{DLB,MLB}},{{ XX, XX}},
      {{MRB,IRB}},{{MRB,DRB}},{{ XX, XX}},
      {{IRB,MRB}},{{DRB,MRB}},{{ XX, XX}},
      {{MLI,IRB}},{{MLI,DRB}},{{ XX, XX}},
      {{IRB,MLI}},{{DRB,MLI}},{{ XX, XX}},
      {{MLI,ILI}},{{MLI,DLI}},{{ XX, XX}},
      {{ILI,MLI}},{{DLI,MLI}},{{ XX, XX}},
      {{ILB,ILB}},{{DLB,DLB}},{{ XX, XX}},
      {{ILI,ILI}},{{DLI,DLI}},{{ XX, XX}},
      {{IRB,IRB}},{{DRB,DRB}},{{ XX, XX}}
#else
#ifdef SLR_STATE
      {{ SR,MLB}},{{ SR,MRB}},{{ XX, XX}},
      {{ SR,DLI}},{{ SR,ILI}},{{ XX, XX}},
      {{ SL,DLB}},{{ SL,ILB}},{{ XX, XX}},
      {{ SR,DLB}},{{ SR,ILB}},{{ SR,DRB}},{{ SR,IRB}},{{ XX, XX}},
#else
      {{ ST,MLB}},{{ ST,MRB}},{{ XX, XX}},
      {{ ST,DLI}},{{ ST,ILI}},{{ XX, XX}},
      {{ ST,DLB}},{{ ST,ILB}},{{ ST,DRB}},{{ ST,IRB}},{{ XX, XX}},
#endif
      {{ MP,DLH}},{{ MP,ILH}},{{ XX, XX}},
      {{ MP,DLI}},{{ MP,ILI}},{{ XX, XX}},
      {{ MP,MLB}},{{ MP,MRB}},{{ XX, XX}},
      {{MLB, MP}},{{MRB, MP}},{{ XX, XX}},
      {{ MP,DLB}},{{ MP,ILB}},{{ MP,DRB}},{{ MP,IRB}},{{ XX, XX}},
      {{DLB, MP}},{{ILB, MP}},{{DRB, MP}},{{IRB, MP}},{{ XX, XX}},
#ifdef PSCFG_SHORT_BULGE
      {{MP,MLB1}},{{MP,MRB1}},{{MLB1,MP}},{{MRB1,MP}},{{ XX, XX}},
#endif
      {{MLI,DRB}},{{MLI,IRB}},{{ XX, XX}},
      {{DLI,MRB}},{{ILI,MRB}},{{ XX, XX}},
      {{DLI, BF}},{{ILI, BF}},{{ XX, XX}},
      {{DRB, BF}},{{IRB, BF}},{{ XX, XX}},
#if defined(PSCFG_LOOP_TYPE_DEPENDENT_SCORE)
      {{MLB,DLB}},{{MLB,ILB}},{{ XX, XX}},
      {{DLB,MLB}},{{ILB,MLB}},{{ XX, XX}},
      {{MLI,DLI}},{{MLI,ILI}},{{ XX, XX}},
      {{DLI,MLI}},{{ILI,MLI}},{{ XX, XX}},
      {{MLH,DLH}},{{MLH,ILH}},{{ XX, XX}},
      {{DLH,MLH}},{{ILH,MLH}},{{ XX, XX}},
      {{MRB,DRB}},{{MRB,IRB}},{{ XX, XX}},
      {{DRB,MRB}},{{IRB,MRB}},{{ XX, XX}},
      {{DLI,DRB}},{{DLI,IRB}},{{ XX, XX}},
      {{ILI,DRB}},{{ILI,IRB}},{{ XX, XX}},
      {{DLB,ILB}},{{ILB,DLB}},{{DLI,ILI}},{{ILI,DLI}},{{DLH,ILH}},{{ILH,DLH}},{{DRB,IRB}},{{IRB,DRB}},{{ XX, XX}},
      {{DLB,DLB}},{{ILB,ILB}},{{ XX, XX}},
      {{DLI,DLI}},{{ILI,ILI}},{{ XX, XX}},
      {{DLH,DLH}},{{ILH,ILH}},{{ XX, XX}},
      {{DRB,DRB}},{{IRB,IRB}},{{ XX, XX}},
#else
      {{MLB,MLB}},{{MLI,MLI}},{{MLH,MLH}},{{MRB,MRB}},{{ XX, XX}},
      {{MLB,DLB}},{{MLB,ILB}},{{MLI,DLI}},{{MLI,ILI}},{{MLH,DLH}},{{MLH,ILH}},{{MRB,DRB}},{{MRB,IRB}},{{ XX, XX}},
      {{DLB,MLB}},{{ILB,MLB}},{{DLI,MLI}},{{ILI,MLI}},{{DLH,MLH}},{{ILH,MLH}},{{DRB,MRB}},{{IRB,MRB}},{{ XX, XX}},
      {{DLI,DRB}},{{DLI,IRB}},{{ILI,DRB}},{{ILI,IRB}},{{ XX, XX}},
      {{DLB,ILB}},{{ILB,DLB}},{{DLI,ILI}},{{ILI,DLI}},{{DLH,ILH}},{{ILH,DLH}},{{DRB,IRB}},{{IRB,DRB}},{{ XX, XX}},
      {{DLB,DLB}},{{ILB,ILB}},{{DLI,DLI}},{{ILI,ILI}},{{DLH,DLH}},{{ILH,ILH}},{{DRB,DRB}},{{IRB,IRB}},{{ XX, XX}},
#endif
      {{DLH, EN}},{{ILH, EN}},{{ XX, XX}}
#endif
#else

#endif
    };
    int n_tsc_constraints = sizeof(tsc_constraints) / sizeof(CArray<State,2>);

    setLoopTypeIndexes(data);
    setBaseStates(data);
    setNChildren(data);
    setNParents(data);
    setChild(data);
    setParent(data);
    setChildIndex(data);
    setNTSC0(data, tsc_constraints, n_tsc_constraints);
    setTSC0Index(data, tsc_constraints, n_tsc_constraints);
    setNTSC0ForTSC0Index();
    setTSC0ForTSC0Index();
    setNStateForBase(data);
    setStateForBase(data);
    setNames(data);

#ifdef DEBUG
    Assert(!isChild(BF_st, END_st));

    // because of ST->BF transition,
    // BEG->ST and BEG->BF cannot co-exist unambigously
    bool bf = false;
    bool st = false;
    for (int i = 0; i < (int)getNChild(BEG_st); i++) {
      const State& s = getChild(BEG_st, i);
      if (s == BF_st) {
	bf = true;
      } else if (!isEmitState(s)) {
	st = true;
      }
    }
    Assert(!(bf && st));
    // tentatively disallow BEG->BF transition due to complication
    Assert(!bf);
#endif
  }
  ~PSCFG_Energy() {}

  CArray<int,NSTATE0> loopTypeIndexes;
  CArray<PSCFG_Base::State,NSTATE0> baseStates;
  CArray<int,NSTATE0> nChildren;
  CArray<int,NSTATE0> nParents;
  CMatrix<State,NSTATE0,(MAX_NCHILD+1)> child;
  CMatrix<State,NSTATE0,(MAX_NPARENT+1)> parent;
  CMatrix<int,NSTATE0,NSTATE0> childIndex;
  int ntsc0;
  CMatrix<int,NSTATE0,NSTATE0> tsc0Index;
  CArray<int,PSCFG_Base::NSTATE0> nStateForBase;
  CMatrix<State, PSCFG_Base::NSTATE0,
	  (MAX_NSTATE_FOR_BASE + 1)> stateForBase;
  CArray<CArray<char,20>, NSTATE0> names;
  CArray<int, NTSC0> nTSC0ForTSC0Index;
  CMatrix<CArray<State, 2>, NTSC0, NTSC0> tsc0ForTSC0Index;

  void setLoopTypeIndexes(const Data& data) {
    for (int i = 0; i < NSTATE0; i++) {
      const State& s = data[i].state;
      loopTypeIndexes[s] = data[i].loopTypeIndex;

#ifdef USE_PSCFG_ENERGY
      Assert(loopTypeIndexes[s] == getLoopTypeIndex0(s));
#endif
    }
  }
  void setBaseStates(const Data& data) {
    for (int i = 0; i < NSTATE0; i++) {
      const State& s = data[i].state;
      baseStates[s] = data[i].baseState;

#ifdef USE_PSCFG_ENERGY
      Assert(baseStates[s] == toPSCFGBaseState0(s));
#endif
    }
  }
  void setNChildren(const Data& data) {
    nChildren.fill(0);
    for (int i = 0; i < NSTATE0; i++) {
      for (int j = 0; data[i].children[j] != XX; j++) {
	if (!isNormalState(data[i].children[j])) break;

	const State& s = data[i].state;
	nChildren[s]++;
      }
    }
#ifdef DEBUG
    int nmax = 0;
    for (int i = 0; i < NSTATE0; i++) {
      const State& s = data[i].state;
      nmax = max(nmax, nChildren[s]);

#ifdef USE_PSCFG_ENERGY
      Assert(nChildren[s] == getNChild0(s));
#endif
    }
    Assert(nmax == MAX_NCHILD, 
	   "nmax=%d, MAX_NCHILD=%d", nmax, MAX_NCHILD);
#endif
  }
  void setNParents(const Data& data) {
    nParents.fill(0);
    for (int i = 0; i < NSTATE0; i++) {
      if (!isNormalState(data[i].state)) break;

      for (int j = 0; data[i].children[j] != XX; j++) {
	const State& sc = data[i].children[j];
	nParents[sc]++;
      }
    }
#ifdef DEBUG
    int nmax = 0;
    for (int i = 0; i < NSTATE0; i++) {
      const State& s = data[i].state;
      nmax = max(nmax, nParents[s]);

#ifdef USE_PSCFG_ENERGY
      Assert(nParents[s] == getNParent0(s));
#endif
    }
    Assert(nmax == MAX_NPARENT,
	   "nmax=%d, MAX_NPARENT=%d", nmax, MAX_NPARENT);
#endif
  }
  void setChild(const Data& data) {
    child.fill(XX);
    for (int i = 0; i < NSTATE0; i++) {
      for (int j = 0; data[i].children[j] != XX; j++) {
	const State& sp = data[i].state;
	const State& sc = data[i].children[j];
	child.ref(sp, j) = sc;
#ifdef USE_PSCFG_ENERGY
	Assert(sc == getChild0(sp, j));
#endif
      }
    }
  }
  void setParent(const Data& data) {
    parent.fill(XX);
    CArray<int,NSTATE0> parentIndexes;
    parentIndexes.fill(0);
    for (int i = 0; i < NSTATE0; i++) {
      for (int j = 0; data[i].children[j] != XX; j++) {
	const State& sp = data[i].state;
	const State& sc = data[i].children[j];
	int& index = parentIndexes[sc];
	parent.ref(sc, index) = sp;
#ifdef USE_PSCFG_ENERGY
	Assert(sp == getParent0(sc, index));
#endif
	index++;
      }
    }
  }
  void setChildIndex(const Data& data) {
    childIndex.fill(-1);
    for (int i = 0; i < NSTATE0; i++) {
      for (int j = 0; data[i].children[j] != XX; j++) {
	const State& sp = data[i].state;
	const State& sc = data[i].children[j];
	childIndex.ref(sp, sc) = j;
#ifdef USE_PSCFG_ENERGY
	Assert(j == getChildIndex0(sp, sc));
#endif
      }
    }
  }
  void setNTSC0(const Data& data, const CArray<State,2> constr[], const int& nconstr) {
    Array<int> group2tscIndex(nconstr);
    Array<int> constr2groupIndex(nconstr);
    group2tscIndex.fill(-1);
    constr2groupIndex.fill(-1);
    bool prev_constr = false;
    int gi = (-1);
    for (int i = 0; i < nconstr; i++) {
      if (constr[i][0] == XX) {
	prev_constr = false;
      } else {
	if (!prev_constr) {
	  gi++;
	  prev_constr = true;
	}
	constr2groupIndex[i] = gi;
      }
    }

    int n = 0;
    for (int i = 0; i < NSTATE0; i++) {
      if (!isNormalState(data[i].state)) break;

      for (int j = 0; data[i].children[j] != XX; j++) {
	//if (!isNormalState(data[i].children[j])) break; 
	//includes transitions to END_st

	const State& sp = data[i].state;
	const State& sc = data[i].children[j];
	int gi = (-1);
	for (int k = 0; k < nconstr; k++) {
	  if ((constr[k][0] == sp) && (constr[k][1] == sc)) {
	    gi = constr2groupIndex[k];
	    break;
	  }
	}
	if (gi >= 0) {
	  if (group2tscIndex[gi] == (-1)) {
	    group2tscIndex[gi] = n++;
	  }
	} else {
	  n++;
	}
      }
    }
    ntsc0 = n;

    Assert(ntsc0 == NTSC0, "ntsc0(%d) != NTSC0(%d)", ntsc0, NTSC0);
#ifdef USE_PSCFG_ENERGY
    if (nconstr == 0) {
      Assert(ntsc0 == getNTSC00());
    }
#endif
  }
  void setTSC0Index(const Data& data, const CArray<State,2> constr[], const int& nconstr) {
    tsc0Index.fill(-1);
    Array<int> group2tscIndex(nconstr);
    Array<int> constr2groupIndex(nconstr);
    group2tscIndex.fill(-1);
    constr2groupIndex.fill(-1);
    int gi = (-1);
    bool prev_constr = false;
    for (int i = 0; i < nconstr; i++) {
      if (constr[i][0] == XX) {
	prev_constr = false;
      } else {
	if (!prev_constr) {
	  gi++;
	  prev_constr = true;
	}
	constr2groupIndex[i] = gi;
      }
    }

    int n = 0;
    for (int i = 0; i < NSTATE0; i++) {
      if (!isNormalState(data[i].state)) break;

      for (int j = 0; data[i].children[j] != XX; j++) {
	//if (!isNormalState(data[i].children[j])) break;
	//includes transitions to END_st

	const State& sp = data[i].state;
	const State& sc = data[i].children[j];
	int gi = (-1);
	for (int k = 0; k < nconstr; k++) {
	  if ((constr[k][0] == sp) && (constr[k][1] == sc)) {
	    gi = constr2groupIndex[k];
	    break;
	  }
	}
	if (gi >= 0) {
	  if (group2tscIndex[gi] == (-1)) {
	    group2tscIndex[gi] = n++;
	  }
	  tsc0Index.ref(sp, sc) = group2tscIndex[gi];
	} else {
	  tsc0Index.ref(sp, sc) = n++;
	}
#ifdef USE_PSCFG_ENERGY
	if (nconstr == 0) {
	  Assert(n == getTSC0Index0(sp, sc));
	}
#endif
      }
    }

#ifdef DEBUG
    Assert(n == NTSC0, "n(%d) != NTSC0(%d)", n, NTSC0);
    Assert(n == ntsc0);
#endif
  }
  void setNTSC0ForTSC0Index() {
    nTSC0ForTSC0Index.fill(0);
    for (int i = 0; i < NSTATE0; i++) {
      const State& sp = State(i);
      if (!isNormalState(sp)) break;

      for (int j = 0; j < NSTATE0; j++) {
	const State& sc = State(j);
	if (!isChild(sp, sc)) continue;

	const int& tsc0Index = getTSC0Index(sp, sc);
	nTSC0ForTSC0Index[tsc0Index]++;
      }
    }
  }
  void setTSC0ForTSC0Index() {
    CArray<State, 2> tmp;
    tmp.fill(X_st);
    tsc0ForTSC0Index.fill(tmp);
    CArray<int, NTSC0> indexes;
    indexes.fill(0);
    for (int i = 0; i < NSTATE0; i++) {
      const State& sp = State(i);
      if (!isNormalState(sp)) break;

      for (int j = 0; j < NSTATE0; j++) {
	const State& sc = State(j);
	if (!isChild(sp, sc)) continue;

	const int& tsc0Index = getTSC0Index(sp, sc);
	tsc0ForTSC0Index.ref(tsc0Index, indexes[tsc0Index])[0] = sp;
	tsc0ForTSC0Index.ref(tsc0Index, indexes[tsc0Index])[1] = sc;
	indexes[tsc0Index]++;
      }
    }
  }
  void setNStateForBase(const Data& data) {
    nStateForBase.fill(0);
    for (int i = 0; i < NSTATE0; i++) {
      nStateForBase[data[i].baseState]++;
    }

#ifdef DEBUG
    int nmax = 0;
    for (int i = 0; i < (int)PSCFG_Base::NSTATE0; i++) {
      nmax = max(nmax, nStateForBase[i]);
#ifdef USE_PSCFG_ENERGY
      Assert(nStateForBase[i] == getNStateForBase0(PSCFG_Base::State(i)));
#endif
    }
    Assert(nmax == MAX_NSTATE_FOR_BASE,
	   "nmax=%d, MAX_NSTATE_FOR_BASE=%d", nmax, MAX_NSTATE_FOR_BASE);
#endif
  }
  void setStateForBase(const Data& data) {
    stateForBase.fill(XX);
    CArray<int,PSCFG_Base::NSTATE0> indexes;
    indexes.fill(0);
    for (int i = 0; i < NSTATE0; i++) {
      const State& s = data[i].state;
      const PSCFG_Base::State& s0 = data[i].baseState;
      int& index = indexes[s0];
      stateForBase.ref(s0, index) = s;
#ifdef USE_PSCFG_ENERGY
      Assert(s == getStateForBase0(s0, index));
#endif
      index++;
    }
  }  
  void setNames(const Data& data) {
    for (int i = 0; i < NSTATE0; i++) {
      const State& s = data[i].state;
      names[s].fill(0);
      strncpy(&names[s][0], data[i].name, (names[s].size() - 1));
#ifdef USE_PSCFG_ENERGY
      Assert(strcmp(&names[s][0], state2str0(s)) == 0);
#endif
    }
  }

public:	 
  inline static const Model& getModel() {
    static const Model model;
    return model;
  }
  inline static const char* state2str(int state) {
    return (isValidState(state) ? &getModel().names[state][0] : "??");
  }	 
  inline const int& getLoopTypeIndex(const State& s) const {
    return loopTypeIndexes[s];
  }	 
  inline const PSCFG_Base::State toPSCFGBaseState(const State& s) const {
#if defined(USE_PSCFG_BASE) && !defined(PSCFG_BASE_HAIRPIN) && !defined(PSCFG_SHORT_BULGE)
    return PSCFG_Base::State(s);
#else
    return baseStates[s];
#endif
  }	 
  inline const int& getNChild(const State& s) const {
    return nChildren[s];
  }	 
  inline const State& getChild(const State& s, const int& i) const {
    return child.ref(s, i);
  }	 
  inline const int& getChildIndex(const State& s, const State& s1) const {
    return childIndex.ref(s, s1);
  }
  inline bool isChild(const State& s, const State& s1) const {
    return (getChildIndex(s, s1) >= 0);
  }
  inline bool isParent(const State& s, const State& s1) const {
    return isChild(s1, s);
  }
  inline const int& getNParent(const State& s) const {
    return nParents[s];
  }	 
  inline const State& getParent(const State& s, const int& i) const {
    return parent.ref(s, i);
  }	 
  inline const int& getNTSC0() const {
    return ntsc0;
  }	 
  inline const int& getTSC0Index(const State& s, const State& s1) const {
    return tsc0Index.ref(s, s1);
  }	
  inline const int& getNTSC0ForTSC0Index(const int& tsc0Index) const {
    return nTSC0ForTSC0Index[tsc0Index];
  }
  inline const State& getTSC0ParentForTSC0Index(const int& tsc0Index, const int& i) const {
    return tsc0ForTSC0Index.ref(tsc0Index, i)[0];
  }
  inline const State& getTSC0ChildForTSC0Index(const int& tsc0Index, const int& i) const {
    return tsc0ForTSC0Index.ref(tsc0Index, i)[1];
  }
  inline const int& getNStateForBase(const State& s) const {
    return nStateForBase[s];
  }	 
  inline const State& 
  getStateForBase(const State& s, const int& i) const {
    return stateForBase.ref(s, i);
  }

  typedef PSCFG_Base EmitModel;
  typedef PSCFG_Base::State EmitState;
  static int getNLoop() {return NLOOP;}
  static int getNLoopType() {return NLOOP;}
  static bool isXXState(const State& s) {return (s == X_st);}
  static State beginState() {return BEGIN_STATE;}
  static State lastState() {return LAST_STATE;}
  static bool isNormalState(const int& s) {return (beginState() <= s && s <= lastState());}
  static bool isValidState(const int& s) {return (beginState() <= s && s <= XX);}
  inline static int getNEmitState() {return NEMIT_STATE;};
  inline static State getEmitState(const int& i) {return State(EMIT_STATE_BEG + i);}
  inline static bool isEmitState(const State& s) {
    return (EMIT_STATE_BEG <= s && s < (EMIT_STATE_BEG + NEMIT_STATE));
  }
  static bool bifurcates(const State& s) {return (s == BF_st);}
  const Rect& getNEmit(const State& s) const {
    return PSCFG_Base::getNEmit(toPSCFGBaseState(s));
  }
  Rect addNEmit(const Rect& rect, const State& s) const {
    return PSCFG_Base::addNEmit(rect, toPSCFGBaseState(s));
  }
  Rect subtractNEmit(const Rect& rect, const State& s) const {
    return PSCFG_Base::subtractNEmit(rect, toPSCFGBaseState(s));
  }
  bool emitB(const State& s) const {
    return PSCFG_Base::emitB(toPSCFGBaseState(s));
  }
  bool emitE(const State& s) const {
    return PSCFG_Base::emitE(toPSCFGBaseState(s));
  }
  bool emitBX(const State& s) const {
    return PSCFG_Base::emitBX(toPSCFGBaseState(s));
  }
  bool emitEX(const State& s) const {
    return PSCFG_Base::emitEX(toPSCFGBaseState(s));
  }
  static int getNESC0() {return NESC0;}
  static int getESC0Index(const State& s) {return s;}

#ifdef SLR_STATE
  inline static int getNSTState() {return 3;}//SL_st,SR_st,BF_st
  inline static int getSTStateIndex(const State& s) {
    switch (s) {
    case SL_st: return 0;
    case SR_st: return 1;
    case BF_st: return 2;
    default:
      Assert(false, "bad st state %d", s);
      return (-1);
    }
  }
#endif

  class Iterator;
  class TBNode;
  class ScoreManager;

#ifdef USE_PSCFG_ENERGY
  //////////////////////////////////////////////////////////////////////
  // old code
  static const char* state2str0(int s) {
    static char names[NSTATE0+1][20] = {
      "ST",

      "MP",

      "MR_Exterior",
      "IR_Exterior",
      "DR_Exterior",

      "MR_Hairpin",
      "IR_Hairpin",
      "DR_Hairpin",

      "MR_Bulge",
      "IR_Bulge",
      "DR_Bulge",
      "ML_Bulge",
      "IL_Bulge",
      "DL_Bulge",

      "MR_Interior",
      "IR_Interior",
      "DR_Interior",
      "ML_Interior",
      "IL_Interior",
      "DL_Interior",

      "MR_Multi1",
      "IR_Multi1",
      "DR_Multi1",
      "ML_Multi1",
      "IL_Multi1",
      "DL_Multi1",

      "MR_Multi",
      "IR_Multi",
      "DR_Multi",
      "ML_Multi",
      "IL_Multi",
      "DL_Multi",

      "BF",

      "BEG",
      "END",
      "X",        
      "??"
    };
    return names[isValidState(s) ? s : NSTATE0];
  }
  static const int& getLoopTypeIndex0(const State& s) {
    //return SCFG_Energy::getLoopTypeIndex(toSCFGEnergyState(s));
    static const CArray<int,NSTATE0> table = {
      (-1),//ST_st
      (-1),//MP_st
      3,//MR_Exterior_st
      3,//IR_Exterior_st
      3,//DR_Exterior_st
      0,//MR_Hairpin_st
      0,//IR_Hairpin_st
      0,//DR_Hairpin_st
      1,//MR_Bulge_st
      1,//IR_Bulge_st
      1,//DR_Bulge_st
      1,//ML_Bulge_st
      1,//IL_Bulge_st
      1,//DL_Bulge_st
      2,//MR_Interior_st
      2,//IR_Interior_st
      2,//DR_Interior_st
      2,//ML_Interior_st
      2,//IL_Interior_st
      2,//DL_Interior_st
      3,//MR_Multi1_st
      3,//IR_Multi1_st
      3,//DR_Multi1_st
      3,//ML_Multi1_st
      3,//IL_Multi1_st
      3,//DL_Multi1_st
      3,//MR_Multi_st
      3,//IR_Multi_st
      3,//DR_Multi_st
      3,//ML_Multi_st
      3,//IL_Multi_st
      3,//DL_Multi_st
      (-1),//BF_st
      (-1),//BEG_st
      (-1),//END_st
      (-1) //X_st            
    };
    return table[s];
  }
    inline static const PSCFG_Base::State&
  toPSCFGBaseState0(const State& s) {
    static const PSCFG_Base::State table[NSTATE0] = {
      PSCFG_Base::ST_st,//ST_st
      PSCFG_Base::MP_st,//MP_st
      PSCFG_Base::MR_st,//MR_Exterior_st
      PSCFG_Base::IR_st,//IR_Exterior_st
      PSCFG_Base::DR_st,//DR_Exterior_st
      PSCFG_Base::MR_st,//MR_Hairpin_st
      PSCFG_Base::IR_st,//IR_Hairpin_st
      PSCFG_Base::DR_st,//DR_Hairpin_st
      PSCFG_Base::MR_st,//MR_Bulge_st
      PSCFG_Base::IR_st,//IR_Bulge_st
      PSCFG_Base::DR_st,//DR_Bulge_st
      PSCFG_Base::ML_st,//ML_Bulge_st
      PSCFG_Base::IL_st,//IL_Bulge_st
      PSCFG_Base::DL_st,//DL_Bulge_st
      PSCFG_Base::MR_st,//MR_Interior_st
      PSCFG_Base::IR_st,//IR_Interior_st
      PSCFG_Base::DR_st,//DR_Interior_st
      PSCFG_Base::ML_st,//ML_Interior_st
      PSCFG_Base::IL_st,//IL_Interior_st
      PSCFG_Base::DL_st,//DL_Interior_st
      PSCFG_Base::MR_st,//MR_Multi1_st
      PSCFG_Base::IR_st,//IR_Multi1_st
      PSCFG_Base::DR_st,//DR_Multi1_st
      PSCFG_Base::ML_st,//ML_Multi1_st
      PSCFG_Base::IL_st,//IL_Multi1_st
      PSCFG_Base::DL_st,//DL_Multi1_st
      PSCFG_Base::MR_st,//MR_Multi_st
      PSCFG_Base::IR_st,//IR_Multi_st
      PSCFG_Base::DR_st,//DR_Multi_st
      PSCFG_Base::ML_st,//ML_Multi_st
      PSCFG_Base::IL_st,//IL_Multi_st
      PSCFG_Base::DL_st,//DL_Multi_st
      PSCFG_Base::BF_st,//BF_st
      PSCFG_Base::BEG_st,//BEG_st
      PSCFG_Base::END_st,//END_st
      PSCFG_Base::X_st //X_st            
    };
    return table[s];
  }
    inline static const State& getChild0(const State& s, const int& i) {
    static const State children[NSTATE0][21] = {
      {  MP,MR1,IR1,DR1,ML1,IL1,DL1,MRM,IRM,DRM,
	MLM,ILM,DLM, BF, XX, XX, XX, XX, XX, XX, XX},//<-ST
      {  MP,MRH,IRH,DRH,MRB,IRB,DRB,MLB,ILB,DLB,
	MRI,IRI,DRI,MRM,IRM,DRM,MLM,ILM,DLM, BF, XX},//<-MP

      { MRE,IRE,DRE, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-MRE
      { MRE,IRE, XX, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-MRE
      { MRE,DRE, XX, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-MRE

      { MRH,IRH,DRH, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-MRH
      { MRH,IRH, XX, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-IRH
      { MRH,DRH, XX, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-DRH

      {  MP,MRB,IRB,DRB, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-MRB
      {  MP,MRB,IRB, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-IRB
      {  MP,MRB,DRB, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-DRB
      {  MP,MLB,ILB,DLB, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-MLB
      {  MP,MLB,ILB, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-ILB
      {  MP,MLB,DLB, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-DLB

      { MRI,IRI,DRI,MLI,ILI,DLI, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-MRI
      { MRI,IRI,MLI,ILI,DLI, XX, XX, XX, XX, XX, 
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-IRI
      { MRI,DRI,MLI,ILI,DLI, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-DRI
      {  MP,MLI,ILI,DLI, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-MLI
      {  MP,MLI,ILI, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-ILI
      {  MP,MLI,DLI, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-DLI

      {  MP,MR1,IR1,DR1,ML1,IL1,DL1, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-MR1
      {  MP,MR1,IR1,ML1,IL1,DL1, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-IR1
      {  MP,MR1,DR1,ML1,IL1,DL1, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-DR1
      {  MP,ML1,IL1,DL1, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-ML1
      {  MP,ML1,IL1, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-IL1
      {  MP,ML1,DL1, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-DL1

      { MRM,IRM,DRM,MLM,ILM,DLM, BF, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-MRM
      { MRM,IRM,MLM,ILM,DLM, BF, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-IRM
      { MRM,DRM,MLM,ILM,DLM, BF, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-DRM
      { MLM,ILM,DLM, BF, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-MLM
      { MLM,ILM, BF, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-ILM
      { MLM,DLM, BF, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-DLM

      {  ST, XX, XX, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-BF

      {  ST,MRE,IRE,DRE, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-BG
      {  XX, XX, XX, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//<-EN
      {  XX, XX, XX, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX} //<-XX
    };
    return children[s][i];
  }
   inline static const int& getNChild0(const State& s) {
    static const int nChild[NSTATE0] = {
      14,//ST->MP,(MID)(RL)1,(MID)(RL)M,BF
      20,//MP->MP,(MID)RH,(MID)(RL)B,(MID)RI,(MID)(RL)M,BF

      3,//MRE->(MID)RE
      2,//IRE->(MI)RE
      2,//DRE->(MD)RE

      3,//MRH->(MID)RH
      2,//IRH->(MI)RH
      2,//DRH->(MD)RH

      4,//MRB->MP,(MID)RB
      3,//IRB->MP,(MI)RB
      3,//DRB->MP,(MD)RB
      4,//MLB->MP,(MID)LB
      3,//ILB->MP,(MI)LB
      3,//DLB->MP,(MD)LB

      6,//MRI->(MID)RI,(MID)LI
      5,//IRI->(MI)RI,(MID)LI
      5,//DRI->(MD)RI,(MID)LI
      4,//MLI->MP,(MID)LI
      3,//ILI->MP,(MI)LI
      3,//DLI->MP,(MD)LI

      7,//MR1->MP,(MID)R1,(MID)L1
      6,//IR1->MP,(MI)R1,(MID)L1
      6,//DR1->MP,(MD)R1,(MID)L1
      4,//ML1->MP,(MID)L1
      3,//IL1->MP,(MI)L1
      3,//DL1->MP,(MD)L1

      7,//MRM->(MID)RM,(MID)LM,BF
      6,//IRM->(MI)RM,(MID)LM,BF
      6,//DRM->(MD)RM,(MID)LM,BF
      4,//MLM->(MID)LM,BF
      3,//ILM->(MI)LM,BF
      3,//DLM->(MD)LM,BF

      1,//BF->ST
      4,//BG->ST,(MID)RE
      0,//EN->
      0 //XX->
    };
    return nChild[s];
  }
  //enum {
  //  NESC0 = NSTATE,
  //  NTSC0 = (14+//ST->MP,(MID)(RL)1,(MID)(RL)M,BF
  //	     20+//MP->MP,(MID)RH,(MID)(RL)B,(MID)RI,(MID)(RL)M,BF
  //	     
  //	     3+//MRE->(MID)RE
  //	     2+//IRE->(MI)RE
  //	     2+//DRE->(MD)RE
  //
  //	     3+//MRH->(MID)RH
  //	     2+//IRH->(MI)RH
  //	     2+//DRH->(MD)RH
  //
  //	     4+//MRB->MP,(MID)RB
  //	     3+//IRB->MP,(MI)RB
  //	     3+//DRB->MP,(MD)RB
  //	     4+//MLB->MP,(MID)LB
  //	     3+//ILB->MP,(MI)LB
  //	     3+//DLB->MP,(MD)LB
  //
  //	     6+//MRI->(MID)RI,(MID)LI
  //	     5+//IRI->(MI)RI,(MID)LI
  //	     5+//DRI->(MD)RI,(MID)LI
  //	     4+//MLI->MP,(MID)LI
  //	     3+//ILI->MP,(MI)LI
  //	     3+//DLI->MP,(MD)LI
  //
  //	     7+//MR1->MP,(MID)R1,(MID)L1
  //	     6+//IR1->MP,(MI)R1,(MID)L1
  //	     6+//DR1->MP,(MD)R1,(MID)L1
  //	     4+//ML1->MP,(MID)L1
  //	     3+//IL1->MP,(MI)L1
  //	     3+//DL1->MP,(MD)L1
  //
  //	     7+//MRM->(MID)RM,(MID)LM,BF
  //	     6+//IRM->(MI)RM,(MID)LM,BF
  //	     6+//DRM->(MD)RM,(MID)LM,BF
  //	     4+//MLM->(MID)LM,BF
  //	     3+//ILM->(MI)LM,BF
  //	     3+//DLM->(MD)LM,BF
  //
  //	     1 //BF->ST
  //
  //	     //4,//BG->ST,(MID)RE
  //	     //0,//EN->
  //	     //0 //XX->
  //	     )
  //  //       = 153
  //};
  inline static const signed char&
  getChildIndex0(const State& s, const State& s1){
    static const struct ChildIndex {
      typedef PSCFG_Energy Model;
      typedef PSCFG_Energy::State State;
      enum {
	LENGTH = Model::NSTATE0,
	//SIZE = (LENGTH * LENGTH),
	NULL_INDEX = (-1)
      };
      signed char data[LENGTH][LENGTH];// index is upto 127

      ChildIndex() {
	for (int i = 0; i < LENGTH; i++) {
	  for (int j = 0; j < LENGTH; j++) {
	    data[i][j] = NULL_INDEX;
	  }
	}
	for (int i = 0; i < LENGTH; i++) {
	  const State& s = (State)i;
	  int k = 0;
	  for (int j = 0; j < Model::getNChild(s); j++) {
	    const State& s1 = Model::getChild(s, j);
	    data[s][s1] = k++;
	  }
	}
      }
      inline const signed char& 
      getVal(const State& s, const State& s1) const {
	return data[s][s1];
      }
    } childIndex;
    return childIndex.getVal(s, s1);
  }
  inline static const State& 
  getStateForBase0(const PSCFG_Base::State& s, const int& i) {
    static const State table[PSCFG_Base::NSTATE0][7] = {
      { ST, XX, XX, XX, XX, XX, XX},//ST
      { MP, XX, XX, XX, XX, XX, XX},//MP
      {MRE,MRH,MRB,MRI,MR1,MRM, XX},//MR
      {IRE,IRH,IRB,IRI,IR1,IRM, XX},//IR
      {DRE,DRH,DRB,DRI,DR1,DRM, XX},//DR
      {MLB,MLI,ML1,MLM, XX, XX, XX},//ML
      {ILB,ILI,IL1,ILM, XX, XX, XX},//IL
      {DLB,DLI,DL1,DLM, XX, XX, XX},//DL
      { BF, XX, XX, XX, XX, XX, XX},//BF
      { BG, XX, XX, XX, XX, XX, XX},//BG
      { EN, XX, XX, XX, XX, XX, XX},//EN
      { XX, XX, XX, XX, XX, XX, XX} //XX
    };
    return table[s][i];
   }
  inline static const int& 
  getNStateForBase0(const PSCFG_Base::State& s) {
    static const int nState[PSCFG_Base::NSTATE0] = {
      1,//ST
      1,//MP
      6,//MR
      6,//IR
      6,//DR
      4,//ML
      4,//IL
      4,//DL
      1,//BF
      1,//BG
      1,//EN
      1 //XX
    };
    return nState[s];
  }
  //inline static const SCFG_Energy::State&
  //toSCFGEnergyState(const State& s) {
  //  static const SCFG_Energy::State table[NSTATE0] = {
  //    SCFG_Energy::ST_st,//ST_st
  //
  //    SCFG_Energy::MP_st,//MP_st
  //
  //    SCFG_Energy::MR_Exterior_st,//MR_Exterior_st
  //    SCFG_Energy::MR_Exterior_st,//IR_Exterior_st
  //    SCFG_Energy::MR_Exterior_st,//DR_Exterior_st
  //
  //    SCFG_Energy::MR_Hairpin_st,//MR_Hairpin_st
  //    SCFG_Energy::MR_Hairpin_st,//IR_Hairpin_st
  //    SCFG_Energy::MR_Hairpin_st,//DR_Hairpin_st
  //
  //    SCFG_Energy::MR_Bulge_st,//MR_Bulge_st
  //    SCFG_Energy::MR_Bulge_st,//IR_Bulge_st
  //    SCFG_Energy::MR_Bulge_st,//DR_Bulge_st
  //    SCFG_Energy::ML_Bulge_st,//ML_Bulge_st
  //    SCFG_Energy::ML_Bulge_st,//IL_Bulge_st
  //    SCFG_Energy::ML_Bulge_st,//DL_Bulge_st
  //
  //    SCFG_Energy::MR_Interior_st,//MR_Interior_st
  //    SCFG_Energy::MR_Interior_st,//IR_Interior_st
  //    SCFG_Energy::MR_Interior_st,//DR_Interior_st
  //    SCFG_Energy::ML_Interior_st,//ML_Interior_st
  //    SCFG_Energy::ML_Interior_st,//IL_Interior_st
  //    SCFG_Energy::ML_Interior_st,//DL_Interior_st
  //
  //    SCFG_Energy::MR_Multi1_st,//MR_Multi1_st
  //    SCFG_Energy::MR_Multi1_st,//IR_Multi1_st
  //    SCFG_Energy::MR_Multi1_st,//DR_Multi1_st
  //    SCFG_Energy::ML_Multi1_st,//ML_Multi1_st
  //    SCFG_Energy::ML_Multi1_st,//IL_Multi1_st
  //    SCFG_Energy::ML_Multi1_st,//DL_Multi1_st
  //
  //    SCFG_Energy::MR_Multi_st,//MR_Multi_st
  //    SCFG_Energy::MR_Multi_st,//IR_Multi_st
  //    SCFG_Energy::MR_Multi_st,//DR_Multi_st
  //    SCFG_Energy::ML_Multi_st,//ML_Multi_st
  //    SCFG_Energy::ML_Multi_st,//IL_Multi_st
  //    SCFG_Energy::ML_Multi_st,//DL_Multi_st
  //  
  //    SCFG_Energy::BF_st,//BF_st
  //    SCFG_Energy::BEG_st,//BEG_st
  //    SCFG_Energy::END_st,//END_st
  //    SCFG_Energy::X_st //X_st
  //  };
  //  return table[s];
  //}
  //inline static const State& 
  //getStateForSCFG(const SCFG_Energy::State& s, const int& i) {
  //  static const State table[SCFG_Energy::NSTATE0][4] = {
  //    { ST, XX, XX, XX},//ST_st
  //    { MP, XX, XX, XX},//MP_st
  //    {MRE,IRE,DRE, XX},//R_Exterior_st
  //    {MRH,IRH,DRH, XX},//R_Hairpin_st
  //    {MRB,IRB,DRB, XX},//R_Bulge_st
  //    {MLB,ILB,DLB, XX},//L_Bulge_st 
  //    {MRI,IRI,DRI, XX},//R_Interior_st
  //    {MLI,ILI,DLI, XX},//L_Interior_st
  //    {MR1,IR1,DR1, XX},//R_Multi1_st
  //    {ML1,IL1,DL1, XX},//L_Multi1_st
  //    {MRM,IRM,DRM, XX},//R_Multi_st
  //    {MLM,ILM,DLM, XX},//L_Multi_st
  //    { BF, XX, XX, XX},//BF_st
  //    { BG, XX, XX, XX},//BEG_st
  //    { EN, XX, XX, XX},//END_st
  //    { XX, XX, XX, XX} //X_st                 
  //  };
  //  return table[s][i];
  //}
  //inline static const int& 
  //getNStateForSCFG(const SCFG_Energy::State& s) {
  //  static const int nState[SCFG_Energy::NSTATE0] = {
  //    1,//ST_st         
  //    1,//MP_st         
  //    3,//R_Exterior_st 
  //    3,//R_Hairpin_st  
  //    3,//R_Bulge_st    
  //    3,//L_Bulge_st    
  //    3,//R_Interior_st 
  //    3,//L_Interior_st 
  //    3,//R_Multi1_st   
  //    3,//L_Multi1_st   
  //    3,//R_Multi_st    
  //    3,//L_Multi_st    
  //    1,//BF_st         
  //    1,//BEG_st        
  //    1,//END_st        
  //    1	//X_st                  
  //  };
  //  return nState[s];
  //}
  static bool isChild0(const State& s, const State& s1) {
    switch (s) {
    case BG:
      return (s1 == ST || s1 == MRE || s1 == IRE || s1 == DRE);
    case EN:
      return (s == MRE || s == MRH);
    default:
      {
	const int& i_end = getNChild(s);
	for (int i = 0; i < i_end; i++) {
	  if (s1 == getChild(s, i)) return true;
	}
	return false;
      }
    }
  }
   inline static const int& getNParent0(const State& s) {
    static const int nParent[NSTATE0] = {
      1,//ST<-BF
      17,//MP<-ST,MP,(MID)(RL)B,(MID)LI,(MID)(RL)1

      3,//MRE<-(MID)RE
      2,//IRE<-(MI)RE
      2,//DRE<-(MD)RE

      4,//MRH<-MP,(MID)RH
      3,//IRH<-MP,(MI)RH
      3,//DRH<-MP,(MD)RH

      4,//MRB<-MP,(MID)RB
      3,//IRB<-MP,(MI)RB
      3,//DRB<-MP,(MD)RB
      4,//MLB<-MP,(MID)LB
      3,//ILB<-MP,(MI)LB
      3,//DLB<-MP,(MD)LB

      4,//MRI<-MP,(MID)RI
      3,//IRI<-MP,(MI)RI
      3,//DRI<-MP,(MD)RI
      6,//MLI<-(MID)RI,(MID)LI
      5,//ILI<-(MI)RI,(MID)LI
      5,//DLI<-(MD)RI,(MID)LI

      4,//MR1<-ST,(MID)R1
      3,//IR1<-ST,(MI)R1
      3,//DR1<-ST,(MD)R1
      7,//ML1<-ST,(MID)R1,(MID)L1
      6,//IL1<-ST,(MID)R1,(MI)L1
      6,//DL1<-ST,(MID)R1,(MD)L1

      5,//MRM<-ST,MP,(MID)RM
      4,//IRM<-ST,MP,(MI)RM
      4,//DRM<-ST,MP,(MD)RM
      8,//MLM<-ST,MP,(MID)RM,(MID)LM
      7,//ILM<-ST,MP,(MID)RM,(MI)LM
      7,//DLM<-ST,MP,(MID)RM,(MD)LM

      8,//BF<-ST,MP,(MID)(RL)M

      0,//BG<-
      2,//EN<-MRE,MRH
      0 //XX
    };
    return nParent[s];
  }
    inline static const State& getParent0(const State& s, const int& i) {
    static const State parents[NSTATE0][21] = {
      {  BF, XX, XX, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->ST
      {  ST, MP,MRB,IRB,DRB,MLB,ILB,DLB,MLI,ILI,
	DLI,MR1,IR1,DR1,ML1,IL1,DL1, XX, XX, XX, XX},//->MP

      { MRE,IRE,DRE, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->MRE
      { MRE,IRE, XX, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->IRE
      { MRE,DRE, XX, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->DRE

      {  MP,MRH,IRH,DRH, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->MRH
      {  MP,MRH,IRH, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->IRH
      {  MP,MRH,DRH, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->DRH

      {  MP,MRB,IRB,DRB, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->MRB
      {  MP,MRB,IRB, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->IRB
      {  MP,MRB,DRB, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->DRB
      {  MP,MLB,ILB,DLB, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->MLB
      {  MP,MLB,ILB, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->ILB
      {  MP,MLB,DLB, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->DLB

      {  MP,MRI,IRI,DRI, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->MRI
      {  MP,MRI,IRI, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->IRI
      {  MP,MRI,DRI, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->DRI
      {  MP,MRI,IRI,DRI,MLI,ILI,DLI, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->MLI
      {  MP,MRI,IRI,DRI,MLI,ILI, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->ILI
      {  MP,MRI,IRI,DRI,MLI,DLI, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->DLI

      {  ST,MR1,IR1,DR1, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->MR1
      {  ST,MR1,IR1, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->IR1
      {  ST,MR1,DR1, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->DR1
      {  ST,MR1,IR1,DR1,ML1,IL1,DL1, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->ML1
      {  ST,MR1,IR1,DR1,ML1,IL1, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->IL1
      {  ST,MR1,IR1,DR1,ML1,DL1, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->DL1

      {  ST, MP,MRM,IRM,DRM, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->MRM
      {  ST, MP,MRM,IRM, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->IRM
      {  ST, MP,MRM,DRM, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->DRM
      {  ST, MP,MRM,IRM,DRM,MLM,ILM,DLM, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->MLM
      {  ST, MP,MRM,IRM,DRM,MLM,ILM, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->ILM
      {  ST, MP,MRM,IRM,DRM,MLM,DLM, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->DLM

      {  ST, MP,MRM,IRM,DRM,MLM,ILM,DLM, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->BF

      {  XX, XX, XX, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->BG
      { MRE,MRH, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX},//->EN
      {  XX, XX, XX, XX, XX, XX, XX, XX, XX, XX,
	 XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX} //->XX
    };
    return parents[s][i];
  }
  static const int getNTSC00() {
    return NTSC0;
  }
  static const int getTSC0Index0(const State& s, const State& s1) {
    // should have accumulated number of NChildren
    int n = 0;
    for (int i = 0; i < s; i++) {
      n += getNChild(State(i));
    }
    n += getChildIndex(s, s1);
    return n;
  }
  // old codes end
  ///////////////////////////////////////////////////////////////////
#endif

#if defined(USE_PSCFG_BASE)

#undef ST 
#undef MP 
#undef MR 
#undef IR 
#undef DR 
#undef ML 
#undef IL 
#undef DL 

#ifdef PSCFG_BASE_HAIRPIN
#undef MRH    
#undef IRH        
#undef DRH 
#endif

#ifdef PSCFG_SHORT_BULGE
#undef MRB1
#undef MLB1
#endif
 
#undef BF 
#undef BG 
#undef EN 
#undef XX 

#else

#undef ST 
#undef MP 

#undef MRE
#undef IRE
#undef DRE

#undef MRH
#undef IRH
#undef DRH

#undef MRB
#undef IRB
#undef DRB
#undef MLB
#undef ILB
#undef DLB

#undef MRI
#undef IRI
#undef DRI
#undef MLI
#undef ILI
#undef DLI

#undef MR1
#undef IR1
#undef DR1
#undef ML1
#undef IL1
#undef DL1

#undef MRM
#undef IRM
#undef DRM
#undef MLM
#undef ILM
#undef DLM

#undef BF 

#undef BG 
#undef EN 
#undef XX 

#endif
};

class 
PSCFG_Energy::
Iterator 
{
public:
  State s;
  Iterator() : s(X_st), next_s(beginState()) {}
  void reset() {s = X_st; next_s = beginState();}
  bool getNext() {
    if (lastState() < next_s) return false;
  
    s = State(next_s++);
    return true;
  }
private:
  int next_s;
};
struct 
PSCFG_Energy::
TBNode 
{
  enum {NBYTE = TBNODE_NBYTE};
  CArray<unsigned char, NBYTE> data;
#ifdef DEBUG
  CArray<State,NSTATE> isSet;
#endif

  void setNull() {
    data.fill(0);
#ifdef DEBUG
    isSet.fill(X_st);
#endif
  }
  static const TBNode& null() {
    static const struct NullNode {
      TBNode nd;
      NullNode() {nd.setNull();}
    } nn;
    return nn.nd;
  }
  static int getNByte() {return NBYTE;}
  struct Conv {
    unsigned int index;
    unsigned int shift;
    unsigned int nbit;
    //unsigned int mask; // ~(~0 << nbit)
    //unsigned int mask_negate // ~(mask << shift)
  };


  struct ConvTable {
    typedef CArray<Conv, NSTATE> Table;
    Table table;
    inline ConvTable() {
      const unsigned int& bit_size = 8;// bit size of one byte
      unsigned int index = 0;
      unsigned int shift = 0;
      for (int s = 0; s < NSTATE; s++) {
	Conv& conv = table[s];
	unsigned int n = getModel().getNChild(State(s));
	unsigned int nbit = 0u;
	Assert(n <= (1u << bit_size));
	if (n > 1) {//n == 1 case is also represented by 0 bits
	  for (nbit = 1u; (1u << nbit) < n; nbit++) {
	    ;// no-op
	  }
	}
	conv.nbit = nbit;
	Assert(nbit <= bit_size);
	if (bit_size < (shift + nbit)) {
	  index++;
	  shift = 0;
	}
	conv.index = index;
	conv.shift = shift;
	shift += nbit;
      }
      Assert((index == (NBYTE - 1)) && (shift <= bit_size),
	     "index=%d,NBYTE=%d,shift=%d,bit_size=%d",
	     index, NBYTE, shift, bit_size);
    }
    ~ConvTable() {}
    inline const Conv& get(const State& s) const {return table[s];}
  };
  inline static const Conv& getConv(const State& s) {
    static const ConvTable convTable;
#ifdef DEBUG
#ifdef USE_PSCFG_ENERGY
    static const Conv conv[NSTATE] = {
      {1, 0, 4}, //14,//ST->MP,(MID)(RL)1,(MID)(RL)M,BF
      {0, 0, 5}, //20,//MP->MP,(MID)RH,(MID)(RL)B,(MID)RI,(MID)(RL)M,BF

      {1, 4, 2}, //3,//MRE->(MID)RE
      {1, 6, 1}, //2,//IRE->(MI)RE
      {1, 7, 1}, //2,//DRE->(MD)RE

      {2, 0, 2}, //3,//MRH->(MID)RH
      {2, 2, 1}, //2,//IRH->(MI)RH
      {2, 3, 1}, //2,//DRH->(MD)RH

      {2, 4, 2}, //4,//MRB->MP,(MID)RB
      {2, 6, 2}, //3,//IRB->MP,(MI)RB
      {3, 0, 2}, //3,//DRB->MP,(MD)RB
      {3, 2, 2}, //4,//MLB->MP,(MID)LB
      {3, 4, 2}, //3,//ILB->MP,(MI)LB
      {3, 6, 2}, //3,//DLB->MP,(MD)LB

      {4, 0, 3}, //6,//MRI->(MID)RI,(MID)LI
      {4, 3, 3}, //5,//IRI->(MI)RI,(MID)LI
      {5, 0, 3}, //5,//DRI->(MD)RI,(MID)LI
      {4, 6, 2}, //4,//MLI->MP,(MID)LI
      {5, 3, 2}, //3,//ILI->MP,(MI)LI
      {5, 5, 2}, //3,//DLI->MP,(MD)LI

      {6, 0, 3}, //7,//MR1->MP,(MID)R1,(MID)L1
      {6, 3, 3}, //6,//IR1->MP,(MI)R1,(MID)L1
      {7, 0, 3}, //6,//DR1->MP,(MD)R1,(MID)L1
      {6, 6, 2}, //4,//ML1->MP,(MID)L1
      {7, 3, 2}, //3,//IL1->MP,(MI)L1
      {7, 5, 2}, //3,//DL1->MP,(MD)L1

      {8, 0, 3}, //7,//MRM->(MID)RM,(MID)LM,BF
      {8, 3, 3}, //6,//IRM->(MI)RM,(MID)LM,BF
      {9, 0, 3}, //6,//DRM->(MD)RM,(MID)LM,BF
      {8, 6, 2}, //4,//MLM->(MID)LM,BF
      {9, 3, 2}, //3,//ILM->(MI)LM,BF
      {9, 5, 2}, //3,//DLM->(MD)LM,BF

      {9, 7, 1}  //1,//BF->ST
                 //4,//BG->ST,(MID)RE
                 //0,//EN->
                 //0 //XX->
    };
    Assert(conv[s] == convTable.get(s));
#endif
#endif
    return convTable.get(s);
  }
  bool isNull() const {return (*this == null());}

  bool operator==(const TBNode& tbn) const {
    if (!equal(data.begin(), data.end(), tbn.data.begin())) {
      return false;
    }
#ifdef DEBUG
    if (!equal(isSet.begin(), isSet.end(), tbn.isSet.begin())) {
      return false;
    }
#endif
    return true;
  }
 
  const State& getVal(const State& s) const {
    Assert(isNormalState(s));
    const Conv& c = getConv(s);
    const unsigned int& d = data[c.index];
    const unsigned int& mask = ~(~0 << c.nbit);
    const unsigned int& i = ((d >> c.shift) & mask);
    const State& s1 = getModel().getChild(s, i);
#ifdef DEBUG
    Assert(isSet[s] != X_st);
    Assert(isSet[s] == s1);
#endif
    return s1;
  }
  inline void setVal(const State& s, const int& childIndex) {
    const Conv& c = getConv(s);
    const unsigned int& mask = ~(~0 << c.nbit);
    const unsigned int& mask_negate = ~(mask << c.shift);
    unsigned char& d = data[c.index];
    d &= mask_negate;
    d |= (childIndex << c.shift);

#ifdef DEBUG
    State s1 = getModel().getChild(s, childIndex);
    Assert(getModel().isNormalState(s) && s != BF_st);
    Assert(getModel().isNormalState(s1));
    Assert(getModel().isChild(s, s1));
    isSet[s] = s1;
    Assert(getVal(s) == s1);
#endif
  }
  bool isValid() const {return true;}

#ifdef DEBUG
  bool valIsSet(const State& s) const {
    Assert(getModel().isNormalState(s));
    return (isSet[s] != X_st);
  }
  bool isAltered() const {
    Iterator it;
    while (it.getNext()) {
      if (valIsSet(it.s)) return true;
    }
    return false;
  }
#endif
  string toString() const {
    ostringstream oss;
    bool hasOne = false;
    oss << "[ ";
    for (int i = 0; i < NSTATE; i++) {
      const State s = State(i);
#ifdef DEBUG
      if (valIsSet(State(s))) {
#endif
	if (hasOne) {
	  oss << ", ";
	} else {
	  hasOne = true;
	}
	oss << getModel().state2str(s) << ": " << getVal(s);
#ifdef DEBUG
      }
#endif
    }
    oss << " ]";
    return oss.str();
  }
  void print() const {
    cout << toString() << flush;
  }
};

#endif
